//
// ElCel Technology Test Harness 
//
// Class: Thread
//




#include "ot/base/System.h"
#include "ot/io/Console.h"
#include "ot/base/Exception.h"
#include "ot/base/StringUtils.h"
#include "ot/util/AttributeListParser.h"

using namespace ot;
using namespace ot::io;
using namespace ot::util;

String getTestAttribute(const String& name);
void testMessage(const String& msg);
void testFailed(const String& test);
void testPassed(const String& test);
void goodCatch(const String& test, const String& eMsg);
void uncaughtException(const String& e, const String& test);


#include "ot/base/IllegalThreadStateException.h"
#include "ot/base/IllegalMonitorStateException.h"
#include "ot/base/InterruptedException.h"
#include "ot/base/Thread.h"
#include "ot/base/Runnable.h"
#include "ot/base/ConditionVariable.h"
#include "ot/base/RecursiveMutex.h"
#include <signal.h>

using namespace ot; 


#ifdef OT_MT

size_t GlobalCount = 0;
size_t Semaphore = 0;
ConditionVariable GlobalCV;
RecursiveMutex GlobalMutex;

//
// class: testInterrupt
//
// The basic idea behind this class is that it tests the Thread::interrupt
// functionality of OpenTop.
//
// A thread is started, which then goes into a sleep.  The main thread interrupts
// this thread, so the sleep() should throw an InterruptedException.  To overcome
// the race condition where the main thread interrupts this thread before the sleep()
// starts, the sleep is performed within a loop.
//
class testInterrupt : public Runnable
{
	virtual void run()
	{
		// test yield
		try
	{
		Thread::Yield(); testPassed(OT_T("Yield"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("Yield"));
	}
		// sleep for a long while (10 secs) to give main line a chance to interrupt us
		bool bInterrupted = false;
		for(int i=0; i<2 && !bInterrupted; ++i)
		{
			try
			{
				Thread::Sleep(10000);
			}
			catch(InterruptedException& /*e*/)
			{
				bInterrupted = true;
			}
		}
		try
	{
		if(bInterrupted) {testPassed(OT_T("thread interrupted"));} else {testFailed(OT_T("thread interrupted"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("thread interrupted"));
	}
	}
};

class testWaiter : public Runnable
{
	virtual void run()
	{
		try
	{
		if(!GlobalMutex.isLocked()) {testPassed(OT_T("isLocked"));} else {testFailed(OT_T("isLocked"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("isLocked"));
	}
		try
	{
		GlobalCV.wait(GlobalMutex);
		testFailed(OT_T("badWait"));
	}
	catch(IllegalMonitorStateException& e)
	{
		goodCatch(OT_T("badWait"), e.toString());
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("badWait"));
	}
		RecursiveMutex::Lock _lock1(GlobalMutex);
		try
	{
		if(GlobalMutex.isLocked()) {testPassed(OT_T("isLocked"));} else {testFailed(OT_T("isLocked"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("isLocked"));
	}
		{
			// create another level of recursion
			RecursiveMutex::Lock _lock2(GlobalMutex);
			GlobalCount++;
			while(!Semaphore)
			{
				try
	{
		GlobalCV.wait(GlobalMutex); testPassed(OT_T("wait"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("wait"));
	}
			}
			Semaphore--;
		}
	}
};

#endif //OT_MT

// Simple class to test use of OT_SYNCHRONIZED in 
// derived (client) classes
struct MySynchronizedObject : public SynchronizedObject
{
	bool doSynchronizedAction() const
	{
		OT_SYNCHRONIZED
		return true;
	}
};


void Thread_Tests()
{
	testMessage(OT_T("Starting tests for Thread"));

	try
	{
		; testPassed(OT_T("Thread::Sleep(10)"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("Thread::Sleep(10)"));
	}
	try
	{
		; testPassed(OT_T("Thread::Sleep(500,500)"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("Thread::Sleep(500,500)"));
	}

	// Synchronized object tests...
	RefPtr<MySynchronizedObject> rpSync;
	try
	{
		rpSync = new MySynchronizedObject; testPassed(OT_T("new Synchronized"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("new Synchronized"));
	}
	try
	{
		if(rpSync->doSynchronizedAction()) {testPassed(OT_T("Synchronized"));} else {testFailed(OT_T("Synchronized"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("Synchronized"));
	}

#ifdef OT_MT

#ifndef WIN32
	try
	{
		Thread::SetInterruptSignal(SIGUSR1); testPassed(OT_T("SetInterruptSignal"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("SetInterruptSignal"));
	}
#endif //WIN32

	RefPtr<Thread> rpThread = new Thread(new testInterrupt);
	// test the destruction of a thread before it has been started
	// by re-using the RefPtr for another object
	rpThread = new Thread(new testInterrupt);

	//
	// If the interrupt() call is supported, then start a thread
	// and attempt to interrupt it from its sleep.  The interrupt()
	// is placed within a loop to help avoid interrupting the 
	// thread before it has entered its sleep state.
	//
	if(rpThread->interruptSupported())
	{
		try
	{
		rpThread->start(); testPassed(OT_T("start"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("start"));
	}
		try
	{
		if(rpThread->isActive()) {testPassed(OT_T("isActive1"));} else {testFailed(OT_T("isActive1"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("isActive1"));
	}
		try
	{
		if(!rpThread->isDaemon()) {testPassed(OT_T("isDaemon"));} else {testFailed(OT_T("isDaemon"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("isDaemon"));
	}
		try
	{
		if(Thread::GetActiveCount()==1) {testPassed(OT_T("GetActiveCount"));} else {testFailed(OT_T("GetActiveCount"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("GetActiveCount"));
	}
		for(int i=0; i<5 && rpThread->isActive(); ++i)
		{
			try
	{
		rpThread->interrupt(); testPassed(OT_T("interrupt"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("interrupt"));
	}
			try
	{
		Thread::Sleep(1000); testPassed(OT_T("main sleep"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("main sleep"));
	}
		}
		try
	{
		rpThread->join(1000); testPassed(OT_T("join"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("join"));
	}
		try
	{
		if(!rpThread->isActive()) {testPassed(OT_T("isActive2"));} else {testFailed(OT_T("isActive2"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("isActive2"));
	}
		try
	{
		if(Thread::GetActiveCount()==0) {testPassed(OT_T("GetActiveCount0"));} else {testFailed(OT_T("GetActiveCount0"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("GetActiveCount0"));
	}
		try
	{
		rpThread->start();
		testFailed(OT_T("start2"));
	}
	catch(IllegalThreadStateException& e)
	{
		goodCatch(OT_T("start2"), e.toString());
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("start2"));
	}
	}

	const size_t waiterThreads = 10;
	for(size_t i=0; i<waiterThreads; ++i)
	{
		(new Thread(new testWaiter))->start();
	}

	// give the threads a chance to start
	// It would be better to use some form of barrier
	for(size_t i1=0; i1<10; ++i1)
	{
		RecursiveMutex::Lock _lock(GlobalMutex);
		if(GlobalCount==waiterThreads)
		{
			break;
		}
		try
	{
		_lock.unlock(); testPassed(OT_T("unlock"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("unlock"));
	}
		try
	{
		Thread::Sleep(500); testPassed(OT_T("sleep"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("sleep"));
	}
	}

	try
	{
		if(Thread::GetActiveCount()==waiterThreads) {testPassed(OT_T("GetActiveCount10"));} else {testFailed(OT_T("GetActiveCount10"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("GetActiveCount10"));
	}

	//
	// Now its time to start waking the threads up.
	// To guard against spurious wake-ups, we maintain a count like a semaphore
	//
	{
		RecursiveMutex::Lock _lock(GlobalMutex);
		Semaphore++;
		try
	{
		GlobalCV.signal(); testPassed(OT_T("signal"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("signal"));
	}
	}

	//
	// Wait for a signalled thread to terminate
	//
	for(size_t i2=0; i2<10; ++i2)
	{
		if(Thread::GetActiveCount()==(waiterThreads-1))
			break;
		// give the chosen thread a chance to die
		try
	{
		Thread::Sleep(500); testPassed(OT_T("sleep2"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("sleep2"));
	}
	}
	try
	{
		if(Thread::GetActiveCount()==(waiterThreads-1)) {testPassed(OT_T("GetActiveCount9"));} else {testFailed(OT_T("GetActiveCount9"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("GetActiveCount9"));
	}

	//
	// Now wake up the rest
	//
	{
		RecursiveMutex::Lock _lock(GlobalMutex);
		Semaphore+=(waiterThreads-1);
		try
	{
		GlobalCV.broadcast(); testPassed(OT_T("broadcast"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("broadcast"));
	}
	}

	//
	// give all the threads a chance to die
	//
	for(size_t i3=0; i3<10; ++i3)
	{
		if(Thread::GetActiveCount()==0)
			break;
		try
	{
		Thread::Sleep(500); testPassed(OT_T("sleep3"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("sleep3"));
	}
	}
	try
	{
		if(Thread::GetActiveCount()==0) {testPassed(OT_T("GetActiveCount3"));} else {testFailed(OT_T("GetActiveCount3"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("GetActiveCount3"));
	}

#endif //OT_MT


	testMessage(OT_T("End of tests for Thread"));
}

