//
// ElCel Technology Test Harness 
//
// Class: Thread
//

include(../common.m4)
OT_TEST_INIT

#`include' "ot/base/IllegalThreadStateException.h"
#`include' "ot/base/IllegalMonitorStateException.h"
#`include' "ot/base/InterruptedException.h"
#`include' "ot/base/Thread.h"
#`include' "ot/base/Runnable.h"
#`include' "ot/base/ConditionVariable.h"
#`include' "ot/base/RecursiveMutex.h"
#`include' <signal.h>

using namespace ot; 


`#ifdef OT_MT'

size_t GlobalCount = 0;
size_t Semaphore = 0;
ConditionVariable GlobalCV;
RecursiveMutex GlobalMutex;

//
// class: testInterrupt
//
// The basic idea behind this class is that it tests the Thread::interrupt
// functionality of OpenTop.
//
// A thread is started, which then goes into a sleep.  The main thread interrupts
// this thread, so the sleep() should throw an InterruptedException.  To overcome
// the race condition where the main thread interrupts this thread before the sleep()
// starts, the sleep is performed within a loop.
//
class testInterrupt : public Runnable
{
	virtual void run()
	{
		// test yield
		OT_TEST_VOID(Yield, Thread::Yield())
		// sleep for a long while (10 secs) to give main line a chance to interrupt us
		bool bInterrupted = false;
		for(int i=0; i<2 && !bInterrupted; ++i)
		{
			try
			{
				Thread::Sleep(10000);
			}
			catch(InterruptedException& /*e*/)
			{
				bInterrupted = true;
			}
		}
		OT_TEST_BOOL(thread interrupted, bInterrupted)
	}
};

class testWaiter : public Runnable
{
	virtual void run()
	{
		OT_TEST_BOOL(isLocked, !GlobalMutex.isLocked())
		OT_TEST_THROW(badWait, GlobalCV.wait(GlobalMutex), IllegalMonitorStateException)
		RecursiveMutex::Lock _lock1(GlobalMutex);
		OT_TEST_BOOL(isLocked, GlobalMutex.isLocked())
		{
			// create another level of recursion
			RecursiveMutex::Lock _lock2(GlobalMutex);
			GlobalCount++;
			while(!Semaphore)
			{
				OT_TEST_VOID(wait, GlobalCV.wait(GlobalMutex))
			}
			Semaphore--;
		}
	}
};

`#endif //OT_MT'

// Simple class to test use of OT_SYNCHRONIZED in 
// derived (client) classes
struct MySynchronizedObject : public SynchronizedObject
{
	bool doSynchronizedAction() const
	{
		OT_SYNCHRONIZED
		return true;
	}
};

OT_START_TESTS(`Thread')

	OT_TEST_VOID(Thread::Sleep(10))
	OT_TEST_VOID(Thread::Sleep(500,500))

	// Synchronized object tests...
	RefPtr<MySynchronizedObject> rpSync;
	OT_TEST_VOID(new Synchronized, rpSync = new MySynchronizedObject)
	OT_TEST_BOOL(Synchronized, rpSync->doSynchronizedAction())

`#ifdef OT_MT'

`#ifndef WIN32'
	OT_TEST_VOID(SetInterruptSignal, Thread::SetInterruptSignal(SIGUSR1))
`#endif //WIN32'

	RefPtr<Thread> rpThread = new Thread(new testInterrupt);
	// test the destruction of a thread before it has been started
	// by re-using the RefPtr for another object
	rpThread = new Thread(new testInterrupt);

	//
	// If the interrupt() call is supported, then start a thread
	// and attempt to interrupt it from its sleep.  The interrupt()
	// is placed within a loop to help avoid interrupting the 
	// thread before it has entered its sleep state.
	//
	if(rpThread->interruptSupported())
	{
		OT_TEST_VOID(start, rpThread->start())
		OT_TEST_BOOL(isActive1, rpThread->isActive())
		OT_TEST_BOOL(isDaemon, !rpThread->isDaemon())
		OT_TEST_BOOL(GetActiveCount, Thread::GetActiveCount()==1)
		for(int i=0; i<5 && rpThread->isActive(); ++i)
		{
			OT_TEST_VOID(interrupt, rpThread->interrupt())
			OT_TEST_VOID(main sleep, Thread::Sleep(1000))
		}
		OT_TEST_VOID(join, rpThread->join(1000))
		OT_TEST_BOOL(isActive2, !rpThread->isActive())
		OT_TEST_BOOL(GetActiveCount0, Thread::GetActiveCount()==0)
		OT_TEST_THROW(start2, rpThread->start(), IllegalThreadStateException)
	}

	const size_t waiterThreads = 10;
	for(size_t i=0; i<waiterThreads; ++i)
	{
		(new Thread(new testWaiter))->start();
	}

	// give the threads a chance to start
	// It would be better to use some form of barrier
	for(size_t i1=0; i1<10; ++i1)
	{
		RecursiveMutex::Lock _lock(GlobalMutex);
		if(GlobalCount==waiterThreads)
		{
			break;
		}
		OT_TEST_VOID(unlock, _lock.unlock())
		OT_TEST_VOID(sleep, Thread::Sleep(500))
	}

	OT_TEST_BOOL(GetActiveCount10, Thread::GetActiveCount()==waiterThreads)

	//
	// Now its time to start waking the threads up.
	// To guard against spurious wake-ups, we maintain a count like a semaphore
	//
	{
		RecursiveMutex::Lock _lock(GlobalMutex);
		Semaphore++;
		OT_TEST_VOID(signal, GlobalCV.signal())
	}

	//
	// Wait for a signalled thread to terminate
	//
	for(size_t i2=0; i2<10; ++i2)
	{
		if(Thread::GetActiveCount()==(waiterThreads-1))
			break;
		// give the chosen thread a chance to die
		OT_TEST_VOID(sleep2, Thread::Sleep(500))
	}
	OT_TEST_BOOL(GetActiveCount9, Thread::GetActiveCount()==(waiterThreads-1))

	//
	// Now wake up the rest
	//
	{
		RecursiveMutex::Lock _lock(GlobalMutex);
		Semaphore+=(waiterThreads-1);
		OT_TEST_VOID(broadcast, GlobalCV.broadcast())
	}

	//
	// give all the threads a chance to die
	//
	for(size_t i3=0; i3<10; ++i3)
	{
		if(Thread::GetActiveCount()==0)
			break;
		OT_TEST_VOID(sleep3, Thread::Sleep(500))
	}
	OT_TEST_BOOL(GetActiveCount3, Thread::GetActiveCount()==0)

`#endif //OT_MT'

OT_END_TESTS(`Thread')

