//
// ElCel Technology Test Harness
//
// Class: FileInputStream
//





#include "ot/base/System.h"
#include "ot/io/Console.h"
#include "ot/base/Exception.h"
#include "ot/base/StringUtils.h"
#include "ot/util/AttributeListParser.h"

using namespace ot;
using namespace ot::io;
using namespace ot::util;

String getTestAttribute(const String& name);
void testMessage(const String& msg);
void testFailed(const String& test);
void testPassed(const String& test);
void goodCatch(const String& test, const String& eMsg);
void uncaughtException(const String& e, const String& test);


#include "ot/io/FileInputStream.h"
#include "ot/io/FileOutputStream.h"
#include "ot/io/File.h"
#include "ot/io/IOException.h"
#include "ot/io/FileNotFoundException.h"

using namespace ot::io;


void FileInputStream_Tests()
{
	testMessage(OT_T("Starting tests for FileInputStream"));

	//
	// Basic API tests
	//

	File badFile1(OT_T(""));
	File badFile2(OT_T("."));
	File testFile(OT_T("test.out"));

	try
	{
		FileInputStream x(badFile1);
		testFailed(OT_T("badFile1"));
	}
	catch(IOException& e)
	{
		goodCatch(OT_T("badFile1"), e.toString());
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("badFile1"));
	}
	try
	{
		FileInputStream x(badFile2);
		testFailed(OT_T("badFile2"));
	}
	catch(IOException& e)
	{
		goodCatch(OT_T("badFile2"), e.toString());
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("badFile2"));
	}
	try
	{
		FileInputStream x(testFile);
		testFailed(OT_T("testFile"));
	}
	catch(FileNotFoundException& e)
	{
		goodCatch(OT_T("testFile"), e.toString());
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("testFile"));
	}

	//
	// create a small file to work on
	//
	Byte buffer[3] = {1, 2, 3};
	RefPtr<FileOutputStream> rpOut;
	try
	{
		rpOut = new FileOutputStream(testFile); testPassed(OT_T("new1"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("new1"));
	}
	if(rpOut)
	{
		try
	{
		rpOut->write(buffer, sizeof(buffer)); testPassed(OT_T("write"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("write"));
	}
		try
	{
		rpOut->close(); testPassed(OT_T("close1"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("close1"));
	}
		try
	{
		if(testFile.exists()) {testPassed(OT_T("exists"));} else {testFailed(OT_T("exists"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("exists"));
	}
		try
	{
		if(testFile.length()==sizeof(buffer)) {testPassed(OT_T("length"));} else {testFailed(OT_T("length"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("length"));
	}
	}

	RefPtr<FileInputStream> rpTest;
	try
	{
		rpTest = new FileInputStream(testFile); testPassed(OT_T("new2"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("new2"));
	}
	if(rpTest)
	{
		// OpenTop API states it returns zero for available on FileInputStream
		try
	{
		if(rpTest->available()==0) {testPassed(OT_T("available"));} else {testFailed(OT_T("available"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("available"));
	}
		try
	{
		if(rpTest->markSupported()==false) {testPassed(OT_T("markSupported"));} else {testFailed(OT_T("markSupported"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("markSupported"));
	}
		try
	{
		rpTest->mark(1);
		testFailed(OT_T("mark"));
	}
	catch(IOException& e)
	{
		goodCatch(OT_T("mark"), e.toString());
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("mark"));
	}
		try
	{
		rpTest->reset();
		testFailed(OT_T("reset"));
	}
	catch(IOException& e)
	{
		goodCatch(OT_T("reset"), e.toString());
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("reset"));
	}

		Byte input[3];

		try
	{
		if(rpTest->read()==1) {testPassed(OT_T("read1"));} else {testFailed(OT_T("read1"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("read1"));
	}
		try
	{
		if(rpTest->read(input+1, 2)==2) {testPassed(OT_T("read2"));} else {testFailed(OT_T("read2"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("read2"));
	}
		try
	{
		if(*(input+1)==2) {testPassed(OT_T("read3"));} else {testFailed(OT_T("read3"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("read3"));
	}
		try
	{
		if(rpTest->skip(1)==0) {testPassed(OT_T("skip"));} else {testFailed(OT_T("skip"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("skip"));
	}
		try
	{
		if(rpTest->read()==InputStream::EndOfFile) {testPassed(OT_T("eof"));} else {testFailed(OT_T("eof"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("eof"));
	}
		try
	{
		rpTest->close(); testPassed(OT_T("close2"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("close2"));
	}
	}

	// Make the file read-only then read it again
	try
	{
		testFile.setReadOnly(true); testPassed(OT_T("setReadOnly"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("setReadOnly"));
	};
	try
	{
		rpTest.release(); rpTest = new FileInputStream(testFile); testPassed(OT_T("open stream"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("open stream"));
	}
	if(rpTest)
	{
		try
	{
		if(rpTest->read()==1) {testPassed(OT_T("read4"));} else {testFailed(OT_T("read4"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("read4"));
	}
		try
	{
		rpTest->close(); testPassed(OT_T("close3"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("close3"));
	}
	}

	// Make the file writable then delete it
	try
	{
		testFile.setReadOnly(false); testPassed(OT_T("setReadOnly"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("setReadOnly"));
	};
	try
	{
		testFile.deleteFile(); testPassed(OT_T("delete"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("delete"));
	}


	testMessage(OT_T("End of tests for FileInputStream"));
}

