//
// ElCel Technology Test Harness
//
// Class: InputStreamReader
//





#include "ot/base/System.h"
#include "ot/io/Console.h"
#include "ot/base/Exception.h"
#include "ot/base/StringUtils.h"
#include "ot/util/AttributeListParser.h"

using namespace ot;
using namespace ot::io;
using namespace ot::util;

String getTestAttribute(const String& name);
void testMessage(const String& msg);
void testFailed(const String& test);
void testPassed(const String& test);
void goodCatch(const String& test, const String& eMsg);
void uncaughtException(const String& e, const String& test);


#include "ot/io/InputStreamReader.h"
#include "ot/io/FileInputStream.h"
#include "ot/io/FileOutputStream.h"
#include "ot/io/File.h"
#include "ot/io/IOException.h"
#include "ot/io/UnsupportedEncodingException.h"
#include "ot/base/NullPointerException.h"

using namespace ot::io;


void InputStreamReader_Tests()
{
	testMessage(OT_T("Starting tests for InputStreamReader"));

	//
	// Basic API tests
	//

	File testFile(OT_T("test.out"));

	//
	// create a small file to work on
	//
	Byte buffer[3] = {'a', 'b', 'c'};
	RefPtr<FileOutputStream> rpOut = new FileOutputStream(testFile);
	try
	{
		rpOut->write(buffer, sizeof(buffer)); testPassed(OT_T("write"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("write"));
	}
	try
	{
		rpOut->close(); testPassed(OT_T("close"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("close"));
	}
	try
	{
		if(testFile.exists()) {testPassed(OT_T("exists"));} else {testFailed(OT_T("exists"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("exists"));
	}
	try
	{
		if(testFile.length()==sizeof(buffer)) {testPassed(OT_T("length"));} else {testFailed(OT_T("length"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("length"));
	}

	RefPtr<FileInputStream> rpIn = new FileInputStream(testFile);

	try
	{
		InputStreamReader x(rpIn.get(), OT_T("qwertyuiop"));
		testFailed(OT_T("unsupported"));
	}
	catch(UnsupportedEncodingException& e)
	{
		goodCatch(OT_T("unsupported"), e.toString());
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("unsupported"));
	}
	try
	{
		InputStreamReader x(0);
		testFailed(OT_T("null"));
	}
	catch(NullPointerException& e)
	{
		goodCatch(OT_T("null"), e.toString());
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("null"));
	}
	try
	{
		InputStreamReader x(rpIn.get(), (CodeConverter*)0);
		testFailed(OT_T("null"));
	}
	catch(NullPointerException& e)
	{
		goodCatch(OT_T("null"), e.toString());
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("null"));
	}

	RefPtr<InputStreamReader> rpTest = new InputStreamReader(rpIn.get(), OT_T("us-ascii"));

	try
	{
		if(rpTest->markSupported()==false) {testPassed(OT_T("markSupported"));} else {testFailed(OT_T("markSupported"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("markSupported"));
	}
	try
	{
		rpTest->mark(1);
		testFailed(OT_T("mark"));
	}
	catch(IOException& e)
	{
		goodCatch(OT_T("mark"), e.toString());
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("mark"));
	}
	try
	{
		rpTest->reset();
		testFailed(OT_T("reset"));
	}
	catch(IOException& e)
	{
		goodCatch(OT_T("reset"), e.toString());
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("reset"));
	}

	CharType input[3];

	try
	{
		if(rpTest->read()=='a') {testPassed(OT_T("read"));} else {testFailed(OT_T("read"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("read"));
	}
	try
	{
		if(rpTest->read(input+1, 2)==2) {testPassed(OT_T("read"));} else {testFailed(OT_T("read"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("read"));
	}
	try
	{
		if(*(input+1)=='b') {testPassed(OT_T("read"));} else {testFailed(OT_T("read"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("read"));
	}
	try
	{
		if(rpTest->skip(1)==0) {testPassed(OT_T("skip"));} else {testFailed(OT_T("skip"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("skip"));
	}
	try
	{
		if(rpTest->read()==InputStream::EndOfFile) {testPassed(OT_T("eof"));} else {testFailed(OT_T("eof"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("eof"));
	}
	try
	{
		if(rpTest->readAtomic().isEOF()) {testPassed(OT_T("eof"));} else {testFailed(OT_T("eof"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("eof"));
	}
	try
	{
		rpTest->close(); testPassed(OT_T("close"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("close"));
	}

	// delete the file, ready for next time
	try
	{
		testFile.deleteFile(); testPassed(OT_T("delete"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("delete"));
	}


	testMessage(OT_T("End of tests for InputStreamReader"));
}

