//
// ElCel Technology Test Harness
//
// Class: Stream
//




#include "ot/base/System.h"
#include "ot/io/Console.h"
#include "ot/base/Exception.h"
#include "ot/base/StringUtils.h"
#include "ot/util/AttributeListParser.h"

using namespace ot;
using namespace ot::io;
using namespace ot::util;

String getTestAttribute(const String& name);
void testMessage(const String& msg);
void testFailed(const String& test);
void testPassed(const String& test);
void goodCatch(const String& test, const String& eMsg);
void uncaughtException(const String& e, const String& test);


#include "ot/base/Character.h"
#include "ot/io/AtomicReadException.h"
#include "ot/io/BufferedInputStream.h"
#include "ot/io/BufferedOutputStream.h"
#include "ot/io/BufferedReader.h"
#include "ot/io/BufferedWriter.h"
#include "ot/io/Reader.h"
#include "ot/io/Writer.h"
#include "ot/io/File.h"
#include "ot/io/IOException.h"
#include "ot/io/FileOutputStream.h"
#include "ot/io/FileInputStream.h"
#include "ot/io/OutputStreamWriter.h"
#include "ot/io/InputStreamReader.h"
#include "ot/io/StringReader.h"
#include "ot/io/StringWriter.h"
#include "ot/io/FilterReader.h"
#include "ot/base/NumUtils.h"
#include "ot/base/IllegalCharacterException.h"
#include "ot/io/CharacterCodingException.h"
#include "ot/io/MalformedInputException.h"

using namespace ot::io;

//
// write a range of characters to a Writer
//
void writeCharRange(Writer* pWriter, unsigned long firstChar, unsigned long lastChar)
{
	RefPtr<BufferedWriter> rpWriter = new BufferedWriter(pWriter);
	for (unsigned long i=firstChar; i<=lastChar; i++)
	{
		if(i<0xD800UL || i>0xDFFFUL)
		{
			Character ch(i);
			rpWriter->write(ch.data(), ch.length());
		}
	}
}

//
// Helper function to generate a file with a name, encoding and filled
// with characters from firstChar  - lastChar
//
void genFile(const File& file, const String& enc, unsigned long firstChar, unsigned long lastChar)
{
	RefPtr<OutputStream> rpOS = new FileOutputStream(file);
	rpOS = new BufferedOutputStream(rpOS.get());
	RefPtr<Writer> rpWriter = new OutputStreamWriter(rpOS.get(), enc, true);
	rpWriter = new BufferedWriter(rpWriter.get());
	try
	{
		writeCharRange(rpWriter.get(), firstChar, lastChar); testPassed(OT_T("write range"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("write range"));
	}
}

//
// Helper function to check that all methods return the correct value
// when  a character stream is at EOF
//
void testEOF(Reader* pReader)
{
	// Ensure all read methods agree we have reached the end of file
	CharType buffer[6];
	try
	{
		if(pReader->readAtomic(buffer, 6)==Reader::EndOfFile) {testPassed(OT_T("read"));} else {testFailed(OT_T("read"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("read"));
	}
	try
	{
		if(pReader->read(buffer, 6)==Reader::EndOfFile) {testPassed(OT_T("read"));} else {testFailed(OT_T("read"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("read"));
	}
	try
	{
		if(pReader->read()==Reader::EndOfFile) {testPassed(OT_T("read"));} else {testFailed(OT_T("read"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("read"));
	}
	try
	{
		if(pReader->skip(1)==0) {testPassed(OT_T("read"));} else {testFailed(OT_T("read"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("read"));
	}
	try
	{
		if(pReader->skipAtomic(1)==0) {testPassed(OT_T("read"));} else {testFailed(OT_T("read"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("read"));
	}
}

//
// Helper function to check the contents returned from a Reader
//
bool testReaderContents(Reader* pReader, unsigned long firstChar, unsigned long lastChar)
{
	bool bRet = true;
	unsigned long i = firstChar;
	while(true)
	{
		Character ch = pReader->readAtomic();
		if(ch == Character::EndOfFileCharacter)
		{
			testEOF(pReader);
			break;
		}
		if(ch.toUnicode() != i || i > lastChar)
		{
			String msg = OT_T("mismatch, got=");
			msg += NumUtils::ToString(ch.toUnicode());
			msg += OT_T(" expected=");
			msg += NumUtils::ToString(i);
			testMessage(msg);
			bRet = false;
			break;
		}
		i++;
		// skip the surrogate pair range
		if(i==0xD800) i=0xE000;
	}
	bRet = bRet && (i==lastChar+1);
	if(!bRet)
	{
		String msg = OT_T("testReaderContents failed, i=");
		msg += NumUtils::ToString(i);
		testMessage(msg);
	}
	return bRet;
}

//
// Helper function to read a file with a name, encoding and filled
// with characters from firstChar - lastChar
//
void testFileContents(const File& file, const String& enc, unsigned long firstChar, unsigned long lastChar)
{
	RefPtr<InputStream> rpIS = new FileInputStream(file);
	RefPtr<Reader> rpReader = new InputStreamReader(rpIS.get(), enc, true);
		// test again using buffering
	rpIS = new FileInputStream(file);
	rpReader = new BufferedReader(new InputStreamReader(rpIS.get(), enc, true));
	try
	{
		if(testReaderContents(rpReader.get(), firstChar, lastChar)) {testPassed(OT_T("testReaderContents"));} else {testFailed(OT_T("testReaderContents"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("testReaderContents"));
	}
}

//
// test that opererations are correctly disallowed on closed readers
//
void testClosedOps(Reader* pReader)
{
	// It should be okay to close a closed stream
	try
	{
		pReader->close(); testPassed(OT_T("closed1"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("closed1"));
	}
	try
	{
		pReader->close(); testPassed(OT_T("closed2"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("closed2"));
	}
	// Once closed, the following methods are not allowed: read, mark, reset
	CharType buffer[10];
	size_t bufLen=10;
	try
	{
		pReader->readAtomic();
		testFailed(OT_T("closed3"));
	}
	catch(IOException& e)
	{
		goodCatch(OT_T("closed3"), e.toString());
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("closed3"));
	}
	try
	{
		pReader->readAtomic(buffer, bufLen);
		testFailed(OT_T("closed4"));
	}
	catch(IOException& e)
	{
		goodCatch(OT_T("closed4"), e.toString());
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("closed4"));
	}
	try
	{
		pReader->read();
		testFailed(OT_T("closed5"));
	}
	catch(IOException& e)
	{
		goodCatch(OT_T("closed5"), e.toString());
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("closed5"));
	}
	try
	{
		pReader->read(buffer, bufLen);
		testFailed(OT_T("closed6"));
	}
	catch(IOException& e)
	{
		goodCatch(OT_T("closed6"), e.toString());
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("closed6"));
	}
	try
	{
		pReader->mark(4);
		testFailed(OT_T("closed7"));
	}
	catch(IOException& e)
	{
		goodCatch(OT_T("closed7"), e.toString());
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("closed7"));
	}
	try
	{
		pReader->reset();
		testFailed(OT_T("closed8"));
	}
	catch(IOException& e)
	{
		goodCatch(OT_T("closed8"), e.toString());
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("closed8"));
	}
	try
	{
		pReader->skip(1);
		testFailed(OT_T("closed9"));
	}
	catch(IOException& e)
	{
		goodCatch(OT_T("closed9"), e.toString());
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("closed9"));
	}
	try
	{
		pReader->skipAtomic(1);
		testFailed(OT_T("closedA"));
	}
	catch(IOException& e)
	{
		goodCatch(OT_T("closedA"), e.toString());
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("closedA"));
	}
}

//
// Perform a bunch of atomic operations on a Reader that
// is positioned on a high unicode character
//
void atomicReaderTests(Reader* pReader)
{
#ifndef OT_UCS4
	CharType buffer[6];
	// test rc=0 for readAtomic with too small buffer
	try
	{
		if(pReader->readAtomic(buffer, 1)==0) {testPassed(OT_T("readAtomic"));} else {testFailed(OT_T("readAtomic"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("readAtomic"));
	}
	try
	{
		if(pReader->skipAtomic(1)==1) {testPassed(OT_T("skipAtomic"));} else {testFailed(OT_T("skipAtomic"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("skipAtomic"));
	}
	try
	{
		if(pReader->readAtomic().toUnicode()==0x10001UL) {testPassed(OT_T("readAtomic2"));} else {testFailed(OT_T("readAtomic2"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("readAtomic2"));
	}
	// skip to an invalid multi-char seq boundary
	try
	{
		if(pReader->skip(1)==1) {testPassed(OT_T("skip"));} else {testFailed(OT_T("skip"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("skip"));
	}
		try
	{
		pReader->skipAtomic(1);
		testFailed(OT_T("skipAtomic"));
	}
	catch(IOException& e)
	{
		goodCatch(OT_T("skipAtomic"), e.toString());
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("skipAtomic"));
	}
#endif //OT_UCS4
}

//
// Perform a bunch of atomic operations on a file with the given encoding
//
void atomicFileTests(const String& enc)
{
	File atomic(OT_T("atomic_file.txt"));
	genFile(atomic, enc, 0x10000, 0x10FFFF);
	try
	{
		if(atomic.isFile()) {testPassed(OT_T("atomic isFile"));} else {testFailed(OT_T("atomic isFile"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("atomic isFile"));
	}

	RefPtr<InputStream> rpIS = new FileInputStream(atomic);
	RefPtr<Reader> rpReader = new InputStreamReader(rpIS.get(), enc, true);
	atomicReaderTests(rpReader.get());
	try
	{
		rpReader->close(); testPassed(OT_T("close"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("close"));
	}
	testClosedOps(rpReader.get());

	rpIS = new FileInputStream(atomic);
	rpReader = new BufferedReader(new InputStreamReader(rpIS.get(), enc, true));
	atomicReaderTests(rpReader.get());
	try
	{
		rpReader->close(); testPassed(OT_T("close"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("close"));
	}
	testClosedOps(rpReader.get());

	try
	{
		atomic.deleteFile(); testPassed(OT_T("delete atomic"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("delete atomic"));
	}
	try
	{
		if(atomic.isFile()==false) {testPassed(OT_T("atomic isFile"));} else {testFailed(OT_T("atomic isFile"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("atomic isFile"));
	}
	try
	{
		if(atomic.exists()==false) {testPassed(OT_T("atomic exists"));} else {testFailed(OT_T("atomic exists"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("atomic exists"));
	}
}

//
// test a variety of Readers which have a controlled
// stream of "Hello World!"
//
void markMyWorld(Reader* pReader)
{
	try
	{
		if(pReader->markSupported()==true) {testPassed(OT_T("markSupported"));} else {testFailed(OT_T("markSupported"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("markSupported"));
	}
	try
	{
		pReader->mark(4); testPassed(OT_T("mark"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("mark"));
	}
	try
	{
		if(pReader->read()=='H') {testPassed(OT_T("read"));} else {testFailed(OT_T("read"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("read"));
	}
	try
	{
		if(pReader->read()=='e') {testPassed(OT_T("read"));} else {testFailed(OT_T("read"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("read"));
	}
	try
	{
		if(pReader->readAtomic().toUnicode()=='l') {testPassed(OT_T("readAtomic"));} else {testFailed(OT_T("readAtomic"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("readAtomic"));
	}
	try
	{
		pReader->reset(); testPassed(OT_T("reset7"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("reset7"));
	}
	try
	{
		if(pReader->read()=='H') {testPassed(OT_T("read"));} else {testFailed(OT_T("read"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("read"));
	}
	try
	{
		if(pReader->skip(11)==11) {testPassed(OT_T("skip"));} else {testFailed(OT_T("skip"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("skip"));
	}
	testEOF(pReader);
}

//
// test that opererations are correctly disallowed on closed writers
//
void testClosedOps(Writer* pWriter)
{
	// It should be okay to close a closed stream
	try
	{
		pWriter->close(); testPassed(OT_T("close"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("close"));
	}
	try
	{
		pWriter->close(); testPassed(OT_T("close"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("close"));
	}
	CharType buffer[10];
	size_t bufLen=10;
	try
	{
		pWriter->write(buffer, bufLen);
		testFailed(OT_T("write"));
	}
	catch(IOException& e)
	{
		goodCatch(OT_T("write"), e.toString());
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("write"));
	}
	try
	{
		pWriter->write(OT_T("hello"));
		testFailed(OT_T("write"));
	}
	catch(IOException& e)
	{
		goodCatch(OT_T("write"), e.toString());
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("write"));
	}
	try
	{
		pWriter->write('H');
		testFailed(OT_T("write"));
	}
	catch(IOException& e)
	{
		goodCatch(OT_T("write"), e.toString());
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("write"));
	}
	try
	{
		pWriter->flush();
		testFailed(OT_T("flush"));
	}
	catch(IOException& e)
	{
		goodCatch(OT_T("flush"), e.toString());
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("flush"));
	}
}


void Stream_Tests()
{
	testMessage(OT_T("Starting tests for Stream"));

	//
	// create three files using the common unicode encodings
	// utf-8, utf-16 and iso-8859-1
	//
	File utf8(OT_T("test_utf8"));
	File utf16(OT_T("test_utf16"));
	File iso88591(OT_T("test_iso8859_1"));

	genFile(utf8, OT_T("utf-8"), 1, 0x10FFFFUL);
	testFileContents(utf8, OT_T("utf-8"), 1, 0x10FFFFUL);

	genFile(utf16, OT_T("utf-16"), 1, 0x10FFFFUL);
	testFileContents(utf16, OT_T("utf-16"), 1, 0x10FFFFUL);

	genFile(iso88591, OT_T("iso-8859-1"), 1, 0x00FEUL);
	testFileContents(iso88591, OT_T("iso-8859-1"), 1, 0x00FEUL);

	//
	// test ability to write illegal surrogate value
	// with strict turned on and then off
	//
	File badutf16(OT_T("badutf16.txt"));
	RefPtr<OutputStreamWriter> badWriter = new OutputStreamWriter(new FileOutputStream(badutf16), OT_T("utf-16"), true);
#ifdef OT_UTF16
	try
	{
		Character bad(0xD800);
		testFailed(OT_T("illegal surrogate"));
	}
	catch(IllegalCharacterException& e)
	{
		goodCatch(OT_T("illegal surrogate"), e.toString());
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("illegal surrogate"));
	}
#else
	Character bad(0xD800);
	try
	{
		badWriter->write(bad.data(), bad.length());
		testFailed(OT_T("surrogate"));
	}
	catch(CharacterCodingException& e)
	{
		goodCatch(OT_T("surrogate"), e.toString());
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("surrogate"));
	}
	try
	{
		badWriter->getEncoder()->setUnmappableCharAction(CodeConverter::replace); testPassed(OT_T("surrogate2"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("surrogate2"));
	}
	try
	{
		; testPassed(OT_T("surrogate3 badWriter->write(bad.data(), bad.length())"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("surrogate3 badWriter->write(bad.data(), bad.length())"));
	}
#endif
	try
	{
		badWriter->close(); testPassed(OT_T("close"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("close"));
	}
	try
	{
		badutf16.deleteFile(); testPassed(OT_T("deleteFile"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("deleteFile"));
	}

	//
	// test ability to handle illegal utf-8 sequences
	//
	File badutf8(OT_T("badutf8.txt"));
	RefPtr<OutputStream> rpBadUTF8OStream  = new FileOutputStream(badutf8);
	Byte bad_text_1[] = {0x80}; // illegal start char
	Byte bad_text_2[] = {0xf0, 0x80, 0x80, 0x80}; // illegal null char
	Byte bad_text_3[] = {0xf0, 0x81, 0x81}; // premature eof

	try
	{
		rpBadUTF8OStream->write(bad_text_1, sizeof(bad_text_1)); testPassed(OT_T("write 1"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("write 1"));
	}
	try
	{
		rpBadUTF8OStream->write(bad_text_2, sizeof(bad_text_2)); testPassed(OT_T("write 2"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("write 2"));
	}
	try
	{
		rpBadUTF8OStream->write(bad_text_3, sizeof(bad_text_3)); testPassed(OT_T("write 3"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("write 3"));
	}
	try
	{
		rpBadUTF8OStream->close(); testPassed(OT_T("close"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("close"));
	}
	//
	// Now, let's see what happens when we read it!
	// We need strict checking switched on to start with
	//
	RefPtr<InputStreamReader> rpBadUtf8Reader = new InputStreamReader(new FileInputStream(badutf8), OT_T("utf-8"), true);
	
	// If the exception is correctly thrown, then the Reader's position should not have changed
	// which means there are three tokens left to read
	try
	{
		rpBadUtf8Reader->readAtomic();
		testFailed(OT_T("illegal start"));
	}
	catch(IOException& e)
	{
		goodCatch(OT_T("illegal start"), e.toString());
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("illegal start"));
	}
	try
	{
		rpBadUtf8Reader->getDecoder()->setInvalidCharAction(CodeConverter::replace); testPassed(OT_T("change converter"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("change converter"));
	}
	try
	{
		if(rpBadUtf8Reader->readAtomic().toUnicode() == 0xFFFD) {testPassed(OT_T("test repl1"));} else {testFailed(OT_T("test repl1"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("test repl1"));
	}
	try
	{
		if(rpBadUtf8Reader->readAtomic().toUnicode() == 0xFFFD) {testPassed(OT_T("test repl2"));} else {testFailed(OT_T("test repl2"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("test repl2"));
	}
	try
	{
		if(rpBadUtf8Reader->readAtomic().toUnicode() == 0xFFFD) {testPassed(OT_T("test repl3"));} else {testFailed(OT_T("test repl3"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("test repl3"));
	}
	try
	{
		if(rpBadUtf8Reader->readAtomic() == Character::EndOfFileCharacter) {testPassed(OT_T("test eof"));} else {testFailed(OT_T("test eof"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("test eof"));
	}

	try
	{
		rpBadUtf8Reader->close(); testPassed(OT_T("close"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("close"));
	}
	testClosedOps(rpBadUtf8Reader.get());
	try
	{
		badutf8.deleteFile(); testPassed(OT_T("deleteFile"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("deleteFile"));
	}

	//
	// Test mark/reset operations
	//

	RefPtr<FileInputStream> rpIS1 = new FileInputStream(utf8);
	try
	{
		if(rpIS1->markSupported()==false) {testPassed(OT_T("markSupported"));} else {testFailed(OT_T("markSupported"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("markSupported"));
	}
	try
	{
		rpIS1->mark(1);
		testFailed(OT_T("mark"));
	}
	catch(IOException& e)
	{
		goodCatch(OT_T("mark"), e.toString());
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("mark"));
	}
	try
	{
		rpIS1->reset();
		testFailed(OT_T("reset"));
	}
	catch(IOException& e)
	{
		goodCatch(OT_T("reset"), e.toString());
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("reset"));
	}

	RefPtr<BufferedInputStream> rpBufIS = new BufferedInputStream(rpIS1.get(), 200);
	try
	{
		if(rpBufIS->markSupported()==true) {testPassed(OT_T("markSupported"));} else {testFailed(OT_T("markSupported"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("markSupported"));
	}
	try
	{
		rpBufIS->mark(100); testPassed(OT_T("mark"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("mark"));
	}
	try
	{
		if(rpBufIS->read()==1) {testPassed(OT_T("read"));} else {testFailed(OT_T("read"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("read"));
	}
	try
	{
		if(rpBufIS->read()==2) {testPassed(OT_T("read"));} else {testFailed(OT_T("read"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("read"));
	}
	//
	// Reset the position and then test the first two bytes again
	//
	try
	{
		rpBufIS->reset(); testPassed(OT_T("reset2"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("reset2"));
	}
	try
	{
		if(rpBufIS->read()==1) {testPassed(OT_T("read"));} else {testFailed(OT_T("read"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("read"));
	}
	try
	{
		if(rpBufIS->read()==2) {testPassed(OT_T("read"));} else {testFailed(OT_T("read"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("read"));
	}
	//
	// Mark again, this time from a different position, and with a buffer squeezing readLimit
	//
	try
	{
		rpBufIS->mark(200); testPassed(OT_T("mark"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("mark"));
	}
	try
	{
		if(rpBufIS->read()==3) {testPassed(OT_T("read"));} else {testFailed(OT_T("read"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("read"));
	}
	try
	{
		rpBufIS->reset(); testPassed(OT_T("reset3"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("reset3"));
	}
	try
	{
		if(rpBufIS->read()==3) {testPassed(OT_T("read"));} else {testFailed(OT_T("read"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("read"));
	}
	//
	// Mark again, this time from a different position, and with a buffer blowing readLimit
	//
	try
	{
		rpBufIS->mark(1000); testPassed(OT_T("mark"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("mark"));
	}
	try
	{
		if(rpBufIS->read()==4) {testPassed(OT_T("read"));} else {testFailed(OT_T("read"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("read"));
	}
	try
	{
		rpBufIS->reset(); testPassed(OT_T("reset4"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("reset4"));
	}
	int startChar=0;
	try
	{
		if((startChar = rpBufIS->read())==4) {testPassed(OT_T("read"));} else {testFailed(OT_T("read"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("read"));
	}
	try
	{
		rpBufIS->reset(); testPassed(OT_T("reset5"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("reset5"));
	}

	// 
	// By default, Readers do not support mark/reset
	// 
	RefPtr<InputStreamReader> rpRdr = new InputStreamReader(rpBufIS.get(), OT_T("utf-8"));
	try
	{
		if(rpRdr->markSupported()==false) {testPassed(OT_T("markSupported"));} else {testFailed(OT_T("markSupported"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("markSupported"));
	}
	try
	{
		rpRdr->mark(1);
		testFailed(OT_T("mark"));
	}
	catch(IOException& e)
	{
		goodCatch(OT_T("mark"), e.toString());
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("mark"));
	}
	try
	{
		rpRdr->reset();
		testFailed(OT_T("reset4"));
	}
	catch(IOException& e)
	{
		goodCatch(OT_T("reset4"), e.toString());
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("reset4"));
	}

	//
	// but BufferedReaders do support mark.reset
	// use a small buffer size so that reset() fails when mark exceeded
	//
	RefPtr<BufferedReader> rpBufRdr = new BufferedReader(rpRdr.get(), 100);
	try
	{
		if(rpBufRdr->markSupported()==true) {testPassed(OT_T("markSupported"));} else {testFailed(OT_T("markSupported"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("markSupported"));
	}
	try
	{
		rpBufRdr->mark(4); testPassed(OT_T("mark"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("mark"));
	}
	try
	{
		if(rpBufRdr->read()==startChar) {testPassed(OT_T("read"));} else {testFailed(OT_T("read"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("read"));
	}
	try
	{
		rpBufRdr->reset(); testPassed(OT_T("reset5"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("reset5"));
	}
	//
	// Bust the buffer size limit
	//
	try
	{
		rpBufRdr->mark(200); testPassed(OT_T("reset5"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("reset5"));
	}
	try
	{
		if(rpBufRdr->read()==startChar) {testPassed(OT_T("read"));} else {testFailed(OT_T("read"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("read"));
	}
	try
	{
		if(rpBufRdr->readAtomic().toUnicode()==(UCS4Char)startChar+1) {testPassed(OT_T("readAtomic"));} else {testFailed(OT_T("readAtomic"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("readAtomic"));
	}
	try
	{
		if(rpBufRdr->skipAtomic(100)==100) {testPassed(OT_T("skipAtomic"));} else {testFailed(OT_T("skipAtomic"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("skipAtomic"));
	}
	try
	{
		if(rpBufRdr->readAtomic().toUnicode()==(UCS4Char)startChar+102) {testPassed(OT_T("readAtomic"));} else {testFailed(OT_T("readAtomic"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("readAtomic"));
	}
	//
	// Reset the position and then test that the 128th character is okay
	//
	try
	{
		rpBufRdr->reset(); testPassed(OT_T("reset5"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("reset5"));
	}
	const size_t charsToSkip = (127-startChar+1);
	try
	{
		if(rpBufRdr->skip(charsToSkip)==charsToSkip) {testPassed(OT_T("skip"));} else {testFailed(OT_T("skip"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("skip"));
	}
	try
	{
		if(rpBufRdr->readAtomic().toUnicode()==128) {testPassed(OT_T("readAtomic"));} else {testFailed(OT_T("readAtomic"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("readAtomic"));
	}

#ifdef OT_UTF8
	// throw the reader into an invalid character position
	try
	{
		rpBufRdr->read(); testPassed(OT_T("read"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("read"));
	}
	// We are in an invalid position, so readAtomic should throw an exception
	try
	{
		rpBufRdr->readAtomic();
		testFailed(OT_T("readAtomic"));
	}
	catch(AtomicReadException& e)
	{
		goodCatch(OT_T("readAtomic"), e.toString());
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("readAtomic"));
	}
#endif //OT_UTF8

	// Ensure we exceed the readLimit
	try
	{
		if(rpBufRdr->skip(200) == 200) {testPassed(OT_T("skip"));} else {testFailed(OT_T("skip"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("skip"));
	}
	// We have exceeded the mark position, so expect an IOException
	try
	{
		rpBufRdr->reset();
		testFailed(OT_T("reset6"));
	}
	catch(IOException& e)
	{
		goodCatch(OT_T("reset6"), e.toString());
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("reset6"));
	}
	//
	// we may not be able to delete the file while it's still open, so
	// explicity close it
	//
	try
	{
		rpBufRdr->close(); testPassed(OT_T("close"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("close"));
	}
	testClosedOps(rpBufRdr.get());

	//
	// Delete the data files created earlier
	//
	try
	{
		utf8.deleteFile(); testPassed(OT_T("deleteFile"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("deleteFile"));
	}
	try
	{
		utf16.deleteFile(); testPassed(OT_T("deleteFile"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("deleteFile"));
	}
	try
	{
		iso88591.deleteFile(); testPassed(OT_T("deleteFile"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("deleteFile"));
	}

	//
	// Test atomic operations
	//
	atomicFileTests(OT_T("utf-8"));
	atomicFileTests(OT_T("utf-16"));

	RefPtr<StringWriter> rpStrWrt = new StringWriter;
	writeCharRange(rpStrWrt.get(), 0x10000, 0x10100);

	String s = rpStrWrt->toString();
	RefPtr<Reader> rpReader = new StringReader(s);
	try
	{
		if(testReaderContents(rpReader.get(), 0x10000, 0x10100)) {testPassed(OT_T("contents"));} else {testFailed(OT_T("contents"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("contents"));
	}
	rpReader = new StringReader(s);
	atomicReaderTests(rpReader.get());
	try
	{
		rpReader->close(); testPassed(OT_T("close"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("close"));
	}
	testClosedOps(rpReader.get());

	//
	// StringReaders do support mark/reset without the need for a BufferedReader
	//
	String hello = OT_T("Hello World!");
	RefPtr<Reader> rpStrRdr = new StringReader(hello);
	markMyWorld(rpStrRdr.get());
	try
	{
		rpStrRdr->close(); testPassed(OT_T("close"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("close"));
	}
	testClosedOps(rpStrRdr.get());

	//
	// FilterReaders should support the same ops as the Reader they are imbued with
	//
	RefPtr<Reader> rpFilterRdr = new FilterReader(new StringReader(hello));
	markMyWorld(rpFilterRdr.get());
	try
	{
		rpFilterRdr->close(); testPassed(OT_T("close"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("close"));
	}
	testClosedOps(rpFilterRdr.get());

	//
	// create a file containing Hello World
	//
	File fHello(OT_T("hello.txt"));
	RefPtr<FileOutputStream> rpOS = new FileOutputStream(fHello);
	RefPtr<OutputStreamWriter> rpWriter = new OutputStreamWriter(rpOS.get(), OT_T("UTF-16"));
	try
	{
		rpWriter->write(hello); testPassed(OT_T("write"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("write"));
	};
	try
	{
		rpWriter->close(); testPassed(OT_T("close"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("close"));
	};
	testClosedOps(rpWriter.get());

	RefPtr<FileInputStream> rpISHello = new FileInputStream(fHello);
	RefPtr<BufferedReader> rpHelloRdr = new BufferedReader(new InputStreamReader(rpISHello.get(), OT_T("UTF-16")));
	markMyWorld(rpHelloRdr.get());
	try
	{
		rpHelloRdr->close(); testPassed(OT_T("close"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("close"));
	}
	testClosedOps(rpHelloRdr.get());
	try
	{
		fHello.deleteFile(); testPassed(OT_T("deleteFile"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("deleteFile"));
	}
	

	testMessage(OT_T("End of tests for Stream"));
}

