//
// ElCel Technology Test Harness
//
// Class: Stream
//

include(../common.m4)
OT_TEST_INIT

#`include' "ot/base/Character.h"
#`include' "ot/io/AtomicReadException.h"
#`include' "ot/io/BufferedInputStream.h"
#`include' "ot/io/BufferedOutputStream.h"
#`include' "ot/io/BufferedReader.h"
#`include' "ot/io/BufferedWriter.h"
#`include' "ot/io/Reader.h"
#`include' "ot/io/Writer.h"
#`include' "ot/io/File.h"
#`include' "ot/io/IOException.h"
#`include' "ot/io/FileOutputStream.h"
#`include' "ot/io/FileInputStream.h"
#`include' "ot/io/OutputStreamWriter.h"
#`include' "ot/io/InputStreamReader.h"
#`include' "ot/io/StringReader.h"
#`include' "ot/io/StringWriter.h"
#`include' "ot/io/FilterReader.h"
#`include' "ot/base/NumUtils.h"
#`include' "ot/base/IllegalCharacterException.h"
#`include' "ot/io/CharacterCodingException.h"
#`include' "ot/io/MalformedInputException.h"

using namespace ot::io;

//
// write a range of characters to a Writer
//
void writeCharRange(Writer* pWriter, unsigned long firstChar, unsigned long lastChar)
{
	RefPtr<BufferedWriter> rpWriter = new BufferedWriter(pWriter);
	for (unsigned long i=firstChar; i<=lastChar; i++)
	{
		if(i<0xD800UL || i>0xDFFFUL)
		{
			Character ch(i);
			rpWriter->write(ch.data(), ch.length());
		}
	}
}

//
// Helper function to generate a file with a name, encoding and filled
// with characters from firstChar  - lastChar
//
void genFile(const File& file, const String& enc, unsigned long firstChar, unsigned long lastChar)
{
	RefPtr<OutputStream> rpOS = new FileOutputStream(file);
	rpOS = new BufferedOutputStream(rpOS.get());
	RefPtr<Writer> rpWriter = new OutputStreamWriter(rpOS.get(), enc, true);
	rpWriter = new BufferedWriter(rpWriter.get());
	OT_TEST_VOID(write range, writeCharRange(rpWriter.get(), firstChar, lastChar))
}

//
// Helper function to check that all methods return the correct value
// when  a character stream is at EOF
//
void testEOF(Reader* pReader)
{
	// Ensure all read methods agree we have reached the end of file
	CharType buffer[6];
	OT_TEST_BOOL(read, pReader->readAtomic(buffer, 6)==Reader::EndOfFile)
	OT_TEST_BOOL(read, pReader->read(buffer, 6)==Reader::EndOfFile)
	OT_TEST_BOOL(read, pReader->read()==Reader::EndOfFile)
	OT_TEST_BOOL(read, pReader->skip(1)==0)
	OT_TEST_BOOL(read, pReader->skipAtomic(1)==0)
}

//
// Helper function to check the contents returned from a Reader
//
bool testReaderContents(Reader* pReader, unsigned long firstChar, unsigned long lastChar)
{
	bool bRet = true;
	unsigned long i = firstChar;
	while(true)
	{
		Character ch = pReader->readAtomic();
		if(ch == Character::EndOfFileCharacter)
		{
			testEOF(pReader);
			break;
		}
		if(ch.toUnicode() != i || i > lastChar)
		{
			String msg = OT_T("mismatch, got=");
			msg += NumUtils::ToString(ch.toUnicode());
			msg += OT_T(" expected=");
			msg += NumUtils::ToString(i);
			OT_TEST_MESSAGE(msg)
			bRet = false;
			break;
		}
		i++;
		// skip the surrogate pair range
		if(i==0xD800) i=0xE000;
	}
	bRet = bRet && (i==lastChar+1);
	if(!bRet)
	{
		String msg = OT_T("testReaderContents failed, i=");
		msg += NumUtils::ToString(i);
		OT_TEST_MESSAGE(msg)
	}
	return bRet;
}

//
// Helper function to read a file with a name, encoding and filled
// with characters from firstChar - lastChar
//
void testFileContents(const File& file, const String& enc, unsigned long firstChar, unsigned long lastChar)
{
	RefPtr<InputStream> rpIS = new FileInputStream(file);
	RefPtr<Reader> rpReader = new InputStreamReader(rpIS.get(), enc, true);
	dnl OT_TEST_BOOL(testReaderContents, testReaderContents(rpReader.get(), firstChar, lastChar))
	// test again using buffering
	rpIS = new FileInputStream(file);
	rpReader = new BufferedReader(new InputStreamReader(rpIS.get(), enc, true));
	OT_TEST_BOOL(testReaderContents, testReaderContents(rpReader.get(), firstChar, lastChar))
}

//
// test that opererations are correctly disallowed on closed readers
//
void testClosedOps(Reader* pReader)
{
	// It should be okay to close a closed stream
	OT_TEST_VOID(closed1, pReader->close())
	OT_TEST_VOID(closed2, pReader->close())
	// Once closed, the following methods are not allowed: read, mark, reset
	CharType buffer[10];
	size_t bufLen=10;
	OT_TEST_THROW(closed3, pReader->readAtomic(), IOException)
	OT_TEST_THROW(closed4, pReader->readAtomic(buffer, bufLen), IOException)
	OT_TEST_THROW(closed5, pReader->read(), IOException)
	OT_TEST_THROW(closed6, pReader->read(buffer, bufLen), IOException)
	OT_TEST_THROW(closed7, pReader->mark(4), IOException)
	OT_TEST_THROW(closed8, pReader->reset(), IOException)
	OT_TEST_THROW(closed9, pReader->skip(1), IOException)
	OT_TEST_THROW(closedA, pReader->skipAtomic(1), IOException)
}

//
// Perform a bunch of atomic operations on a Reader that
// is positioned on a high unicode character
//
void atomicReaderTests(Reader* pReader)
{
#`ifndef' OT_UCS4
	CharType buffer[6];
	// test rc=0 for readAtomic with too small buffer
	OT_TEST_BOOL(readAtomic, pReader->readAtomic(buffer, 1)==0)
	OT_TEST_BOOL(skipAtomic, pReader->skipAtomic(1)==1)
	OT_TEST_BOOL(readAtomic2, pReader->readAtomic().toUnicode()==0x10001UL)
	// skip to an invalid multi-char seq boundary
	OT_TEST_BOOL(skip, pReader->skip(1)==1)
	dnl OT_TEST_THROW(skipAtomic, pReader->skipAtomic(1), AtomicReadException)
	OT_TEST_THROW(skipAtomic, pReader->skipAtomic(1), IOException)
#`endif' //OT_UCS4
}

//
// Perform a bunch of atomic operations on a file with the given encoding
//
void atomicFileTests(const String& enc)
{
	File atomic(OT_T("atomic_file.txt"));
	genFile(atomic, enc, 0x10000, 0x10FFFF);
	OT_TEST_BOOL(atomic isFile, atomic.isFile())

	RefPtr<InputStream> rpIS = new FileInputStream(atomic);
	RefPtr<Reader> rpReader = new InputStreamReader(rpIS.get(), enc, true);
	atomicReaderTests(rpReader.get());
	OT_TEST_VOID(close, rpReader->close())
	testClosedOps(rpReader.get());

	rpIS = new FileInputStream(atomic);
	rpReader = new BufferedReader(new InputStreamReader(rpIS.get(), enc, true));
	atomicReaderTests(rpReader.get());
	OT_TEST_VOID(close, rpReader->close())
	testClosedOps(rpReader.get());

	OT_TEST_VOID(delete atomic, atomic.deleteFile())
	OT_TEST_BOOL(atomic isFile, atomic.isFile()==false)
	OT_TEST_BOOL(atomic exists, atomic.exists()==false)
}

//
// test a variety of Readers which have a controlled
// stream of "Hello World!"
//
void markMyWorld(Reader* pReader)
{
	OT_TEST_BOOL(markSupported, pReader->markSupported()==true)
	OT_TEST_VOID(mark, pReader->mark(4))
	OT_TEST_BOOL(read, pReader->read()=='H')
	OT_TEST_BOOL(read, pReader->read()=='e')
	OT_TEST_BOOL(readAtomic, pReader->readAtomic().toUnicode()=='l')
	OT_TEST_VOID(reset7, pReader->reset())
	OT_TEST_BOOL(read, pReader->read()=='H')
	OT_TEST_BOOL(skip, pReader->skip(11)==11)
	testEOF(pReader);
}

//
// test that opererations are correctly disallowed on closed writers
//
void testClosedOps(Writer* pWriter)
{
	// It should be okay to close a closed stream
	OT_TEST_VOID(close, pWriter->close())
	OT_TEST_VOID(close, pWriter->close())
	CharType buffer[10];
	size_t bufLen=10;
	OT_TEST_THROW(write, pWriter->write(buffer, bufLen), IOException)
	OT_TEST_THROW(write, pWriter->write(OT_T("hello")), IOException)
	OT_TEST_THROW(write, pWriter->write('H'), IOException)
	OT_TEST_THROW(flush, pWriter->flush(), IOException)
}

OT_START_TESTS(`Stream')

	//
	// create three files using the common unicode encodings
	// utf-8, utf-16 and iso-8859-1
	//
	File utf8(OT_T("test_utf8"));
	File utf16(OT_T("test_utf16"));
	File iso88591(OT_T("test_iso8859_1"));

	genFile(utf8, OT_T("utf-8"), 1, 0x10FFFFUL);
	testFileContents(utf8, OT_T("utf-8"), 1, 0x10FFFFUL);

	genFile(utf16, OT_T("utf-16"), 1, 0x10FFFFUL);
	testFileContents(utf16, OT_T("utf-16"), 1, 0x10FFFFUL);

	genFile(iso88591, OT_T("iso-8859-1"), 1, 0x00FEUL);
	testFileContents(iso88591, OT_T("iso-8859-1"), 1, 0x00FEUL);

	//
	// test ability to write illegal surrogate value
	// with strict turned on and then off
	//
	File badutf16(OT_T("badutf16.txt"));
	RefPtr<OutputStreamWriter> badWriter = new OutputStreamWriter(new FileOutputStream(badutf16), OT_T("utf-16"), true);
#`ifdef' OT_UTF16
	OT_TEST_THROW(illegal surrogate, Character bad(0xD800), IllegalCharacterException)
#`else'
	Character bad(0xD800);
	OT_TEST_THROW(surrogate, badWriter->write(bad.data(), bad.length()), CharacterCodingException)
	OT_TEST_VOID(surrogate2, badWriter->getEncoder()->setUnmappableCharAction(CodeConverter::replace))
	OT_TEST_VOID(surrogate3 badWriter->write(bad.data(), bad.length()))
#`endif'
	OT_TEST_VOID(close, badWriter->close())
	OT_TEST_VOID(deleteFile, badutf16.deleteFile())

	//
	// test ability to handle illegal utf-8 sequences
	//
	File badutf8(OT_T("badutf8.txt"));
	RefPtr<OutputStream> rpBadUTF8OStream  = new FileOutputStream(badutf8);
	Byte bad_text_1[] = {0x80}; // illegal start char
	Byte bad_text_2[] = {0xf0, 0x80, 0x80, 0x80}; // illegal null char
	Byte bad_text_3[] = {0xf0, 0x81, 0x81}; // premature eof

	OT_TEST_VOID(write 1, rpBadUTF8OStream->write(bad_text_1, sizeof(bad_text_1)))
	OT_TEST_VOID(write 2, rpBadUTF8OStream->write(bad_text_2, sizeof(bad_text_2)))
	OT_TEST_VOID(write 3, rpBadUTF8OStream->write(bad_text_3, sizeof(bad_text_3)))
	OT_TEST_VOID(close, rpBadUTF8OStream->close())
	//
	// Now, let's see what happens when we read it!
	// We need strict checking switched on to start with
	//
	RefPtr<InputStreamReader> rpBadUtf8Reader = new InputStreamReader(new FileInputStream(badutf8), OT_T("utf-8"), true);
	
	// If the exception is correctly thrown, then the Reader's position should not have changed
	// which means there are three tokens left to read
	OT_TEST_THROW(illegal start, rpBadUtf8Reader->readAtomic(), IOException)
	OT_TEST_VOID(change converter, rpBadUtf8Reader->getDecoder()->setInvalidCharAction(CodeConverter::replace))
	OT_TEST_BOOL(test repl1, rpBadUtf8Reader->readAtomic().toUnicode() == 0xFFFD)
	OT_TEST_BOOL(test repl2, rpBadUtf8Reader->readAtomic().toUnicode() == 0xFFFD)
	OT_TEST_BOOL(test repl3, rpBadUtf8Reader->readAtomic().toUnicode() == 0xFFFD)
	OT_TEST_BOOL(test eof, rpBadUtf8Reader->readAtomic() == Character::EndOfFileCharacter)

	OT_TEST_VOID(close, rpBadUtf8Reader->close())
	testClosedOps(rpBadUtf8Reader.get());
	OT_TEST_VOID(deleteFile, badutf8.deleteFile())

	//
	// Test mark/reset operations
	//

	RefPtr<FileInputStream> rpIS1 = new FileInputStream(utf8);
	OT_TEST_BOOL(markSupported, rpIS1->markSupported()==false)
	OT_TEST_THROW(mark, rpIS1->mark(1), IOException)
	OT_TEST_THROW(reset, rpIS1->reset(), IOException)

	RefPtr<BufferedInputStream> rpBufIS = new BufferedInputStream(rpIS1.get(), 200);
	OT_TEST_BOOL(markSupported, rpBufIS->markSupported()==true)
	OT_TEST_VOID(mark, rpBufIS->mark(100))
	OT_TEST_BOOL(read, rpBufIS->read()==1)
	OT_TEST_BOOL(read, rpBufIS->read()==2)
	//
	// Reset the position and then test the first two bytes again
	//
	OT_TEST_VOID(reset2, rpBufIS->reset())
	OT_TEST_BOOL(read, rpBufIS->read()==1)
	OT_TEST_BOOL(read, rpBufIS->read()==2)
	//
	// Mark again, this time from a different position, and with a buffer squeezing readLimit
	//
	OT_TEST_VOID(mark, rpBufIS->mark(200))
	OT_TEST_BOOL(read, rpBufIS->read()==3)
	OT_TEST_VOID(reset3, rpBufIS->reset())
	OT_TEST_BOOL(read, rpBufIS->read()==3)
	//
	// Mark again, this time from a different position, and with a buffer blowing readLimit
	//
	OT_TEST_VOID(mark, rpBufIS->mark(1000))
	OT_TEST_BOOL(read, rpBufIS->read()==4)
	OT_TEST_VOID(reset4, rpBufIS->reset())
	int startChar=0;
	OT_TEST_BOOL(read, (startChar = rpBufIS->read())==4)
	OT_TEST_VOID(reset5, rpBufIS->reset())

	// 
	// By default, Readers do not support mark/reset
	// 
	RefPtr<InputStreamReader> rpRdr = new InputStreamReader(rpBufIS.get(), OT_T("utf-8"));
	OT_TEST_BOOL(markSupported, rpRdr->markSupported()==false)
	OT_TEST_THROW(mark, rpRdr->mark(1), IOException)
	OT_TEST_THROW(reset4, rpRdr->reset(), IOException)

	//
	// but BufferedReaders do support mark.reset
	// use a small buffer size so that reset() fails when mark exceeded
	//
	RefPtr<BufferedReader> rpBufRdr = new BufferedReader(rpRdr.get(), 100);
	OT_TEST_BOOL(markSupported, rpBufRdr->markSupported()==true)
	OT_TEST_VOID(mark, rpBufRdr->mark(4))
	OT_TEST_BOOL(read, rpBufRdr->read()==startChar)
	OT_TEST_VOID(reset5, rpBufRdr->reset())
	//
	// Bust the buffer size limit
	//
	OT_TEST_VOID(reset5, rpBufRdr->mark(200))
	OT_TEST_BOOL(read, rpBufRdr->read()==startChar)
	OT_TEST_BOOL(readAtomic, rpBufRdr->readAtomic().toUnicode()==(UCS4Char)startChar+1)
	OT_TEST_BOOL(skipAtomic, rpBufRdr->skipAtomic(100)==100)
	OT_TEST_BOOL(readAtomic, rpBufRdr->readAtomic().toUnicode()==(UCS4Char)startChar+102)
	//
	// Reset the position and then test that the 128th character is okay
	//
	OT_TEST_VOID(reset5, rpBufRdr->reset())
	const size_t charsToSkip = (127-startChar+1);
	OT_TEST_BOOL(skip, rpBufRdr->skip(charsToSkip)==charsToSkip)
	OT_TEST_BOOL(readAtomic, rpBufRdr->readAtomic().toUnicode()==128)

#`ifdef' OT_UTF8
	// throw the reader into an invalid character position
	OT_TEST_VOID(read, rpBufRdr->read())
	// We are in an invalid position, so readAtomic should throw an exception
	OT_TEST_THROW(readAtomic, rpBufRdr->readAtomic(), AtomicReadException)
#`endif' //OT_UTF8

	// Ensure we exceed the readLimit
	OT_TEST_BOOL(skip, rpBufRdr->skip(200) == 200)
	// We have exceeded the mark position, so expect an IOException
	OT_TEST_THROW(reset6, rpBufRdr->reset(), IOException)
	//
	// we may not be able to delete the file while it's still open, so
	// explicity close it
	//
	OT_TEST_VOID(close, rpBufRdr->close())
	testClosedOps(rpBufRdr.get());

	//
	// Delete the data files created earlier
	//
	OT_TEST_VOID(deleteFile, utf8.deleteFile())
	OT_TEST_VOID(deleteFile, utf16.deleteFile())
	OT_TEST_VOID(deleteFile, iso88591.deleteFile())

	//
	// Test atomic operations
	//
	atomicFileTests(OT_T("utf-8"));
	atomicFileTests(OT_T("utf-16"));

	RefPtr<StringWriter> rpStrWrt = new StringWriter;
	writeCharRange(rpStrWrt.get(), 0x10000, 0x10100);

	String s = rpStrWrt->toString();
	RefPtr<Reader> rpReader = new StringReader(s);
	OT_TEST_BOOL(contents, testReaderContents(rpReader.get(), 0x10000, 0x10100))
	rpReader = new StringReader(s);
	atomicReaderTests(rpReader.get());
	OT_TEST_VOID(close, rpReader->close())
	testClosedOps(rpReader.get());

	//
	// StringReaders do support mark/reset without the need for a BufferedReader
	//
	String hello = OT_T("Hello World!");
	RefPtr<Reader> rpStrRdr = new StringReader(hello);
	markMyWorld(rpStrRdr.get());
	OT_TEST_VOID(close, rpStrRdr->close())
	testClosedOps(rpStrRdr.get());

	//
	// FilterReaders should support the same ops as the Reader they are imbued with
	//
	RefPtr<Reader> rpFilterRdr = new FilterReader(new StringReader(hello));
	markMyWorld(rpFilterRdr.get());
	OT_TEST_VOID(close, rpFilterRdr->close())
	testClosedOps(rpFilterRdr.get());

	//
	// create a file containing Hello World
	//
	File fHello(OT_T("hello.txt"));
	RefPtr<FileOutputStream> rpOS = new FileOutputStream(fHello);
	RefPtr<OutputStreamWriter> rpWriter = new OutputStreamWriter(rpOS.get(), OT_T("UTF-16"));
	OT_TEST_VOID(write, rpWriter->write(hello));
	OT_TEST_VOID(close, rpWriter->close());
	testClosedOps(rpWriter.get());

	RefPtr<FileInputStream> rpISHello = new FileInputStream(fHello);
	RefPtr<BufferedReader> rpHelloRdr = new BufferedReader(new InputStreamReader(rpISHello.get(), OT_T("UTF-16")));
	markMyWorld(rpHelloRdr.get());
	OT_TEST_VOID(close, rpHelloRdr->close())
	testClosedOps(rpHelloRdr.get());
	OT_TEST_VOID(deleteFile, fHello.deleteFile())
	
OT_END_TESTS(`Stream')

