//
// ElCel Technology Test Harness 
//
// Class: XMLReader
//




#include "ot/base/System.h"
#include "ot/io/Console.h"
#include "ot/base/Exception.h"
#include "ot/base/StringUtils.h"
#include "ot/util/AttributeListParser.h"

using namespace ot;
using namespace ot::io;
using namespace ot::util;

String getTestAttribute(const String& name);
void testMessage(const String& msg);
void testFailed(const String& test);
void testPassed(const String& test);
void goodCatch(const String& test, const String& eMsg);
void uncaughtException(const String& e, const String& test);


#include "ot/sax/XMLReaderFactory.h"
#include "ot/sax/XMLReader.h"
#include "ot/sax/Locator.h"
#include "ot/sax/SAXFeatures.h"
#include "ot/sax/DefaultHandler.h"
#include "ot/sax/SAXNotSupportedException.h"
#include "ot/sax/SAXParseException.h"
#include "ot/net/UnknownHostException.h"
#include "ot/io/StringReader.h"

using namespace ot::sax; 
using namespace ot::net; 
using namespace ot::io; 

size_t NamespaceAttributes = 0;

class MyXMLReaderHandler : public DefaultHandler
{
public:

	virtual void setDocumentLocator(Locator* pLocator)
	{
		m_rpLocator = pLocator;
	}
	virtual void startDocument()
	{
	}
	virtual void startElement(const String& ns, const String& local, const String& qname, const Attributes& atts)
	{
		try
	{
		if(ns == OT_T("http://www.opentop.org/ns")) {testPassed(OT_T("check ns uri"));} else {testFailed(OT_T("check ns uri"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("check ns uri"));
	}
		try
	{
		if(local == OT_T("test")) {testPassed(OT_T("check local"));} else {testFailed(OT_T("check local"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("check local"));
	}
		try
	{
		if(qname == OT_T("ot:test")) {testPassed(OT_T("check qname"));} else {testFailed(OT_T("check qname"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("check qname"));
	}
		try
	{
		if(atts.getLength() == (1 + NamespaceAttributes)) {testPassed(OT_T("check atts"));} else {testFailed(OT_T("check atts"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("check atts"));
	}
		try
	{
		if(atts.getType(0) == OT_T("CDATA")) {testPassed(OT_T("check att type"));} else {testFailed(OT_T("check att type"));}
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("check att type"));
	}
	}
	virtual void error(const SAXParseException& e)
	{
		throw e;
	}	

	RefPtr<Locator> m_rpLocator;
	bool m_bIsStandalone;
};


void XMLReader_Tests()
{
	testMessage(OT_T("Starting tests for XMLReader"));

	const String dtd   = OT_T("<!DOCTYPE test [")
	                     OT_T("<!ELEMENT test (#PCDATA)*>")
	                     OT_T("<!ATTLIST test attr1 CDATA #REQUIRED attr2 (yes|no) 'yes'>")
	                     OT_T("]>");
	const String doc   = OT_T("<ot:test xmlns:ot='http://www.opentop.org/ns' attr1='attr1'/>");
	const String valid = dtd + doc;

	RefPtr<XMLReader> rpReader = XMLReaderFactory::CreateXMLReader();
	RefPtr<MyXMLReaderHandler> rpHandler = new MyXMLReaderHandler();

	//
	// Set the XMLreader into a validating parser
	//
	try
	{
		rpReader->setFeature(SAXFeatures::validation, true); testPassed(OT_T("set validation"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("set validation"));
	}

	//
	// Register a content handler, which tests correct namespace processing
	//
	rpReader->setContentHandler(rpHandler.get());
	{
		RefPtr<StringReader> rpSR = new StringReader(dtd);
		RefPtr<InputSource> rpIS = new InputSource(rpSR.get());
		try
	{
		rpReader->parse(rpIS.get());
		testFailed(OT_T("parse notwf"));
	}
	catch(SAXParseException& e)
	{
		goodCatch(OT_T("parse notwf"), e.toString());
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("parse notwf"));
	}
	}
	{
		RefPtr<StringReader> rpSR = new StringReader(doc);
		RefPtr<InputSource> rpIS = new InputSource(rpSR.get());
		try
	{
		rpReader->parse(rpIS.get()); testPassed(OT_T("parse invalid"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("parse invalid"));
	}
	}
	{
		RefPtr<StringReader> rpSR = new StringReader(valid);
		RefPtr<InputSource> rpIS = new InputSource(rpSR.get());
		try
	{
		rpReader->parse(rpIS.get()); testPassed(OT_T("parse valid"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("parse valid"));
	}
	}

	//
	// Now let's try that again, this time with namespace prefix reporting switched on
	//
	try
	{
		rpReader->setFeature(SAXFeatures::namespace_prefixes, true); testPassed(OT_T("switch on prefix reporting"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("switch on prefix reporting"));
	}
	NamespaceAttributes = 1;
	{
		RefPtr<StringReader> rpSR = new StringReader(valid);
		RefPtr<InputSource> rpIS = new InputSource(rpSR.get());
		try
	{
		rpReader->parse(rpIS.get()); testPassed(OT_T("parse ns-prefix"));
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("parse ns-prefix"));
	}
	}

	//
	// Register an error handler, which throws a SAXParseException on validity errors
	//
	rpReader->setErrorHandler(rpHandler.get());
	{
		RefPtr<StringReader> rpSR = new StringReader(doc);
		RefPtr<InputSource> rpIS = new InputSource(rpSR.get());
		try
	{
		rpReader->parse(rpIS.get());
		testFailed(OT_T("parse invalid 2"));
	}
	catch(SAXParseException& e)
	{
		goodCatch(OT_T("parse invalid 2"), e.toString());
	}
	catch(Exception& e)
	{
		uncaughtException(e.toString(), OT_T("parse invalid 2"));
	}
	}


	testMessage(OT_T("End of tests for XMLReader"));
}

