//==============================================================================
// This file is part of the OpenTop C++ Library.
// (c) Copyright 2000-2002 ElCel Technology Limited
//
// This file may be distributed and/or modified under the terms of the
// GNU General Public License version 2 as published by the Free Software
// Foundation and appearing in the file LICENSE.GPL included in the
// packaging of this file.
//
// Licensees holding valid OpenTop Commercial licences may use this file
// in accordance with the OpenTop Commercial License Agreement included
// with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// See http://www.elcel.com/products/opentop/pricing.html or email
// sales@elcel.com for information about OpenTop Commercial License Agreements.
//
//==============================================================================
//
// $Revision: 1.2 $
// $Date: 2003/08/28 19:47:48 $
//
//==============================================================================

#include "FileMessageFactory.h"

#include "ot/base/System.h"
#include "ot/base/NumUtils.h"
#include "ot/base/SystemUtils.h"
#include "ot/io/BufferedReader.h"
#include "ot/io/File.h"
#include "ot/io/InputStreamReader.h"
#include "ot/io/FileInputStream.h"

OT_AUXIL_NAMESPACE_BEGIN

using namespace io;
using namespace util;

//==============================================================================
// FileMessageFactory::Install
//
// Static function to install a FileMessageFactory if the appropriate
// environment is present.
//==============================================================================
void FileMessageFactory::Install()
{
	//
	// If the environment variable: ET_MSG_DIR is set then hey! let's
	// use internationalized messages
	//
	const String& sMsgDir = System::GetEnvironmentString(OT_T("ET_MSG_DIR"));
	if(!sMsgDir.empty())
	{
		const String& sLang = System::GetEnvironmentString(OT_T("LANG"));
		System::SetMessageFactory(new FileMessageFactory(sMsgDir, sLang));
	}
}

//==============================================================================
// FileMessageFactory::FileMessageFactory
//
//==============================================================================
FileMessageFactory::FileMessageFactory(const String& msgDirectory, const String lang) :
	m_lang(lang),
	m_bDirectoryOK(false)
{
	//
	// A common user error will be to put a trailing separator char on the end of
	// the directory, so we'll remove it now.
	//
	if(!msgDirectory.empty())
	{
		size_t pos = msgDirectory.find_last_not_of(OT_T("/\\"));
		if(pos != String::npos)
		{
			m_msgDir = msgDirectory.substr(0, pos+1);
		}
	}

	m_bDirectoryOK = File(m_msgDir).isDirectory();
}

//==============================================================================
// FileMessageFactory::getMessageSet
//
//==============================================================================
RefPtr<MessageSet> FileMessageFactory::getMessageSet(const String& org, const String& app)
{
	RefPtr<MessageSet> rpSet = new MessageSet(org, app);

	if(m_bDirectoryOK)
	{
		//
		// construct the name of the message file that we will read to create
		// the requested message set
		//
		// filename ::= msg-dir '/' [org-name '/'] [lang'/'] app-nane '.msg'
		//
		const String sDelim = OT_T(" \t");

		String filename = m_msgDir;
		filename += File::GetSeparatorChar();

		if(!org.empty())
		{
			filename += org;
			filename += File::GetSeparatorChar();
		}

		if(!m_lang.empty())
		{
			filename += m_lang;
			filename += File::GetSeparatorChar();
		}

		filename += app;
		filename += OT_T(".msg");


		//
		// Now attempt to open the file.  All failures are silently ignored, with the
		// result that messages are presented to the user using the English prototype.
		//
		try
		{
			//
			// Make use of the ElCel I/O facilities.  The input file must be coded
			// in UTF-8, so we use a UTF-8 Reader
			//
			const String sUTF8 = OT_T("UTF-8");

			RefPtr<BufferedReader> rpReader = new BufferedReader(
			                                  new InputStreamReader(
											  new FileInputStream(filename), sUTF8));

			//
			// Read the message file, one line at a time.
			//
			// Note that multi-line messages are permitted.  To cater for this
			// we do not add the message to the MessageSet until we have read all 
			// messages for the given number.
			//

			String line;
			String message;
			size_t messageID=0;

			while(rpReader->readLine(line) != Reader::EndOfFile)
			{
				if(line[0] == '#')
					continue;

				//
				// It was the intention to allow the input file to be relatively free-format
				// with a number followed by a string with any number of separator characters
				// in between.  The problem with this approach is that some messages need
				// leading spaces (such as the verbose usage display when the --help option
				// is given to a command-line program).
				//
				// For this reason we only permit one separator character between the
				// number and the start of the message.  The message can start at any
				// position, but it must be the first non-blank token on a line.
				//
				size_t numPos = line.find_first_not_of(sDelim);
				if(numPos != String::npos)
				{
					size_t msgPos = line.find_first_of(sDelim, numPos+1);

					size_t newMessageID;
					String newMessage;

					if(msgPos == String::npos)
					{
						newMessageID = NumUtils::ToLong(line.substr(numPos));
					}
					else
					{
						newMessageID = NumUtils::ToLong(line.substr(numPos, msgPos-numPos));
						newMessage = line.substr(msgPos+1);
					}

					//
					// okay, we have the message number (which cannot be zero)
					//
					if(newMessageID)
					{
						//
						// If it the same message ID as last time round the loop, then
						// append the message to the saved message...
						//
						if(newMessageID == messageID)
						{
							message += OT_T("\n");
							message += newMessage;
						}
						else 
						{
							//
							// ...otherwise add the saved message to the MessageSet
							//
							if(!message.empty())
							{
								rpSet->addMessage(messageID, message);
							}
							//
							// and save the current message for the next time round the loop
							//
							message = newMessage;
							messageID = newMessageID;
						}
					}
				}
			}
			// add the last one if present
			if(!message.empty())
			{
				rpSet->addMessage(messageID, message);
			}
		}
		catch(Exception& /*e*/)
		{
		}
	}
	
	return rpSet;
}

OT_AUXIL_NAMESPACE_END
