//==============================================================================
// This file is part of the OpenTop C++ Library.
// (c) Copyright 2000-2002 ElCel Technology Limited
//
// This file may be distributed and/or modified under the terms of the
// GNU General Public License version 2 as published by the Free Software
// Foundation and appearing in the file LICENSE.GPL included in the
// packaging of this file.
//
// Licensees holding valid OpenTop Commercial licences may use this file
// in accordance with the OpenTop Commercial License Agreement included
// with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// See http://www.elcel.com/products/opentop/pricing.html or email
// sales@elcel.com for information about OpenTop Commercial License Agreements.
//
//==============================================================================
//
// $Revision: 1.3 $
// $Date: 2003/05/04 17:47:30 $
//
//==============================================================================
// Class: MemCheckSystemMonitor
// 
/**
	@class ot::auxil::MemCheckSystemMonitor
	
	@brief Auxiliary helper class to terminate the @OpenTop library cleanly and
	       employ the memory-leak detection facilities of the C runtime library
	       (if any such facilities exist on a given platform)
	       to aid in application memory leak detection.

	Normally instantiated in the @c main() function, this class uses its constructor
	to initialize memory leak detection in the C runtime library.

	This class derives from SystemMonitor, so its destructor will 
	facilitate a clean shutdown of the application by calling System::Terminate().

    @code
    #include "ot/auxil/MemCheckSystemMonitor.h"
    using namespace ot;
    using namespace ot::auxil;

    int main(int argc, char* argv[])
    {
        // create a MemCheckSystemMonitor instance to ensure clean termination
        // and detect memory leaks on some platforms
        MemCheckSystemMonitor monitor;
        
        // create a deliberate memory leak...
        int* pInt = new int;

        return (0);
    }
    @endcode

	@note The only compiler currently supplying memory-leak detection within the
	      C-runtime library is Visual C++ under Microsoft Windows.  When running a 
	      program under the Visual Studio debugger, messages will be produced during
	      program termination to indicate the existence of any detected memory leaks.
	      The @c MemCheckSystemMonitor class does not currently perform 
	      memory-leak detection on any other platforms/compilers.
	@sa SystemMonitor
*/
//==============================================================================

#include "MemCheckSystemMonitor.h"

#ifdef _MSC_VER
#include <crtdbg.h>
#endif

OT_AUXIL_NAMESPACE_BEGIN

//==============================================================================
// MemCheckSystemMonitor::~MemCheckSystemMonitor
//
/**
   Constructor.

   Under some platforms (e.g. Visual C++ under Windows), the constructor
   initializes the C-runtime library to perform memory-leak detection at 
   application termination.
*/
//==============================================================================
MemCheckSystemMonitor::MemCheckSystemMonitor()
{

#ifdef _MSC_VER
	int tmpDbgFlag = _CrtSetDbgFlag(_CRTDBG_REPORT_FLAG);
	tmpDbgFlag |= _CRTDBG_LEAK_CHECK_DF;
	_CrtSetDbgFlag(tmpDbgFlag);
	// use _CrtSetBreakAlloc(nAlloc) to trap a memory allocation;
#endif 

}

OT_AUXIL_NAMESPACE_END
