//==============================================================================
// This file is part of the OpenTop C++ Library.
// (c) Copyright 2000-2002 ElCel Technology Limited
//
// This file may be distributed and/or modified under the terms of the
// GNU General Public License version 2 as published by the Free Software
// Foundation and appearing in the file LICENSE.GPL included in the
// packaging of this file.
//
// Licensees holding valid OpenTop Commercial licences may use this file
// in accordance with the OpenTop Commercial License Agreement included
// with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// See http://www.elcel.com/products/opentop/pricing.html or email
// sales@elcel.com for information about OpenTop Commercial License Agreements.
//
//==============================================================================
//
// $Revision: 1.3 $
// $Date: 2003/09/04 21:13:50 $
//
//==============================================================================

#include "TraceHelper.h"

#include "ot/base/System.h"
#include "ot/base/Thread.h"
#include "ot/base/NumUtils.h"
#include "ot/base/StringUtils.h"
#include "ot/util/AttributeListParser.h"
#include "ot/io/FileOutputStream.h"
#include "ot/io/OutputStreamWriter.h"
#include "ot/net/Socket.h"

OT_AUXIL_NAMESPACE_BEGIN

using namespace util;
using namespace io;
using namespace net;

const CharType* szFile = OT_T("file");
const CharType* szHost = OT_T("host");
const CharType* szPort = OT_T("port");
const CharType* szFlush = OT_T("flush");
const CharType* szAll  = OT_T("all");
const CharType FieldSep = OT_T('|');

//==============================================================================
// TraceHelper::TraceHelper
//
//==============================================================================
TraceHelper::TraceHelper(const String& name, const String& command) :
	m_nAllLevel(0),
	m_name(name),
	m_bAutoFlush(false)
{
	//
	// Initialize our section table
	//
	for(short i=0; i<MaxSections; ++i)
	{
		m_sectionTable[i] = 0;
	}
	
	//
	// Parse the command string, which is of the form:
	// "File=filename flush=flush <section>=<number> <section>=<number>..."
	//
	if(!command.empty())
	{
		AttributeListParser parser;
		parser.parseString(command);
		const String& filename=parser.getAttributeValueICase(szFile);
		const String& host=parser.getAttributeValueICase(szHost);
		if(filename.size())
		{
			m_rpWriter = new OutputStreamWriter(new FileOutputStream(filename));
		}
		else if(host.size())
		{
			int port = 1122;
			const String& sPort=parser.getAttributeValueICase(szPort);
			if(sPort.size())
				port = NumUtils::ToInt(sPort);
			RefPtr<Socket> rpSocket = new Socket(host, port);
			m_rpWriter = new OutputStreamWriter(rpSocket->getOutputStream().get());
		}

		parser.removeAttribute(szFile);
		parser.removeAttribute(szHost);
		parser.removeAttribute(szPort);

		if(m_rpWriter)
		{
			m_rpWriter->write(name);
			m_rpWriter->write(OT_T(" trace started at "));
			m_rpWriter->write(DateTime::GetSystemTime().toString());
			m_rpWriter->write(System::GetLineEnding());
			m_rpWriter->write(OT_T("Trace options: "));
			m_rpWriter->write(command);
			m_rpWriter->write(System::GetLineEnding());
			m_rpWriter->flush();

			for(size_t i=0; i<parser.getAttributeCount(); ++i)
			{
				const String& section = parser.getAttributeName(i);

				if(StringUtils::CompareNoCase(section, szFlush)==0)
				{
					m_bAutoFlush = true;
				}
				else
				{
					int level = NumUtils::ToInt(parser.getAttributeValue(i));
					short nSection = GetSectionNumber(section);
					if(nSection != -1)
					{
						doActivate(nSection, level);
					}
				}
			}
		}
	}
}

//==============================================================================
// TraceHelper::~TraceHelper
//
//==============================================================================
TraceHelper::~TraceHelper()
{
	Enable(false);

	if(m_rpWriter)
	{
		try
		{
			m_rpWriter->write(OT_T("stopped at "));
			m_rpWriter->write(DateTime::GetSystemTime().toString());
			m_rpWriter->write(System::GetLineEnding());
			m_rpWriter->close();
		}
		catch(Exception& /*e*/)
		{
		}
	}
}

//==============================================================================
// TraceHelper::doTrace
//
//==============================================================================
void TraceHelper::doTrace(short nSection, short nLevel, const CharType* message, size_t len)
{
	if(nSection > 0 && nSection < MaxSections && m_rpWriter && (nLevel <= m_nAllLevel || nLevel <= m_sectionTable[nSection]))
	{
		try
		{
			OT_SYNCHRONIZED_PTR(m_rpWriter->getLock().get())
			// Disable tracing to avoid infinite recursion where
			// tracing mechanism itself causes tracing output
			Enable(false);
			formatOutput(nSection, nLevel);
			m_rpWriter->write(message, len);
			m_rpWriter->write(System::GetLineEnding());
			if(m_bAutoFlush)
				m_rpWriter->flush();
			Enable(true);
		}
		catch(Exception&)
		{
			Enable(true);
		}
	}
}

//==============================================================================
// TraceHelper::doTrace
//
//==============================================================================
void TraceHelper::doTraceBytes(short nSection, short nLevel,
                               const String& message, const Byte* bytes, size_t len)
{
	if(nSection > 0 && nSection < MaxSections && m_rpWriter && (nLevel <= m_nAllLevel || nLevel <= m_sectionTable[nSection]))
	{
		try
		{
			OT_SYNCHRONIZED_PTR(m_rpWriter->getLock().get())
			// Disable tracing to avoid infinite recursion where
			// tracing mechanism itself causes tracing output
			Enable(false);
			formatOutput(nSection, nLevel);
			m_rpWriter->write(message);
			m_rpWriter->write(StringUtils::FromLatin1((char*)bytes, len));
			m_rpWriter->write(System::GetLineEnding());
			if(m_bAutoFlush)
				m_rpWriter->flush();
			Enable(true);
		}
		catch(Exception&)
		{
			Enable(true);
		}
	}
}

//==============================================================================
// TraceHelper::formatOutput
//
// MT note: the writer's lock should be held prior to calling this
//==============================================================================
void TraceHelper::formatOutput(short nSection, short nLevel)
{
	m_rpWriter->write(FieldSep);
	m_rpWriter->write(NumUtils::ToString(DateTime::GetSystemTime().toAnsiTime()));
	m_rpWriter->write(FieldSep);

#ifdef OT_MT
	m_rpWriter->write(Thread::CurrentThreadId().toString());
	m_rpWriter->write(FieldSep);
#endif //OT_MT

	const CharType* pSectionName = GetSectionName(nSection);
	if(pSectionName)
	{
		m_rpWriter->write(pSectionName);
	}
	else
	{
		m_rpWriter->write(OT_T("unknown"));
	}
	m_rpWriter->write(StringUtils::FromLatin1(StringUtils::Format("|%03d|", (int)nLevel)));
}

//==============================================================================
// TraceHelper::doActivate
//
// todo protect this, which means we should set the writer lock equal to our id
//==============================================================================
void TraceHelper::doActivate(short nSection, short nLevel)
{
	if(nSection == Tracer::All)
	{
		m_nAllLevel = nLevel;
	}
	else if(nSection > 0 && nSection < MaxSections)
	{
		m_sectionTable[nSection] = nLevel;
	}

	if(nLevel)
	{
		Tracer::Enable(true);
	}
}

OT_AUXIL_NAMESPACE_END
