//==============================================================================
// This file is part of the OpenTop C++ Library.
// (c) Copyright 2000-2002 ElCel Technology Limited
//
// This file may be distributed and/or modified under the terms of the
// GNU General Public License version 2 as published by the Free Software
// Foundation and appearing in the file LICENSE.GPL included in the
// packaging of this file.
//
// Licensees holding valid OpenTop Commercial licences may use this file
// in accordance with the OpenTop Commercial License Agreement included
// with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// See http://www.elcel.com/products/opentop/pricing.html or email
// sales@elcel.com for information about OpenTop Commercial License Agreements.
//
//==============================================================================
//
// $Revision: 1.1.1.1 $
// $Date: 2002/12/14 00:24:34 $
//
//==============================================================================
//
// Class: AtomicCounter
// 
/**
	@class ot::AtomicCounter
	
	@brief Integer value that can be incremented and
	decremented.  In a multi-threaded environment, access to the counter is
	synchronized to prevent lost updates.

*/
//==============================================================================

#include "AtomicCounter.h"

#if defined(WIN32)
	#include "ot/base/winincl.h"
#endif

OT_BASE_NAMESPACE_BEGIN

//==============================================================================
// AtomicCounter::AtomicCounter
//
/**
   Default constructor.

   Initializes the counter to zero.
*/
//==============================================================================
AtomicCounter::AtomicCounter() : 
	m_count(0)
{
}

//==============================================================================
// AtomicCounter::AtomicCounter
//
/**
   Constructs an AtomicCounter with an initial value.

   @param n the initial value of the counter.
*/
//==============================================================================
AtomicCounter::AtomicCounter(long n) :
	m_count(n)
{
}

//==============================================================================
// AtomicCounter::operator
//
/**
   Assignment operator.

   Assigns a new value to the AtomicCounter.
   @param n the new value of the counter.
   @returns the new value
*/
//==============================================================================
unsigned long AtomicCounter::operator=(long n)
{
	return m_count = n;
}

//==============================================================================
// AtomicCounter::operator++
//
/**
   Prefix increment operator.

   Increments the counter by one and returns the new value as a single atomic
   operation.
*/
//==============================================================================
unsigned long AtomicCounter::operator++()
{
#if defined(WIN32)
	// Note: not guaranteed to return correct result on Win95/NT3.51
	return ::InterlockedIncrement((long*)&m_count);
#else
	#if defined(OT_MT)
	FastMutex::Lock lock(m_mutex);
	#endif //OT_MT
	return ++m_count;
#endif
}

//==============================================================================
// AtomicCounter::operator++(int)
//
/**
   Postfix increment operator.

   Increments the counter by one and returns the original value as a 
   synchronized atomic operation.
*/
//==============================================================================
unsigned long AtomicCounter::operator++(int)
{
#if defined(WIN32)
	// Note: not guaranteed to return correct result on Win95/NT3.51
	return ::InterlockedIncrement((long*)&m_count)-1;
#else
	#if defined(OT_MT)
	FastMutex::Lock lock(m_mutex);
	#endif //OT_MT
	return m_count++;
#endif
}

//==============================================================================
// AtomicCounter::operator--
//
/**
   Prefix decrement operator.

   Decrements the counter by one and returns the new value as a 
   synchronized atomic operation.
*/
//==============================================================================
unsigned long AtomicCounter::operator--()
{
#if defined(WIN32)
	// Note: not guaranteed to return correct result on Win95/NT3.51
	return ::InterlockedDecrement((long*)&m_count);
#else
	#if defined(OT_MT)
	FastMutex::Lock lock(m_mutex);
	#endif //OT_MT
	return --m_count;
#endif
}

//==============================================================================
// AtomicCounter::operator--(int)
//
/**
   Postfix decrement operator.

   Decrements the counter by one and returns the original value as a 
   synchronized atomic operation.
*/
//==============================================================================
unsigned long AtomicCounter::operator--(int)
{
#if defined(WIN32)
	// Note: not guaranteed to return correct result on Win95/NT3.51
	return ::InterlockedDecrement((long*)&m_count)+1;
#else
	#if defined(OT_MT)
	FastMutex::Lock lock(m_mutex);
	#endif //OT_MT
	return m_count--;
#endif
}

//==============================================================================
// AtomicCounter::operator long
//
/**
   Conversion operator.

   Returns the current value of the counter.
*/
//==============================================================================
AtomicCounter::operator unsigned long() const
{
	return m_count;
}

OT_BASE_NAMESPACE_END
