//==============================================================================
// This file is part of the OpenTop C++ Library.
// (c) Copyright 2000-2002 ElCel Technology Limited
//
// This file may be distributed and/or modified under the terms of the
// GNU General Public License version 2 as published by the Free Software
// Foundation and appearing in the file LICENSE.GPL included in the
// packaging of this file.
//
// Licensees holding valid OpenTop Commercial licences may use this file
// in accordance with the OpenTop Commercial License Agreement included
// with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// See http://www.elcel.com/products/opentop/pricing.html or email
// sales@elcel.com for information about OpenTop Commercial License Agreements.
//
//==============================================================================
//
// $Revision: 1.1.1.1 $
// $Date: 2002/12/14 00:24:35 $
//
//==============================================================================
// Class Character
/**
	@class ot::Character
	
	@brief Represents a Unicode character using an internal sequence of one or more 
	::CharType characters.

	It provides optimized routines for converting Unicode characters into
	a sequence of one or more ::CharType characters and for decoding a multi-character
	sequence into a Unicode code-point ::UCS4Char value.

    The Character class also contains a number of convenient methods for
	querying the characteristics of the encoded Unicode character.  These routines
	such as isHexDigit() and isSpace() are simply wrappers for functions in the 
	Unicode class.  They have counterparts in the standard C++ library,
	but the standard library routines rely on the capabilities of a locale
	which may not be available for Unicode.  The Unicode class does
	not suffer from this drawback.
*/
//==============================================================================

#include "Character.h"
#include "SystemCodeConverter.h"
#include "IllegalCharacterException.h"
#include "UnicodeCharacterType.h"

OT_BASE_NAMESPACE_BEGIN

//==============================================================================
// Character::EndOfFileCharacter
//
/**
   Character representing the 'end of file' condition.

   This is a special Character that can be returned from functions that read
   a single Character when the end of file condition has been reached.
*/
//==============================================================================
Character Character::EndOfFileCharacter;

//==============================================================================
// Character::Character
//
/**
   Constructs a Character with an internal ::CharType sequence equivalent to 
   the Unicode character represented by the value of @c ch.

   @throws IllegalCharacterException if @c ch is not a legal Unicode character
           in the range U+0000-U+10FFFF.
*/
//==============================================================================
Character::Character(UCS4Char ch)
{
	if(ch > 0x10FFFFUL) throw IllegalCharacterException();

	CharType* pNext;
	const CharType* pEnd = m_data+MaxSeqLen;
	SystemCodeConverter::Result res = 
		SystemCodeConverter::ToInternalEncoding(ch, m_data, pEnd, pNext);
	if(res == SystemCodeConverter::ok)
	{
		m_length = (pNext-m_data);
	}
	else
	{
		throw IllegalCharacterException();
	}
}

OT_BASE_NAMESPACE_END
