//==============================================================================
// This file is part of the OpenTop C++ Library.
// (c) Copyright 2000-2002 ElCel Technology Limited
//
// This file may be distributed and/or modified under the terms of the
// GNU General Public License version 2 as published by the Free Software
// Foundation and appearing in the file LICENSE.GPL included in the
// packaging of this file.
//
// Licensees holding valid OpenTop Commercial licences may use this file
// in accordance with the OpenTop Commercial License Agreement included
// with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// See http://www.elcel.com/products/opentop/pricing.html or email
// sales@elcel.com for information about OpenTop Commercial License Agreements.
//
//==============================================================================
//
// $Revision: 1.3 $
// $Date: 2003/08/25 08:16:48 $
//
//==============================================================================
//
// Class: Exception
// 
/**
	@class ot::Exception
	
	@brief Base class for all exceptions thrown from the @OpenTop library.

	Functions within @OpenTop rarely throw an exception of type @c Exception,
	but usually throw one of the many derivatives such as net::SocketException
	when a socket error occurs.  However, applications may @a catch exceptions
	of type @c Exception and extract a meaningful error message by calling the
	toString() member function.

    @code
    try
    {
        //  open a FileInputStream (may throw a FileNotFoundException)
        RefPtr<FileInputStream> rpFIS = new FileInputStream(OT_T("test.txt"));
    }
    catch(Exception& e)
    {
        std::err << StringUtils::ToAscii(e.toString()) << std::endl;
    }
    @endcode
*/
//==============================================================================

#include "Exception.h"

OT_BASE_NAMESPACE_BEGIN

const String sColon = OT_T(": ");

//==============================================================================
// Exception::Exception
//
/**
   Constructs an Exception without any additional message description.
*/
//==============================================================================
Exception::Exception()
{
}

//==============================================================================
// Exception::Exception
//
/**
   Constructs an Exception with a description message.
   @param message a String providing further information about the exception
*/
//==============================================================================
Exception::Exception(const String& message) :
	m_message(message)
{
}

//==============================================================================
// Exception::~Exception
//
/**
   Virtual destructor.  Ensures that derived class' destructor will be called
   if a pointer to an Exception is ever deleted.  Note, however, that @OpenTop
   always throws exceptions by value and catches them by reference.
*/
//==============================================================================
Exception::~Exception()
{
}

//==============================================================================
// Exception::setMessage
//
/**
   Sets the message text for an Exception.  Intended to be called by the
   constructors of derived exception classes.
   @param message the message text for the exception
*/
//==============================================================================
void Exception::setMessage(const String& message)
{
	m_message = message;
}

//==============================================================================
// Exception::getMessage
//
/**
   Returns the message text for this Exception.
*/
//==============================================================================
String Exception::getMessage() const
{
	return m_message;
}

//==============================================================================
// Exception::toString
//
/**
   Returns a string representation of this Exception.

   If this Exception has a non-empty message text, the return value is
   made up as follows:-
   - the description returned from getDescription()
   - a colon followed by a space (": ")
   - the message text

   If the message text is empty then toString() returns the same value as
   getDescription().
*/
//==============================================================================
String Exception::toString() const
{
	const String& msg = getMessage();

	if(msg.empty())
	{
		return getDescription();
	}
	else
	{
		return String(getDescription() + sColon + msg);
	}
}

//==============================================================================
// Exception::getExceptionType
//
/**
   Returns the class name as a String.

   This method should be overridden in all derived classes to return the name
   of the derived class.
   
   @sa getDescription()
*/
//==============================================================================
String Exception::getExceptionType() const
{
	return OT_T("Exception");
}

//==============================================================================
// Exception::getLocalizedDescription
//
/**
   Returns a localized description of the Exception type or, if a native
   language translation is unavailable, an empty string.

   The base class implementation always returns an empty string.
   
   @sa getDescription()
*/
//==============================================================================
String Exception::getLocalizedDescription() const
{
	return String();
}

//==============================================================================
// Exception::getDescription
//
/**
   Returns a description of this Exception type.

   If getLocalizedDescription() returns a non-empty string, then it is
   returned, otherwise the exception class name returned from getExceptionType()
   is returned.
*/
//==============================================================================
String Exception::getDescription() const
{
	const String& localizedDescr = getLocalizedDescription();
	if(localizedDescr.empty())
	{
		return getExceptionType();
	}
	else
	{
		return localizedDescr;
	}
}

OT_BASE_NAMESPACE_END
