//==============================================================================
// This file is part of the OpenTop C++ Library.
// (c) Copyright 2000-2002 ElCel Technology Limited
//
// This file may be distributed and/or modified under the terms of the
// GNU General Public License version 2 as published by the Free Software
// Foundation and appearing in the file LICENSE.GPL included in the
// packaging of this file.
//
// Licensees holding valid OpenTop Commercial licences may use this file
// in accordance with the OpenTop Commercial License Agreement included
// with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// See http://www.elcel.com/products/opentop/pricing.html or email
// sales@elcel.com for information about OpenTop Commercial License Agreements.
//
//==============================================================================
//
// $Revision: 1.1.1.1 $
// $Date: 2002/12/14 00:24:39 $
//
//==============================================================================
// Class ObjectManager
/**
	@class ot::ObjectManager
	
	@brief Manages the lifetime of ManagedObject instances
	that need to exist for the duration of the application. 
	
	@OpenTop contains several classes that use global objects which can be
	created by application code or internally within the
	library.  To enable these objects (and others) to be deleted at system termination,
	they are registered with a singleton ObjectManager which is responsible for
	keeping a reference to them until the process terminates.

    The task of deleting objects at process termination is seldom strictly
	required because the underlying operating system will free any resources held
	by the process upon termination.  However, it is sometimes useful in a development
	environment to help to detect memory leaks in long-running processes.

	@sa System::GetObjectManager()
*/
//==============================================================================

#include "ObjectManager.h"
#include "ManagedObject.h"

OT_BASE_NAMESPACE_BEGIN

//==============================================================================
// ObjectManager::registerObject
//
/**
   Registers an instance of ManagedObject with this ObjectManager.

   The ObjectManager adds a reference to the passed object to its collection
   of object references.  In so doing, the reference count of the passed object 
   is incremented to prevent it from being destroyed until system termination
   or until the object is removed by calling unregisterObject().

   @sa unregisterObject()
   @param pObject pointer to a ManagedObject whose lifetime needs to be managed
          by the ObjectManager
*/
//==============================================================================
void ObjectManager::registerObject(ManagedObject* pObject)
{
	OT_AUTO_LOCK(FastMutex, m_mutex);
	m_list.push_back(pObject);
}

//==============================================================================
// ObjectManager::unregisterObject
//
/**
   Removes the reference to @c pObject from the managed collection of object
   references.
   
   It is not an error if the object designated by @c pObject does not exist
   in the managed collection.

   @param pObject pointer to a ManagedObject which will be removed from the
          managed collection of object references

*/
//==============================================================================
void ObjectManager::unregisterObject(ManagedObject* pObject)
{
	OT_AUTO_LOCK(FastMutex, m_mutex);
	m_list.remove(pObject);
}

//==============================================================================
// ObjectManager::unregisterAllObjects
//
// Private function used by System::Terminate to sweep through and unregister
// all objects before this ObjectManager is destroyed.
//==============================================================================
void ObjectManager::unregisterAllObjects()
{
	//
	// Note: when an object is unregistered, it's destrictor may be called
	// which could possibly choose to unregister another object.  This is
	// not recommended behaviour (the object will have been unregistered
	// anyway due to system termination), but to prevent a hang on the
	// mutex lock if this situation were to occur, a copy of the registered
	// objects list is made while under mutex control, the mutex is then
	// released and the copied list allowed to go out of scope.

#ifdef OT_MT

	FastMutex::Lock scoped_lock(m_mutex);
	ObjectList copy = m_list;
	m_list.clear();
	scoped_lock.unlock();

#else  // !OT_MT

	m_list.clear();

#endif //OT_MT
}

OT_BASE_NAMESPACE_END
