//==============================================================================
// This file is part of the OpenTop C++ Library.
// (c) Copyright 2000-2002 ElCel Technology Limited
//
// This file may be distributed and/or modified under the terms of the
// GNU General Public License version 2 as published by the Free Software
// Foundation and appearing in the file LICENSE.GPL included in the
// packaging of this file.
//
// Licensees holding valid OpenTop Commercial licences may use this file
// in accordance with the OpenTop Commercial License Agreement included
// with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// See http://www.elcel.com/products/opentop/pricing.html or email
// sales@elcel.com for information about OpenTop Commercial License Agreements.
//
//==============================================================================
//
// $Revision: 1.2 $
// $Date: 2003/05/05 17:25:00 $
//
//==============================================================================

#include "SystemUtils.h"
#include "IllegalArgumentException.h"
#include "NullPointerException.h"
#include "NumUtils.h"
#include "StringUtils.h"
#include "Tracer.h"
#include "debug.h"

#include <algorithm>
#include <limits.h>
#include <stdlib.h>
#include <errno.h>

OT_BASE_NAMESPACE_BEGIN

//==============================================================================
// SystemUtils::GetSystemErrorString
//
// Portable routine to return an error message associated with the passed
// error code.
//==============================================================================
String SystemUtils::GetSystemErrorString(long errorNum)
{
	errorNum = errorNum ? errorNum : errno;
	return StringUtils::FromNativeMBCS(strerror(errorNum))
		+ StringUtils::FromLatin1(StringUtils::Format(" (%ld)", errorNum));
}

//==============================================================================
// SystemUtils::TraceSystemCall
//
//==============================================================================
void SystemUtils::TraceSystemCall(short nSection, short nLevel, const String& message, int rc)
{
	static String sRC = OT_T(" rc=");
	static String sErrno = OT_T(", errno=");
	static String sComma = OT_T(", ");
	int err=errno;

	String traceMsg = message + sRC + NumUtils::ToString(rc);
	if(rc!=0)
	{
		traceMsg += sErrno  + NumUtils::ToString(err)
	             + sComma + GetSystemErrorString(err);
	}

	Tracer::Trace(nSection, nLevel, traceMsg);
}

//==============================================================================
// SystemUtils::TestBufferIsValid
//
// Test that the passed buffer is appropriate for calls to read() operations.
//==============================================================================
void SystemUtils::TestBufferIsValid(const void* pBuffer, size_t& bufLen)
{
	if(!pBuffer) throw NullPointerException();
	if(!bufLen)  throw IllegalArgumentException(OT_T("zero buffer length"));
	if(bufLen > LONG_MAX)
		bufLen = LONG_MAX;
}

#ifdef WIN32

//==============================================================================
// SystemUtils::GetWin32ErrorString
//
// Return an error message associated with the passed WIN32 error code.
//==============================================================================
String SystemUtils::GetWin32ErrorString(DWORD errNo)
{
	LPVOID lpMessageBuffer = 0;

	::FormatMessage(
		FORMAT_MESSAGE_ALLOCATE_BUFFER |
		FORMAT_MESSAGE_FROM_SYSTEM,
		NULL,
		errNo,
		MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), //default language
		(LPTSTR) &lpMessageBuffer,
		0,
		NULL);

	String ret = StringUtils::FromWin32String((LPCTSTR)lpMessageBuffer);

	::LocalFree((HLOCAL) lpMessageBuffer);

	//
	// Strip trailing .crlf if present
	//
	if(ret.size()>2 && (ret[ret.size()-3] == '.'))
	{
		ret.erase(ret.size()-3);
	}
	return ret;
}

#endif // !WIN32

OT_BASE_NAMESPACE_END

