//==============================================================================
// This file is part of the OpenTop C++ Library.
// (c) Copyright 2000-2002 ElCel Technology Limited
//
// This file may be distributed and/or modified under the terms of the
// GNU General Public License version 2 as published by the Free Software
// Foundation and appearing in the file LICENSE.GPL included in the
// packaging of this file.
//
// Licensees holding valid OpenTop Commercial licences may use this file
// in accordance with the OpenTop Commercial License Agreement included
// with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// See http://www.elcel.com/products/opentop/pricing.html or email
// sales@elcel.com for information about OpenTop Commercial License Agreements.
//
//==============================================================================
//
// $Revision: 1.1.1.1 $
// $Date: 2002/12/14 00:24:47 $
//
//==============================================================================
//
// Class: ThreadId
// 
/**
	@class ot::ThreadId
	
	@brief Represents an abstract thread identifier. 

	A ThreadId uniquely identifies a single Thread for the duration that the 
	Thread is active.  After a Thread has terminated its ThreadId may be reused
	by a new Thread. 
*/
//==============================================================================

#include "ThreadId.h"
#include "NumUtils.h"

#ifdef OT_MT

OT_BASE_NAMESPACE_BEGIN

//==============================================================================
// ThreadId::ThreadId
//
/**
   Default constructor.
*/
//==============================================================================
ThreadId::ThreadId() :
	m_nativeId(0)
{
}

//==============================================================================
// ThreadId::ThreadId
//
/**
   Copy constructor.
   @param rhs the ThreadId to copy
*/
//==============================================================================
ThreadId::ThreadId(const ThreadId& rhs) :
	m_nativeId(rhs.m_nativeId)
{
}

//==============================================================================
// ThreadId::ThreadId
//
/**
   Constructs a ThreadId from a native thread identifier as used by the underlying
   threading library.
   @param id the native thread identifier
*/
//==============================================================================
ThreadId::ThreadId(NativeId id) :
	m_nativeId(id)
{
}

//==============================================================================
// ThreadId::operator=
//
/**
   Assignment operator.  Sets this ThreadId equal to the native thread identifier
   @c id.
*/
//==============================================================================
ThreadId& ThreadId::operator=(NativeId id)
{
	m_nativeId = id;
	return *this;
}

//==============================================================================
// ThreadId::operator==
//
/**
   Equality operator.  Returns true if this ThreadId is equal to the
   ThreadId @c rhs.
   @returns true if the two ThreadId objects represent the same Thread; false
            otherwise
*/
//==============================================================================
bool ThreadId::operator==(const ThreadId& rhs) const
{
#if defined(OT_WIN32_THREADS)

	return (m_nativeId == rhs.m_nativeId);

#elif defined(OT_POSIX_THREADS)

	return (::pthread_equal(m_nativeId, rhs.m_nativeId) != 0);

#else
	#error Unsupported configuration
#endif
}

//==============================================================================
// ThreadId::operator!=
//
/**
   Inequality operator.  Returns true if this ThreadId is not equal to the
   ThreadId @c rhs.
   @returns false if the two ThreadId objects represent the same Thread; true
            otherwise
*/
//==============================================================================
bool ThreadId::operator!=(const ThreadId& rhs) const
{
	return !(*this==rhs);
}

//==============================================================================
// ThreadId::NativeId 
//
/**
   Returns the native thread identifier.
*/
//==============================================================================
ThreadId::NativeId ThreadId::getNativeId() const
{
	return m_nativeId;
}

//==============================================================================
// ThreadId::toString
//
/**
   Returns a string representation of the object.
*/
//==============================================================================
String ThreadId::toString() const
{
	return NumUtils::ToString((unsigned long)m_nativeId);
}

OT_BASE_NAMESPACE_END

#endif //OT_MT
