//==============================================================================
// This file is part of the OpenTop C++ Library.
// (c) Copyright 2000-2002 ElCel Technology Limited
//
// This file may be distributed and/or modified under the terms of the
// GNU General Public License version 2 as published by the Free Software
// Foundation and appearing in the file LICENSE.GPL included in the
// packaging of this file.
//
// Licensees holding valid OpenTop Commercial licences may use this file
// in accordance with the OpenTop Commercial License Agreement included
// with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// See http://www.elcel.com/products/opentop/pricing.html or email
// sales@elcel.com for information about OpenTop Commercial License Agreements.
//
//==============================================================================
//
// $Revision: 1.1.1.1 $
// $Date: 2002/12/14 00:24:47 $
//
//==============================================================================
//
// Class: ThreadLocal
//
/**
	@class ot::ThreadLocal
	
	@brief Provides a class interface to native thread-local storage which
	provides each thread with its own copy of a variable.
	
	Each thread's copy of a ThreadLocal variable is initially set to zero.

    The variable type is a @c void pointer that can be used to hold a pointer
	or it can be cast to a long integer value.

    In order for a ThreadLocal variable to be available to all threads, it is
	often convenient for instances of ThreadLocal to be created statically
	during program initialization.  In this way the variable exists for the lifetime
	of the application, and any thread can access its individual value at any time.

    In the following example, a static ThreadLocal variable is used to hold
	a pointer to an instance of an imaginary class @c WorkerThreadInfo.
	In this case, if the current thread does not have an instance of @c WorkerThreadInfo
	then it creates one and stores the pointer for subsequent use.

    @code
    static ThreadLocal ThreadInfoPointer;

    WorkerThreadInfo* getThreadInfo()
    {
        WorkerThreadInfo* pInfo = ThreadInfoPointer.get();
        if(!pInfo)
        {
            pInfo = new WorkerThreadInfo(Thread::CurrentThreadId());
            ThreadInfoPointer.set(pInfo);
        }
        return pInfo;
    }
    @endcode

    Note that the above example contains a potential memory leak. When a thread terminates,
	the ThreadLocal variable for that particular thread becomes inaccessible.  If
	the variable contains a pointer to allocated memory then the memory is @em not
	automatically freed.  For this reason it is a good idea to put any necessary clean-up
	code into the Runnable::run() method.
*/
//==============================================================================

#include "ThreadLocal.h"
#include "OSException.h"

#ifdef WIN32
#include "Win32Exception.h"
#endif //WIN32

#ifdef OT_MT

OT_BASE_NAMESPACE_BEGIN

//==============================================================================
// ThreadLocal::ThreadLocal
//
/**
   Default constructor.  Uses the operating-system's threading library to 
   allocate a new thread-local variable.

   The value of the variable is automatically initialized to zero for every
   thread.
*/
//==============================================================================
ThreadLocal::ThreadLocal()
{
#if defined(OT_WIN32_THREADS)

	m_key = ::TlsAlloc();
	if(m_key == (DWORD)-1)
	{
		throw Win32Exception(::GetLastError());
	}

#elif defined(OT_POSIX_THREADS)

	int status = ::pthread_key_create(&m_key, 0);
	if(status != 0)
	{
		throw OSException(status, OT_T("pthread_key_create"));
	}

#endif
}

//==============================================================================
// ThreadLocal::~ThreadLocal
//
/**
   Destructor.  Frees the thread-local variable, making it inaccessible from
   any thread.
*/
//==============================================================================
ThreadLocal::~ThreadLocal()
{
#if defined(OT_WIN32_THREADS)
	::TlsFree(m_key);
#elif defined(OT_POSIX_THREADS)
	::pthread_key_delete(m_key);
#endif
}
	
//==============================================================================
// ThreadLocal::set
//
/**
   Sets the value of this ThreadLocal for the currently executing thread.
   @param value the value to set this ThreadLocal variable to for the current thread
   @sa get()
*/
//==============================================================================
void ThreadLocal::set(void* value) const
{
#if defined(OT_WIN32_THREADS)

	::TlsSetValue(m_key, value);

#elif defined(OT_POSIX_THREADS)

	::pthread_setspecific(m_key, value);

#endif
}

//==============================================================================
// ThreadLocal::get
//
/**
   Returns the current value of this ThreadLocal variable for the current thread.
   @sa set()
*/
//==============================================================================
void* ThreadLocal::get() const
{
#if defined(OT_WIN32_THREADS)
	
	return ::TlsGetValue(m_key);

#elif defined(OT_POSIX_THREADS)

	return ::pthread_getspecific(m_key);

#endif
}

OT_BASE_NAMESPACE_END

#endif //OT_MT
