//==============================================================================
// This file is part of the OpenTop C++ Library.
// (c) Copyright 2000-2002 ElCel Technology Limited
//
// This file may be distributed and/or modified under the terms of the
// GNU General Public License version 2 as published by the Free Software
// Foundation and appearing in the file LICENSE.GPL included in the
// packaging of this file.
//
// Licensees holding valid OpenTop Commercial licences may use this file
// in accordance with the OpenTop Commercial License Agreement included
// with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// See http://www.elcel.com/products/opentop/pricing.html or email
// sales@elcel.com for information about OpenTop Commercial License Agreements.
//
//==============================================================================
//
// $Revision: 1.2 $
// $Date: 2003/08/10 22:02:37 $
//
//==============================================================================
//
// Class: ASCIIConverter
// 
// US-ASCII characters are 7-bit characters that share the same bit representation
// as the first 128 ISO10646 (UNICODE) characters, therefore no conversion is
// necessary when decoding from ASCII to Unicode.  Only characters up to 0x7F 
// can be encoded anything higher is an error.
//
//==============================================================================

#include "ASCIIConverter.h"

#include "ot/base/SystemCodeConverter.h"

OT_CVT_NAMESPACE_BEGIN

//==============================================================================
// ASCIIConverter::decode
//
//==============================================================================
CodeConverter::Result ASCIIConverter::decode(const Byte *from, const Byte *from_end,
                                            const Byte *& from_next,
                                            CharType *to, CharType *to_limit,
                                            CharType *& to_next)
{
	Result ret = ok;
	from_next = from, to_next = to;

	//
	// decode Bytes into Characters until 
	// one of the buffers is exhausted
	//
	while(from_next < from_end && to_next < to_limit && ret == ok)
	{
		//
		// If the input byte is ASCII then we can convert it directly into
		// an internal character - all internal encodings are equal for the ASCII range
		//
		if(*from_next < 0x80U)
		{
			*to_next++ = *from_next++;
		}
		else
		{
			if(getInvalidCharAction() == abort)
			{
				handleInvalidByteSequence(from_next, 1);
				ret = error;
			}
			else
			{
				OT_DBG_ASSERT(getInvalidCharAction() == replace);
				if( (ret=SystemCodeConverter::ToInternalEncoding(
					getInvalidCharReplacement(), to_next, to_limit, to_next))==ok)
				{
					++from_next;
				}
			}
		}
	}

	return ret; 
}

//==============================================================================
// ASCIIConverter::encode
//
//==============================================================================
CodeConverter::Result ASCIIConverter::encode(const CharType *from, const CharType *from_end,
                                             const CharType *& from_next,
                                             Byte *to, Byte *to_limit,
                                             Byte *& to_next)
{
	Result ret = ok;
	from_next = from, to_next = to;

	//
	// encode Characters into Bytes until 
	// one of the buffers is exhausted
	//
	while(from_next < from_end && to_next < to_limit && ret == ok)
	{
		UCS4Char ch;
		const CharType * from_next_copy;
		if( (ret=SystemCodeConverter::FromInternalEncoding(ch, from_next, from_end, from_next_copy)) == ok)
		{
			if (ch < 0x80U)	// needs just 1 byte
			{
				*to_next++ = Byte(ch);
				from_next = from_next_copy;
			}
			else // greater than 0x7F is an error for US-ASCII
			{
				if( (ret = handleUnmappableCharacter(ch, to_next, to_limit, to_next)) == ok)
				{
					from_next = from_next_copy;
				}
			}
		}
		else if(ret == error)
		{
			internalEncodingError(from_next, from_end-from_next);
		}
	}
	
	return ret;
}

//==============================================================================
// ASCIIConverter::getDecodedLength
//
// Returns the number of Unicode character that an external array of bytes will
// generate once decoded.
//==============================================================================
size_t ASCIIConverter::getDecodedLength(const Byte *from, const Byte *from_end) const
{
	return (from_end - from);
}

bool ASCIIConverter::alwaysNoConversion() const
{
	// even though ASCII will not need converting, it does require
	// validation.  By returning false here, we ensure that the encode/decode
	// methods are always called.
	return (false);
}

size_t ASCIIConverter::getMaxEncodedLength() const
{
	return 1;
}

//==============================================================================
// ASCIIConverter::getEncodingName
//
//==============================================================================
String ASCIIConverter::getEncodingName() const
{
	return OT_T("US-ASCII");
}

OT_CVT_NAMESPACE_END

