//==============================================================================
// This file is part of the OpenTop C++ Library.
// (c) Copyright 2000-2002 ElCel Technology Limited
//
// This file may be distributed and/or modified under the terms of the
// GNU General Public License version 2 as published by the Free Software
// Foundation and appearing in the file LICENSE.GPL included in the
// packaging of this file.
//
// Licensees holding valid OpenTop Commercial licences may use this file
// in accordance with the OpenTop Commercial License Agreement included
// with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// See http://www.elcel.com/products/opentop/pricing.html or email
// sales@elcel.com for information about OpenTop Commercial License Agreements.
//
//==============================================================================
//
// $Revision: 1.2 $
// $Date: 2003/08/10 22:02:37 $
//
//==============================================================================
//
// Class: ISO88591Converter
// 
// ISO-8859-1 characters share the same bit representation as the first 256
// Unicode characters, therefore no conversion is necessary when decoding
// from ISO-8859-1 to Unicode.  Only characters up to 0xFF can be encoded
// anything higher is an error.
//
//==============================================================================

#include "ISO88591Converter.h"

#include "ot/base/SystemCodeConverter.h"

OT_CVT_NAMESPACE_BEGIN

//==============================================================================
// ISO88591Converter::decode
//
//==============================================================================
CodeConverter::Result ISO88591Converter::decode(const Byte *from, const Byte *from_end,
                                            const Byte *& from_next,
                                            CharType *to, CharType *to_limit,
                                            CharType *& to_next)
{
	Result ret = ok;
	from_next = from, to_next = to;
	
	//
	// decode Bytes into Characters until 
	// one of the buffers is exhausted
	//
	while(ret == ok && from_next < from_end && to_next < to_limit)
	{
		// For ultimate efficiency, test if we are in the ascii range,
		// if so, there is no internal encoding reqd
		if(*from_next < 0x80U)
		{
			*to_next++ = *from_next++;
		}
		else
		{
			if( (ret=SystemCodeConverter::ToInternalEncoding(*from_next, to_next, to_limit, to_next)) == ok)
			{
				from_next++;
			}
		}
	}

	return (ret); 
}

//==============================================================================
// ISO88591Converter::encode
//
//==============================================================================
CodeConverter::Result ISO88591Converter::encode(const CharType *from, const CharType *from_end,
                                                const CharType *& from_next,
                                                Byte *to, Byte *to_limit,
                                                Byte *& to_next)
{
	Result ret = ok;
	from_next = from, to_next = to;

	//
	// encode Characters into Bytes until 
	// one of the buffers is exhausted
	//
	while(from_next < from_end && to_next < to_limit && ret == ok)
	{
		UCS4Char ch;
		const CharType * from_next_copy;
		if( (ret=SystemCodeConverter::FromInternalEncoding(ch, from_next, from_end, from_next_copy))==ok)
		{
			if (ch < 0xFFU)	// needs just 1 byte
			{
				*to_next++ = Byte(ch);
				from_next = from_next_copy;
			}
			else // greater than 0xFE is an error for ISO-8859-1
			{
				if( (ret = handleUnmappableCharacter(ch, to_next, to_limit, to_next)) == ok)
				{
					from_next = from_next_copy;
				}
			}
		}
		else if(ret == error)
		{
			internalEncodingError(from_next, from_end-from_next);
		}
	}
	
	return ret;
}

//==============================================================================
// ISO88591Converter::getDecodedLength
//
// Returns the number of Unicode characters that an external array of bytes will
// generate once decoded.
//==============================================================================
size_t ISO88591Converter::getDecodedLength(const Byte *from, const Byte *from_end) const
{
	return (from_end - from);
}

bool ISO88591Converter::alwaysNoConversion() const
{
	// Even though ISO-8859-1 is Unicode encoded into a single octet,
	// This doesn't mean that ISO-8859-1 files don't need decoding
	// into the encoding of the application.  In fact, there is not a Latin1
	// internal encoding, so conversion is always required.
	return (false);
}

size_t ISO88591Converter::getMaxEncodedLength() const
{
	return 1;
}

//==============================================================================
// ISO88591Converter::getEncodingName
//
//==============================================================================
String ISO88591Converter::getEncodingName() const
{
	return OT_T("ISO-8859-1");
}

OT_CVT_NAMESPACE_END

