//==============================================================================
// This file is part of the OpenTop C++ Library.
// (c) Copyright 2000-2002 ElCel Technology Limited
//
// This file may be distributed and/or modified under the terms of the
// GNU General Public License version 2 as published by the Free Software
// Foundation and appearing in the file LICENSE.GPL included in the
// packaging of this file.
//
// Licensees holding valid OpenTop Commercial licences may use this file
// in accordance with the OpenTop Commercial License Agreement included
// with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// See http://www.elcel.com/products/opentop/pricing.html or email
// sales@elcel.com for information about OpenTop Commercial License Agreements.
//
//==============================================================================
//
// $Revision: 1.1.1.1 $
// $Date: 2002/12/14 00:25:11 $
//
//==============================================================================
// 
// Class: BufferedWriter
// 
/**
	@class ot::io::BufferedWriter
	
	@brief A BufferedWriter wraps another Writer object and provides
	       buffering. 

    When the BufferedWriter is constructed, an internal buffer is created. The
	size of the internal buffer can be specified by using the overloaded
	constructor.

    When characters are written to this Writer, they are copied into
	the internal buffer instead of being written directly to the contained Writer.
	When the internal buffer becomes full it is automatically flushed to
	the contained Writer.  In this way i/o efficiency
	may be improved by translating multiple small write operations into one
	large one.
	
	To further improve efficiency and avoid the needless copying of data, if the
	application	performs a write request that is larger than the internal buffer,
	the internal buffer is bypassed and characters are written directly to the
	contained Writer.

    @mt
    The contained Writer is used as the lock object for synchronized methods.
*/
//=============================================================================


#include "BufferedWriter.h"

#include "ot/base/NullPointerException.h"

OT_IO_NAMESPACE_BEGIN

const size_t DefaultBufferSize = 1024;

//==============================================================================
// BufferedWriter::BufferedWriter
//
/**
   Constructs a BufferedWriter using the default buffer size and @c pWriter
   as the contained Writer.
   
   @throws NullPointerException if @c pWriter is null.
*/
//==============================================================================
BufferedWriter::BufferedWriter(Writer* pWriter) :
	m_rpWriter(pWriter)
{
	if(!pWriter) throw NullPointerException();
	m_rpLock = pWriter->getLock();

	init(DefaultBufferSize);
}

//==============================================================================
// BufferedWriter::BufferedWriter
//
/**
   Constructs a BufferedWriter using a buffer size of @c bufSize and 
   @c pWriter as the contained Writer.  If @c bufSize is zero, the
   default buffer size is used.

   @throws NullPointerException if @c pWriter is null.
*/
//==============================================================================
BufferedWriter::BufferedWriter(Writer* pWriter, size_t bufSize) :
	m_rpWriter(pWriter)
{
	if(!pWriter) throw NullPointerException();
	m_rpLock = pWriter->getLock();

	init(bufSize ? bufSize : DefaultBufferSize);
}

//==============================================================================
// BufferedWriter::~BufferedWriter
//
/**
   Flushes and destroys the character buffer before destroying this
   BufferedWriter.
*/
//==============================================================================
BufferedWriter::~BufferedWriter()
{
	try
	{
		flush();
	}
	catch(...)
	{
	}
	delete [] m_pBuffer;
}

//==============================================================================
// BufferedWriter::init
//
// Private initialization function.
//==============================================================================
void BufferedWriter::init(size_t bufferSize)
{
	m_bufferSize = bufferSize;
	m_pBuffer = new CharType[m_bufferSize];
	m_used=0;
}

//==============================================================================
// BufferedWriter::flush
//
//==============================================================================
void BufferedWriter::flush()
{
	OT_SYNCHRONIZED_PTR(m_rpLock)

	flushBuffersImpl();
	m_rpWriter->flush();
}

//==============================================================================
// BufferedWriter::flushBuffers
//
/**
   The internal character buffer is written to the contained Writer without
   requesting that Writer to flush().

   @sa flush()
   @synchronized
*/
//==============================================================================
void BufferedWriter::flushBuffers()
{
	OT_SYNCHRONIZED_PTR(m_rpLock)

	flushBuffersImpl();
}

//==============================================================================
// BufferedWriter::flushBuffersImpl
//
// MT Note: m_rpLock must be held before calling this method.
//==============================================================================
void BufferedWriter::flushBuffersImpl()
{
	m_rpWriter->write(m_pBuffer, m_used);
	m_used=0;
}

//==============================================================================
// BufferedWriter::write
//
//==============================================================================
void BufferedWriter::write(const CharType* pStr, size_t len)
{
	if(!pStr) throw NullPointerException();
	
	OT_SYNCHRONIZED_PTR(m_rpLock)
	
	if(m_used + len > m_bufferSize)
	{
		flushBuffersImpl();
	}

	if(len >= m_bufferSize)
	{
		OT_DBG_ASSERT(m_used == 0);
		m_rpWriter->write(pStr, len);
	}
	else
	{
		OT_DBG_ASSERT(len + m_used <= m_bufferSize);
		::memcpy(m_pBuffer+m_used, pStr, len*sizeof(CharType));
		m_used+=len;
	}
}

//==============================================================================
// BufferedWriter::close
//
//==============================================================================
void BufferedWriter::close()
{
	OT_SYNCHRONIZED_PTR(m_rpLock)

	flushBuffersImpl();
	m_rpWriter->close();
	// force writes through to closed stream
	delete [] m_pBuffer; m_pBuffer = 0;
	m_bufferSize = 0; 
}

OT_IO_NAMESPACE_END
