//==============================================================================
// This file is part of the OpenTop C++ Library.
// (c) Copyright 2000-2002 ElCel Technology Limited
//
// This file may be distributed and/or modified under the terms of the
// GNU General Public License version 2 as published by the Free Software
// Foundation and appearing in the file LICENSE.GPL included in the
// packaging of this file.
//
// Licensees holding valid OpenTop Commercial licences may use this file
// in accordance with the OpenTop Commercial License Agreement included
// with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// See http://www.elcel.com/products/opentop/pricing.html or email
// sales@elcel.com for information about OpenTop Commercial License Agreements.
//
//==============================================================================
//
// $Revision: 1.1.1.1 $
// $Date: 2002/12/14 00:25:11 $
//
//==============================================================================
//
// Class: ByteArrayInputStream
/**
	@class ot::io::ByteArrayInputStream
	
	@brief An input stream which uses a byte array as its input source.

	@sa ByteArrayOutputStream
*/
//==============================================================================

#include "ByteArrayInputStream.h"
#include "IOException.h"

#include "ot/base/NullPointerException.h"
#include "ot/base/SystemUtils.h"

OT_IO_NAMESPACE_BEGIN

//==============================================================================
// ByteArrayInputStream::ByteArrayInputStream
//
/**
   Constructs a ByteArrayInputStream with a copy of the byte array starting at
   @c pData for a length of @c dataLen.
   @throws NullPointerException if @c pData is null.
*/
//==============================================================================
ByteArrayInputStream::ByteArrayInputStream(const Byte* pData, size_t dataLen) :
	m_apBuffer(new Byte[dataLen]),
	m_bufSize(dataLen),
	m_pos(0),
	m_markPos(-1),
	m_bClosed(false)
{
	if(!pData) throw NullPointerException();

	::memcpy(m_apBuffer.get(), pData, dataLen);
}

//==============================================================================
// ByteArrayInputStream::close
//
//==============================================================================
void ByteArrayInputStream::close()
{
	m_markPos = -1;
	m_pos = m_bufSize = 0;
	delete [] m_apBuffer.release();
	m_bClosed = true;
}

//==============================================================================
// ByteArrayInputStream::mark
//
// Marks the current position in this stream. A subsequent call to the 
// reset method re-positions the stream at the last marked position so that 
// subsequent reads re-read the same bytes. 
//
// The readLimit has no effect for ByteArrayInputStreams, as the entire data
// buffer is always available until the stream is closed. 
//==============================================================================
void ByteArrayInputStream::mark(size_t /*readLimit*/)
{
	if(m_bClosed) throw IOException(OT_T("stream is closed"));

	m_markPos = m_pos;
}

//==============================================================================
// ByteArrayInputStream::reset
//
// Reset the position to the last mark() operation.  If there is no mark 
// (due to no mark() call then throw an IOException.
//==============================================================================
void ByteArrayInputStream::reset()
{
	if(m_markPos == -1)
	{
		throw IOException(OT_T("unable to reset input stream, no marked position"));
	}
	else
	{
		m_pos = m_markPos;
	}
}

//==============================================================================
// ByteArrayInputStream::markSupported
//
/**
   Tests whether the ByteArrayInputStream supports the mark() operation,
   which it does.
   
   @sa mark()
   @sa reset()
   @returns @c true for ByteArrayInputStream.
*/
//==============================================================================
bool ByteArrayInputStream::markSupported() const
{
	return true;
}

//==============================================================================
// ByteArrayInputStream::read
//
//==============================================================================
long ByteArrayInputStream::read(Byte* pBuffer, size_t bufLen)
{
	SystemUtils::TestBufferIsValid(pBuffer, bufLen);
	
	if(m_bClosed) throw IOException(OT_T("stream is closed"));

	const size_t bytesAvailable = (m_bufSize - m_pos);

	if(bytesAvailable)
	{
		const size_t count = (bufLen < bytesAvailable)
		                   ? bufLen
		                   : bytesAvailable;

		::memcpy(pBuffer, m_apBuffer.get()+m_pos, count);
		m_pos+=count;
		return count;
	}
	else
	{
		return EndOfFile;
	}
}

//==============================================================================
// ByteArrayInputStream::available
//
//==============================================================================
size_t ByteArrayInputStream::available()
{
	return (m_bufSize - m_pos);
}

OT_IO_NAMESPACE_END
