//==============================================================================
// This file is part of the OpenTop C++ Library.
// (c) Copyright 2000-2002 ElCel Technology Limited
//
// This file may be distributed and/or modified under the terms of the
// GNU General Public License version 2 as published by the Free Software
// Foundation and appearing in the file LICENSE.GPL included in the
// packaging of this file.
//
// Licensees holding valid OpenTop Commercial licences may use this file
// in accordance with the OpenTop Commercial License Agreement included
// with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// See http://www.elcel.com/products/opentop/pricing.html or email
// sales@elcel.com for information about OpenTop Commercial License Agreements.
//
//==============================================================================
//
// $Revision: 1.2 $
// $Date: 2003/08/10 22:07:41 $
//
//==============================================================================
// Class: Console
/**
	@class ot::io::Console
	
	@brief Class module providing access to the standard console i/o streams:
	In (@c stdin), Out (@c stdout) and Err (@c stderr).

	It cannot be instantiated - all methods are static.

    @mtsafeClass
*/
//==============================================================================

#include "Console.h"

#include "FileOutputStream.h"
#include "FileInputStream.h"
#include "FileDescriptor.h"
#include "InputStreamReader.h"
#include "FileSystem.h"
#include "PrintWriter.h"

#include "ot/base/System.h"
#include "ot/base/ObjectManager.h"

#include <memory>

OT_IO_NAMESPACE_BEGIN

using namespace io;

//==================================================================
// Multi-threaded locking strategy
//
// To ensure that singleton resources are not created by multiple
// concurrent threads, the pointers are protected using a
// single static mutex.
//
// To minimize the time that the ConsoleMutex is held, we do
// not lock the mutex for the creation of a singleton object,
// only when a complete object has been created and we need to swap
// the address into our static pointer variable.
//==================================================================
#ifdef OT_MT
	FastMutex ConsoleMutex;
#endif //OT_MT

InputStreamReader* OT_MT_VOLATILE Console::s_pIn = 0;
PrintWriter* OT_MT_VOLATILE Console::s_pOut = 0;
PrintWriter* OT_MT_VOLATILE Console::s_pErr = 0;

//==============================================================================
// Console::Out
//
/**
   Returns a PrintWriter connected to the @c stdout file handle.
   @mtsafe
*/
//==============================================================================
RefPtr<PrintWriter> Console::Out()
{
	//==================================================================
	// Multi-threaded locking strategy
	//
	// This uses the "double-checked locking pattern" (Schmidt 1996)
	// with a volatile storage member to minimise race conditions due
	// to the so-called "relaxed memory model"..
	//==================================================================
	if(!s_pOut)
	{
		OT_AUTO_LOCK(FastMutex, ConsoleMutex);
		// test again now that mutex has been locked
		if(!s_pOut)
		{
			s_pOut = new PrintWriter(
				new FileOutputStream(FileDescriptor::Out().get()), false);
			System::GetObjectManager().registerObject(s_pOut);
		}
	}

	return s_pOut;
}

//==============================================================================
// Console::Err
//
/**
   Returns a PrintWriter connected to the @c stderr file handle.
   
   @mtsafe
*/
//==============================================================================
RefPtr<PrintWriter> Console::Err()
{
	//==================================================================
	// Multi-threaded locking strategy
	//
	// This uses the "double-checked locking pattern" (Schmidt 1996)
	// with a volatile storage member to minimise race conditions due
	// to the so-called "relaxed memory model"..
	//==================================================================
	if(!s_pErr)
	{
		OT_AUTO_LOCK(FastMutex, ConsoleMutex);
		// test again now that mutex has been locked
		if(!s_pErr)
		{
			s_pErr = new PrintWriter(
				new FileOutputStream(FileDescriptor::Err().get()), true);
			System::GetObjectManager().registerObject(s_pErr);
		}
	}

	return s_pErr;
}

//==============================================================================
// Console::In
//
/**
   Returns an InputStreamReader connected to the @c stdin
   file handle.

   @mtsafe
*/
//==============================================================================
RefPtr<InputStreamReader> Console::In()
{
	//==================================================================
	// Multi-threaded locking strategy
	//
	// This uses the "double-checked locking pattern" (Schmidt 1996)
	// with a volatile storage member to minimise race conditions due
	// to the so-called "relaxed memory model"..
	//==================================================================
	if(!s_pIn)
	{
		OT_AUTO_LOCK(FastMutex, ConsoleMutex);
		// test again now that mutex has been locked
		if(!s_pIn)
		{
			s_pIn = new InputStreamReader(
				new FileInputStream(FileDescriptor::In().get()));
			System::GetObjectManager().registerObject(s_pIn);
		}
	}

	return s_pIn;
}

OT_IO_NAMESPACE_END
