//==============================================================================
// This file is part of the OpenTop C++ Library.
// (c) Copyright 2000-2002 ElCel Technology Limited
//
// This file may be distributed and/or modified under the terms of the
// GNU General Public License version 2 as published by the Free Software
// Foundation and appearing in the file LICENSE.GPL included in the
// packaging of this file.
//
// Licensees holding valid OpenTop Commercial licences may use this file
// in accordance with the OpenTop Commercial License Agreement included
// with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// See http://www.elcel.com/products/opentop/pricing.html or email
// sales@elcel.com for information about OpenTop Commercial License Agreements.
//
//==============================================================================
//
// $Revision: 1.1.1.1 $
// $Date: 2002/12/14 00:25:13 $
//
//==============================================================================
//
// Class: FileDescriptor
/**
	@class ot::io::FileDescriptor
	
	@brief An abstract base class representing open file handles.

    It @a derives from ManagedObject and uses the standard @OpenTop reference-counting
	scheme.  By using ResourceDescriptors, @OpenTop is able to share file handles
	between multiple objects, preventing the file from being closed until
	all references to the FileDescriptor have been removed.

    A FileDescriptor is associated with a FileSystem object when it is created.
	When the file needs to be closed, the FileSystem associated with the file
	performs the action.
*/
//==============================================================================

#include "FileDescriptor.h"
#include "FileSystem.h"

#include "ot/base/NullPointerException.h"

OT_IO_NAMESPACE_BEGIN

//==============================================================================
// FileDescriptor::FileDescriptor
//
/**
   Constructs a FileDescriptor.

   @param pFS the FileSystem associated with the open file designated by this
          FileDescriptor.
   @param bAutoClose true if the file should be closed when this FileDescriptor
          is destroyed
   @throws NullPointerException if @c pFS is null
*/
//==============================================================================
FileDescriptor::FileDescriptor(const FileSystem* pFS, bool bAutoClose) :
	ResourceDescriptor(bAutoClose),
	m_rpFS(const_cast<FileSystem*>(pFS))
{
	if(!pFS) throw NullPointerException();
}

//==============================================================================
// FileDescriptor::~FileDescriptor
//
/**
   Destructor.  This method does nothing, but @a derived classes are 
   expected to call getAutoClose() and close the related resource if
   it returns @c true.
   
   To indicate that they have performed the required task,
   all @a derived class destructors should call setAutoClose(false)
   before exiting.
*/
//==============================================================================
FileDescriptor::~FileDescriptor()
{
}

//==============================================================================
// FileDescriptor::close
//
/**
   Closes the file.
*/
//==============================================================================
void FileDescriptor::close()
{
	m_rpFS->closeFile(this);
	setAutoClose(false);
}

//==============================================================================
// FileDescriptor::getFileSystem
//
/**
   Returns the FileSystem associated with this FileDescriptor.
*/
//==============================================================================
RefPtr<FileSystem> FileDescriptor::getFileSystem() const
{
	return m_rpFS;
}

//==============================================================================
// FileDescriptor::Err
//
/**
   Returns a FileDescriptor for the standard Console stream @c 'stderr'.

   FileDescriptors for standard Console streams are not auto-close descriptors
   which means that the stream is (correctly) left open when then the
   returned FileDescriptor is destroyed.
*/
//==============================================================================
RefPtr<FileDescriptor> FileDescriptor::Err()
{
	return FileSystem::GetFileSystem()->getConsoleFD(FileSystem::Stderr);
}

//==============================================================================
// FileDescriptor::Out
//
/**
   Returns a FileDescriptor for the standard Console stream @c 'stdout'.

   FileDescriptors for standard Console streams are not auto-close descriptors
   which means that the stream is (correctly) left open when then the
   returned FileDescriptor is destroyed.
*/
//==============================================================================
RefPtr<FileDescriptor> FileDescriptor::Out()
{
	return FileSystem::GetFileSystem()->getConsoleFD(FileSystem::Stdout);
}

//==============================================================================
// FileDescriptor::In
//
/**
   Returns a FileDescriptor for the standard Console stream @c 'stdin'.

   FileDescriptors for standard Console streams are not auto-close descriptors
   which means that the stream is (correctly) left open when then the
   returned FileDescriptor is destroyed.
*/
//==============================================================================
RefPtr<FileDescriptor> FileDescriptor::In()
{
	return FileSystem::GetFileSystem()->getConsoleFD(FileSystem::Stdin);
}

OT_IO_NAMESPACE_END
