//==============================================================================
// This file is part of the OpenTop C++ Library.
// (c) Copyright 2000-2002 ElCel Technology Limited
//
// This file may be distributed and/or modified under the terms of the
// GNU General Public License version 2 as published by the Free Software
// Foundation and appearing in the file LICENSE.GPL included in the
// packaging of this file.
//
// Licensees holding valid OpenTop Commercial licences may use this file
// in accordance with the OpenTop Commercial License Agreement included
// with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// See http://www.elcel.com/products/opentop/pricing.html or email
// sales@elcel.com for information about OpenTop Commercial License Agreements.
//
//==============================================================================
//
// $Revision: 1.1.1.1 $
// $Date: 2002/12/14 00:25:15 $
//
//==============================================================================
//
// Class: FilterInputStream
/**
	@class ot::io::FilterInputStream
	
	@brief A base class that @a derives from InputStream but also
	@a contains another InputStream which is used as the input source.
	
    FilterInputStream and its sister class FilterOutputStream provide
	an extensible framework for building chains of processing.

    The FilterInputStream class overrides all methods of InputStream with 
	versions that pass requests to the contained input stream.
	Derived classes of FilterInputStream are expected to further override
	some of these methods to perform some useful function before the data
	is returned to the caller (which may be yet another FilterInputStream).

    @note Derived classes that override the any of the read() methods
	would be advised to override all the read() methods in addition to the
	skip() method.  Failure to do so may produce unexpected results as
	these methods would otherwise be delegated to the contained InputStream - 
	perhaps bypassing the required logic.
*/
//==============================================================================

#include "FilterInputStream.h"

#include "ot/base/NullPointerException.h"

OT_IO_NAMESPACE_BEGIN

//==============================================================================
// FilterInputStream::FilterInputStream
//
/**
   Constructs a FilterInputStream using @c pInputStream as the contained
   input stream.

   This constructor is protected to prevent the accidental
   creation of instances of FilterInputStream.  It should
   only be necessary to create instances of @a derived classes
   which have been customized to perform some useful behaviour.

   @throws NullPointerException if @c pInputStream is null.
*/
//==============================================================================
FilterInputStream::FilterInputStream(InputStream* pInputStream) :
	m_rpInputStream(pInputStream)
{
	if(!pInputStream) throw NullPointerException();
}

void FilterInputStream::mark(size_t readLimit)
{
	m_rpInputStream->mark(readLimit);
}

bool FilterInputStream::markSupported() const
{
	return m_rpInputStream->markSupported();
}

void FilterInputStream::reset()
{
	m_rpInputStream->reset();
}

size_t FilterInputStream::available()
{
	return m_rpInputStream->available();
}

void FilterInputStream::close()
{
	m_rpInputStream->close();
}

int FilterInputStream::read()
{
	return m_rpInputStream->read();
}

long FilterInputStream::read(Byte* pBuffer, size_t bufLen)
{
	return m_rpInputStream->read(pBuffer, bufLen);
}

size_t FilterInputStream::skip(size_t n)
{
	return m_rpInputStream->skip(n);
}

//==============================================================================
// FilterInputStream::getInputStream
//
/**
   Returns the contained input stream.
*/
//==============================================================================
RefPtr<InputStream> FilterInputStream::getInputStream() const
{
	return m_rpInputStream;
}

OT_IO_NAMESPACE_END
