//==============================================================================
// This file is part of the OpenTop C++ Library.
// (c) Copyright 2000-2002 ElCel Technology Limited
//
// This file may be distributed and/or modified under the terms of the
// GNU General Public License version 2 as published by the Free Software
// Foundation and appearing in the file LICENSE.GPL included in the
// packaging of this file.
//
// Licensees holding valid OpenTop Commercial licences may use this file
// in accordance with the OpenTop Commercial License Agreement included
// with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// See http://www.elcel.com/products/opentop/pricing.html or email
// sales@elcel.com for information about OpenTop Commercial License Agreements.
//
//==============================================================================
//
// $Revision: 1.1.1.1 $
// $Date: 2002/12/14 00:25:16 $
//
//==============================================================================
//
// Class: FilterReader
/**
	@class ot::io::FilterReader
	
	@brief A base class that @a derives from Reader but also
	@a contains another Reader which is used as the input source.
	
    FilterReader and its sister class FilterWriter provide
	an extensible framework for building chains of character processing.

    The FilterReader class overrides all methods of Reader with 
	versions that pass requests to the contained Reader.
	Derived classes of FilterReader are expected to further override
	some of these methods to perform some useful function before the characters
	are returned to the caller (which may be yet another FilterReader).

    @note Derived classes that override the any of the read() or readAtomic()
	methods would be advised to override all the read() and readAtomic()
	methods in addition to the skip() and skipAtomic() methods.  Failure to do
	so may produce unexpected results as
	these methods would otherwise be delegated to the contained Reader - 
	perhaps bypassing the required logic.

    @mt
    The contained Reader is used as the lock object for synchronized methods.
*/
//==============================================================================

#include "FilterReader.h"

#include "ot/base/NullPointerException.h"

OT_IO_NAMESPACE_BEGIN

//==============================================================================
// FilterReader::FilterReader
//
/**
   Constructs a FilterReader using @c pReader as the contained Reader.

   @throws NullPointerException if @c pReader is null.
*/
//==============================================================================
FilterReader::FilterReader(Reader* pReader) :
	m_rpReader(pReader)
{
	if(!pReader) throw NullPointerException();
	m_rpLock = pReader->getLock();
}

//==============================================================================
// FilterReader::close
//
//==============================================================================
void FilterReader::close()
{
	m_rpReader->close();
}

//==============================================================================
// FilterReader::mark
//
//==============================================================================
void FilterReader::mark(size_t readLimit)
{
	m_rpReader->mark(readLimit);
}
     
//==============================================================================
// FilterReader::markSupported
//
/**
   Tests whether the Reader supports the mark() operation.
   
   The FilterReader class contains an implementation that passes the
   request to the contained Reader.

   @sa mark()
   @sa reset()
   @returns true if the contained Reader supports the mark() operation;
            false otherwise
*/
//==============================================================================
bool FilterReader::markSupported() const
{
	return m_rpReader->markSupported();
}

//==============================================================================
// FilterReader::read
//
//==============================================================================
IntType FilterReader::read()
{
	return m_rpReader->read();
}

//==============================================================================
// FilterReader::read
//
//==============================================================================
long FilterReader::read(CharType* pBuffer, size_t bufLen)
{
	return m_rpReader->read(pBuffer, bufLen);
}

//==============================================================================
// FilterReader::readAtomic
//
//==============================================================================
long FilterReader::readAtomic(CharType* pBuffer, size_t bufLen)
{
	return m_rpReader->readAtomic(pBuffer, bufLen);
}

Character FilterReader::readAtomic()
{
	return m_rpReader->readAtomic();
}

//==============================================================================
// FilterReader::reset
//
//==============================================================================
void FilterReader::reset()
{
	m_rpReader->reset();
}

//==============================================================================
// FilterReader::skip
//
//==============================================================================
size_t FilterReader::skip(size_t n)
{
	return m_rpReader->skip(n);
}

//==============================================================================
// FilterReader::skipAtomic
//
//==============================================================================
size_t FilterReader::skipAtomic(size_t n)
{
	return m_rpReader->skipAtomic(n);
}

//==============================================================================
// FilterReader::getReader
//
/**
   Returns a reference to the contained Reader.
*/
//==============================================================================
RefPtr<Reader> FilterReader::getReader() const
{
	return m_rpReader;
}

OT_IO_NAMESPACE_END
