//==============================================================================
// This file is part of the OpenTop C++ Library.
// (c) Copyright 2000-2002 ElCel Technology Limited
//
// This file may be distributed and/or modified under the terms of the
// GNU General Public License version 2 as published by the Free Software
// Foundation and appearing in the file LICENSE.GPL included in the
// packaging of this file.
//
// Licensees holding valid OpenTop Commercial licences may use this file
// in accordance with the OpenTop Commercial License Agreement included
// with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// See http://www.elcel.com/products/opentop/pricing.html or email
// sales@elcel.com for information about OpenTop Commercial License Agreements.
//
//==============================================================================
//
// $Revision: 1.1.1.1 $
// $Date: 2002/12/14 00:25:16 $
//
//==============================================================================
//
// Class: FilterWriter
// 
/**
	@class ot::io::FilterWriter
	
	@brief A base class that @a derives from Writer but also
	@a contains another Writer which is used as the output sink.
	
    FilterWriter and its sister class FilterReader provide
	an extensible framework for building chains of character processing.

    The FilterWriter class overrides all methods of Writer with 
	versions that pass requests to the contained Writer.
	Derived classes of FilterWriter are expected to further override
	some of these methods to perform some useful function before the characters
	are written out to the sink (which may be yet another FilterWriter).

    @mt
    The contained Writer is used as the lock object for synchronized methods.
*/
//==============================================================================

#include "FilterWriter.h"

#include "ot/base/NullPointerException.h"

OT_IO_NAMESPACE_BEGIN

//==============================================================================
// FilterWriter::FilterWriter
//
/**
   Construct a FilterWriter with @c pWriter as the contained Writer.

   @throws NullPointerException if pWriter is null.
*/
//==============================================================================
FilterWriter::FilterWriter(Writer* pWriter) :
	m_rpWriter(pWriter)
{
	if(!pWriter) throw NullPointerException();
	m_rpLock = pWriter->getLock();
}
	
void FilterWriter::close()
{
	m_rpWriter->close();
}

void FilterWriter::flush()
{
	m_rpWriter->flush();
}

void FilterWriter::flushBuffers()
{
	m_rpWriter->flushBuffers();
}

void FilterWriter::write(const CharType* pStr, size_t len)
{
	m_rpWriter->write(pStr, len);
}

void FilterWriter::write(CharType c)
{
	m_rpWriter->write(c);
}

void FilterWriter::write(const Character& ch)
{
	m_rpWriter->write(ch);
}

void FilterWriter::write(const String& str)
{
	m_rpWriter->write(str);
}

//==============================================================================
// FilterWriter::getWriter
//
/**
   Returns the Writer contained by this FilterWriter.
*/
//==============================================================================
RefPtr<Writer> FilterWriter::getWriter() const
{
	return m_rpWriter;
}

OT_IO_NAMESPACE_END
