//==============================================================================
// This file is part of the OpenTop C++ Library.
// (c) Copyright 2000-2002 ElCel Technology Limited
//
// This file may be distributed and/or modified under the terms of the
// GNU General Public License version 2 as published by the Free Software
// Foundation and appearing in the file LICENSE.GPL included in the
// packaging of this file.
//
// Licensees holding valid OpenTop Commercial licences may use this file
// in accordance with the OpenTop Commercial License Agreement included
// with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// See http://www.elcel.com/products/opentop/pricing.html or email
// sales@elcel.com for information about OpenTop Commercial License Agreements.
//
//==============================================================================
//
// Class OutputStream
/**
	@class ot::io::OutputStream
	
	@brief Abstract base class with the ability to write bytes to a data sink.

    Concrete classes derived from OutputStream may be used 
    to write bytes to a wide range of data sinks including the console, a file
    or a network socket connection.
    
    Unlike the standard C++ ostream, this class does not concern itself with
    formatting, rather it limits itself to the writing out of data.  In this respect
    it is closer to the standard streambuf class - but does not have the
    added complication of supporting a read capability.

    OutputStreams only concern themselves with the writing of bytes - not characters.
    If the application uses characters (as most do) then an OutputStream can be
    wrapped by an OutputStreamWriter that knows how the OutputStream should be
    encoded.
*/
//==============================================================================

#include "OutputStream.h"

OT_IO_NAMESPACE_BEGIN

//==============================================================================
// OutputStream::close
//
/**
   Closes the output stream and releases system resources associated with
   the stream.  

   Once an OutputStream is closed further calls to write(), flush() or
   flushBuffers() will result in an IOException being thrown. Further calls
   to close() are legal but have no effect.
   
   @throws IOException if an I/O error occurs.
*/
//==============================================================================
void OutputStream::close()
{
}

//==============================================================================
// OutputStream::flush
//
/**
   Forces all buffered data to be written out and sent to the final data sink.

   @sa flushBuffers()
   @throws IOException if an I/O error occurs.
*/
//==============================================================================
void OutputStream::flush()
{
}

//==============================================================================
// OutputStream::flushBuffers
//
/**
   Forces all data buffers associated with this output stream to be written out.
   This may not cause the data to be sent to the final sink if the underlying
   sink maintains its own buffering scheme.

   @sa flush()
   @throws IOException if an I/O error occurs.
*/
//==============================================================================
void OutputStream::flushBuffers()
{
}

//==============================================================================
// OutputStream::write
//
/**
   Writes the single byte @c x to this output stream.
   @param x the byte to write
   @throws IOException if an I/O error occurs.
*/
//==============================================================================
void OutputStream::write(Byte x)
{
	write(&x, 1);
}

#ifdef OT_DOCUMENTATION_ONLY
//=============================================================================
//
// Documentation for pure virtual methods follows:
//
//=============================================================================

//==============================================================================
// OutputStream::write
//
/**
   Writes an array of bytes to this output stream.

   @param pBuffer pointer to the start of an array of bytes to be written
   @param bufLen length of the byte array
   @throws NullPointerException if @c pBuffer is null.
   @throws IOException if an I/O error occurs.
*/
//==============================================================================
void OutputStream::write(const Byte* pBuffer, size_t bufLen);

#endif  //OT_DOCUMENTATION_ONLY

OT_IO_NAMESPACE_END
