//==============================================================================
// This file is part of the OpenTop C++ Library.
// (c) Copyright 2000-2002 ElCel Technology Limited
//
// This file may be distributed and/or modified under the terms of the
// GNU General Public License version 2 as published by the Free Software
// Foundation and appearing in the file LICENSE.GPL included in the
// packaging of this file.
//
// Licensees holding valid OpenTop Commercial licences may use this file
// in accordance with the OpenTop Commercial License Agreement included
// with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// See http://www.elcel.com/products/opentop/pricing.html or email
// sales@elcel.com for information about OpenTop Commercial License Agreements.
//
//==============================================================================
//
// $Revision: 1.3 $
// $Date: 2003/08/10 22:05:56 $
//
//==============================================================================
//
// Class: ResourceDescriptor
/**
	@class ot::io::ResourceDescriptor
	
	@brief An abstract base class used to connect @OpenTop with operating system
	resources such as open files and network sockets.

	A ResourceDescriptor represents an open handle.  It @a derives from
	ManagedObject and uses the standard @OpenTop reference-counting scheme.  By using
	ResourceDescriptors, @OpenTop is able to share handles between multiple
	objects, preventing the underlying resource from being closed until
	all references to the ResourceDescriptor have been destroyed.  This
	is often more efficient and flexible than duplicating operating system 
	handles for the same purpose.

	ResourceDescriptor declares a private copy constructor and assignment
	operator to prevent the copying of instance objects.
*/
//==============================================================================

#include "ResourceDescriptor.h"

OT_IO_NAMESPACE_BEGIN

//==============================================================================
// ResourceDescriptor::ResourceDescriptor
//
/**
   Constructs a ResourceDescriptor, specifying whether the related operating
   system resource should be closed when this ResourceDescriptor is destroyed.

   @param bAutoClose true if the resource should be closed when this
          ResourceDescriptor is destroyed; false otherwise
*/
//==============================================================================
ResourceDescriptor::ResourceDescriptor(bool bAutoClose) :
	m_bAutoClose(bAutoClose)
{
}

//==============================================================================
// ResourceDescriptor::~ResourceDescriptor
//
/**
   Destructor.  This method does nothing but @a derived classes are 
   expected to call getAutoClose() and close the related resource if
   it returns @c true.
   
   To indicate that they have performed the required task,
   all @a derived class destructors should call setAutoClose(false)
   before exiting.
*/
//==============================================================================
ResourceDescriptor::~ResourceDescriptor()
{
	// Derived classes should set m_bAutoClose to false in their
	// destructors.  Failure to do so implies that they have not
	// honoured the committment to close the managed resource.
	OT_DBG_ASSERT(!m_bAutoClose);
}

//==============================================================================
// ResourceDescriptor::getAutoClose
//
/**
   Returns a boolean value to indicate whether the underlying resource will
   be closed when the reference count for this ResourceDescriptor reaches zero.

   @returns true if the resource will be closed automatically; false otherwise
*/
//==============================================================================
bool ResourceDescriptor::getAutoClose() const
{
	return m_bAutoClose;
}

//==============================================================================
// ResourceDescriptor::setAutoClose
//
/**
   Sets the value of the auto close indicator.

   @sa getAutoClose()
*/
//==============================================================================
void ResourceDescriptor::setAutoClose(bool bAutoClose)
{
	m_bAutoClose = bAutoClose;
}

#ifdef OT_DOCUMENTATION_ONLY

//==============================================================================
// ResourceDescriptor::toString
//
/**
   Returns a string representation of this ResourceDescriptor.  
   
   @OpenTop uses the string representation to identify the resource descriptor
   when tracing via the Tracer class.
   @since 1.3
*/
//==============================================================================
String ResourceDescriptor::toString() const;

#endif //OT_DOCUMENTATION_ONLY

OT_IO_NAMESPACE_END
