//==============================================================================
// This file is part of the OpenTop C++ Library.
// (c) Copyright 2000-2002 ElCel Technology Limited
//
// This file may be distributed and/or modified under the terms of the
// GNU General Public License version 2 as published by the Free Software
// Foundation and appearing in the file LICENSE.GPL included in the
// packaging of this file.
//
// Licensees holding valid OpenTop Commercial licences may use this file
// in accordance with the OpenTop Commercial License Agreement included
// with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// See http://www.elcel.com/products/opentop/pricing.html or email
// sales@elcel.com for information about OpenTop Commercial License Agreements.
//
//==============================================================================
//
// $Revision: 1.1.1.1 $
// $Date: 2002/12/14 00:25:22 $
//
//==============================================================================
//
// Class: StringReader
/**
	@class ot::io::StringReader
	
	@brief A Reader which uses a String as its input source.
	@sa StringWriter
*/
//==============================================================================

#include "StringReader.h"
#include "AtomicReadException.h"

#include "ot/base/NullPointerException.h"
#include "ot/base/IllegalCharacterException.h"
#include "ot/base/SystemUtils.h"

OT_IO_NAMESPACE_BEGIN

//==============================================================================
// StringReader::StringReader
//
/**
   Constructs a StringReader with a copy of @c string as the contained String.
*/
//==============================================================================
StringReader::StringReader(const String& string) :
	m_string(string),
	m_pos(0),
	m_markPos(-1),
	m_bClosed(false)
{
}

//==============================================================================
// StringReader::StringReader
//
/**
   Constructs a StringReader, copying the ::CharType array beginning
   at @c pStr into a String which is used as the input source.

   Unicode characters represented by the ::CharType array should already be
   encoded using the internal @OpenTop encoding scheme.

   @param pStr pointer to the start of the ::CharType array to copy
   @param length of the ::CharType array
*/
//==============================================================================
StringReader::StringReader(const CharType* pStr, size_t length) :
	m_pos(0),
	m_markPos(-1),
	m_bClosed(false)
{
	if(length)
	{
		if(!pStr) throw NullPointerException();
		m_string.assign(pStr, length);
	}
}

//==============================================================================
// StringReader::close
//
//==============================================================================
void StringReader::close()
{
	OT_SYNCHRONIZED_PTR(m_rpLock)

	m_markPos = -1;
	m_bClosed = true;
}

//==============================================================================
// StringReader::mark
//
// Marks the current position in this reader. A subsequent call to the 
// reset method re-positions the reader at the last marked position so that 
// subsequent reads re-read the same characters. 
//
// The readLimit has no effect for StringReaders, as the entire string
// is always available until the Reader is closed. 
//==============================================================================
void StringReader::mark(size_t /*readLimit*/)
{
	OT_SYNCHRONIZED_PTR(m_rpLock)

	if(m_bClosed) throw IOException(OT_T("stream is closed"));
	m_markPos = m_pos;
}

//==============================================================================
// StringReader::reset
//
// Reset the position to the last mark() operation.  If there is no mark 
// (due to no mark() call then throw an IOException.
//==============================================================================
void StringReader::reset()
{
	OT_SYNCHRONIZED_PTR(m_rpLock)

	if(m_markPos == -1)
	{
		throw IOException(OT_T("unable to reset input stream, no marked position"));
	}
	else
	{
		m_pos = m_markPos;
	}
}

//==============================================================================
// StringReader::markSupported
//
/**
   Tests whether the Reader supports the mark() operation, which it does.
   
   @sa mark()
   @sa reset()
   @returns always returns true for StringReader.
   @synchronized
*/
//==============================================================================
bool StringReader::markSupported() const
{
	return true;
}

//==============================================================================
// StringReader::read
//
//==============================================================================
long StringReader::read(CharType* pBuffer, size_t bufLen)
{
	SystemUtils::TestBufferIsValid(pBuffer, bufLen);
	
	OT_SYNCHRONIZED_PTR(m_rpLock)

	if(m_bClosed) throw IOException(OT_T("stream is closed"));

	const size_t charsAvailable = (m_string.length() - m_pos);

	if(charsAvailable)
	{
		const size_t count = (bufLen < charsAvailable)
		                   ? bufLen
		                   : charsAvailable;

		::memcpy(pBuffer, m_string.data()+m_pos, count*sizeof(CharType));
		m_pos += count;
		return count;
	}
	else
	{
		return EndOfFile;
	}
}

//==============================================================================
// StringReader::readAtomic
//
//==============================================================================
long StringReader::readAtomic(CharType* pBuffer, size_t bufLen)
{
	SystemUtils::TestBufferIsValid(pBuffer, bufLen);

	OT_SYNCHRONIZED_PTR(m_rpLock)

	if(m_bClosed) throw IOException(OT_T("stream is closed"));

	const size_t stringSize = m_string.size();
	
	if(m_pos < stringSize)
	{
		const CharType* pData = m_string.data();

		if(!SystemCodeConverter::IsSequenceStartChar(pData[m_pos]))
		{
			throw AtomicReadException(OT_T("not on character sequence boundary"));
		}

		//
		// If we can slip our entire buffer into the passed buffer then 
		// that's great because our buffer *should* hold an integral number of
		// unicode characters.  But if the passed buffer is too small, we have
		// a choice: either
		// i)  scan the buffer counting characters; or
		// ii) search back for a sequence start character
		// All internal encodings are guaranteed to support the recognition of
		// sequence start characters, so (ii) will be more efficient for any buffer
		// of a significant size. 
		//
		size_t charCount = stringSize - m_pos;
		if(charCount > bufLen)
		{
			charCount = bufLen;
			do {--charCount;}
			while(charCount && !SystemCodeConverter::IsSequenceStartChar(*(pData+m_pos+charCount)));

			size_t charSeqLen = SystemCodeConverter::GetCharSequenceLength(*(pData+m_pos+charCount));
			if((charCount + charSeqLen) <= bufLen)
			{
				charCount += charSeqLen;
			}
		}

		//
		// It's quite possible that we are unable to read any characters because
		// there is insufficient buffer space to read an integral number of 
		// characters.  This is legal.
		//
		if(charCount)
		{
			::memcpy(pBuffer, pData+m_pos, charCount*sizeof(CharType));
			m_pos += charCount;
		}
		return charCount;
	}
	else
	{
		return EndOfFile;
	}
}

//==============================================================================
// StringReader::readAtomic
//
//==============================================================================
Character StringReader::readAtomic()
{
	OT_SYNCHRONIZED_PTR(m_rpLock)

	if(m_bClosed) throw IOException(OT_T("stream is closed"));
	
	if(m_pos < m_string.length())
	{
		if(!SystemCodeConverter::IsSequenceStartChar(m_string[m_pos]))
		{
			throw AtomicReadException(OT_T("not on character sequence boundary"));
		}

		const CharType nextChar = m_string[m_pos];
		size_t seqLen = SystemCodeConverter::GetCharSequenceLength(nextChar);

		if((m_pos+seqLen) <= m_string.length())
		{
			size_t oldOffset = m_pos;
			m_pos += seqLen;
			return Character(m_string.data() + oldOffset, seqLen);
		}
		else
		{
			throw IllegalCharacterException(OT_T("unterminated character sequence in String"));
		}
	}
	return Character::EndOfFileCharacter;
}

OT_IO_NAMESPACE_END
