//==============================================================================
// This file is part of the OpenTop C++ Library.
// (c) Copyright 2000-2002 ElCel Technology Limited
//
// This file may be distributed and/or modified under the terms of the
// GNU General Public License version 2 as published by the Free Software
// Foundation and appearing in the file LICENSE.GPL included in the
// packaging of this file.
//
// Licensees holding valid OpenTop Commercial licences may use this file
// in accordance with the OpenTop Commercial License Agreement included
// with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// See http://www.elcel.com/products/opentop/pricing.html or email
// sales@elcel.com for information about OpenTop Commercial License Agreements.
//
//==============================================================================
//
// $Revision: 1.1.1.1 $
// $Date: 2002/12/14 00:25:22 $
//
//==============================================================================
//
// Class: StringWriter
/**
	@class ot::io::StringWriter
	
	@brief A Writer which collects output characters in a buffer, which can then
	be used to create a String.
	@sa StringReader
*/
//==============================================================================

#include "StringWriter.h"
#include "IOException.h"

OT_IO_NAMESPACE_BEGIN

//==============================================================================
// StringWriter::StringWriter
//
/**
   Constructs a StringWriter with no initial size.
*/
//==============================================================================
StringWriter::StringWriter() :
	m_bClosed(false)
{
}

//==============================================================================
// StringWriter::StringWriter
//
/**
   Constructs a StringWriter with an initial size of @c initialSize.
*/
//==============================================================================
StringWriter::StringWriter(size_t initialSize) :
	m_buffer(initialSize),
	m_bClosed(false)
{
}

//==============================================================================
// StringWriter::close
//
/**
   Closes the StringWriter.

   The internal string buffer is not released so that it remains available for
   subsequent calls to toString().

   @synchronized
*/
//==============================================================================
void StringWriter::close()
{
	OT_SYNCHRONIZED_PTR(m_rpLock)

	m_bClosed = true;
}

//==============================================================================
// StringWriter::write
//
/**
   Writes the character buffer @c pBuf with a length of @c len

   @throws IOException if the StringWriter has been closed
   @synchronized
*/
//==============================================================================
void StringWriter::write(const CharType* pBuf, size_t len)
{
	OT_SYNCHRONIZED_PTR(m_rpLock)

	if(m_bClosed) throw IOException(OT_T("cannot write to a closed stream"));

	m_buffer.append(pBuf, len);
}

//==============================================================================
// StringWriter::toString
//
/**
   Creates a string from the internal character buffer.

   @returns A String containing all the characters that have been written to the
            StringWriter so far.
   @synchronized
*/
//==============================================================================
String StringWriter::toString() const
{
	OT_SYNCHRONIZED_PTR(m_rpLock)

	return String(m_buffer.data(), m_buffer.size());
}

OT_IO_NAMESPACE_END
