//==============================================================================
// This file is part of the OpenTop C++ Library.
// (c) Copyright 2000-2002 ElCel Technology Limited
//
// This file may be distributed and/or modified under the terms of the
// GNU General Public License version 2 as published by the Free Software
// Foundation and appearing in the file LICENSE.GPL included in the
// packaging of this file.
//
// Licensees holding valid OpenTop Commercial licences may use this file
// in accordance with the OpenTop Commercial License Agreement included
// with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// See http://www.elcel.com/products/opentop/pricing.html or email
// sales@elcel.com for information about OpenTop Commercial License Agreements.
//
//==============================================================================
//
// $Revision: 1.1.1.1 $
// $Date: 2002/12/14 00:25:23 $
//
//==============================================================================
//
// Class Writer
/**
	@class ot::io::Writer
	
	@brief Abstract base class representing an output stream with the ability
	to write Unicode characters to a sink.

 	Concrete sub-classes of Writer may be used to translate Unicode characters
	into byte sequences which are then written to an OutputStream.  In this case,
	an instance of CodeConverter may be used to translate between Unicode
	characters and bytes.

	Depending on the configuration chosen, Unicode characters may be represented
	internally using 32-bit UCS-4 characters, 16-bit UTF-16 characters or
	8-bit UTF-8 characters.  See @encoding for further information.

	The Writer interface mirrors OutputStream, but deals in characters where
	OutputStream deals in bytes.

    @mt
    As shown on the inheritance graph, Writer @a derives from 
	SynchronizedObject, which gives it the ability to protect its internal
	state from concurrent access from multiple threads.  All public methods 
	are synchronized for safe concurrent access.
*/
//==============================================================================

#include "Writer.h"

#include "ot/base/NullPointerException.h"

OT_IO_NAMESPACE_BEGIN

//==============================================================================
// Writer::Writer
//
/**
   Default constructor.
*/
//==============================================================================
Writer::Writer() :
	m_rpLock(this, this)
{
}

//==============================================================================
// Writer::Writer
//
/**
   Constructor taking a SynchronizedObject which will be used as the lock
   for synchronized functions.

   @throws NullPointerException if @c pLockObject is null
*/
//==============================================================================
Writer::Writer(SynchronizedObject* pLockObject) :
	m_rpLock(this, pLockObject)
{
	if(!pLockObject) throw NullPointerException();
}

//==============================================================================
// Writer::write
//
/**
   Writes the single ::CharType character @c c.

   Depending on the character encoding being employed, a single
   ::CharType character may not represent a complete Unicode character.  In this
   case the characters from subsequent write operations will be used to complete
   the multi-character sequence.

   @param c the character to write.
   @throws IOException if an I/O error occurs.
   @sa write(const Character& ch)
   @synchronized
*/
//==============================================================================
void Writer::write(CharType c)
{
	write(&c, 1);
}

//==============================================================================
// Writer::write
//
/**
   Writes the Unicode character @c ch to the output stream.

   Depending on the character encoding being employed, a single Unicode
   character may be represented internally using a sequence of more than one
   ::CharType characters.  In this case, as many ::CharType characters as are required
   to encode the Unicode character are written.

   @param ch the Unicode character to write.
   @throws IOException if an I/O error occurs.
   @sa write(CharType c)
   @synchronized
*/
//==============================================================================
void Writer::write(const Character& ch)
{
	write(ch.data(), ch.length());
}

//==============================================================================
// Writer::write
//
/**
   Writes the sequence of ::CharType characters contained in the String @c str.

   @param str the String to write.
   @throws IOException if an I/O error occurs.
   @synchronized
*/
//==============================================================================
void Writer::write(const String& str)
{
	if(!str.empty())
	{
		write(str.data(), str.length());
	}
}

//==============================================================================
// Writer::flush
//
/**
   Flushes any output buffers before forcing the output to its final
   destination.

   @throws IOException if an I/O error occurs.
   @sa flushBuffers()
   @synchronized
*/
//==============================================================================
void Writer::flush()
{
}

//==============================================================================
// Writer::flushBuffers
//
/**
   If this Writer maintains an output buffer, the buffer is emptied and written
   to the output destination without requesting the final destination
   to flush the output.

   The base class implementation does nothing.

   @sa flush()
   @synchronized
*/
//==============================================================================
void Writer::flushBuffers()
{
}

//==============================================================================
// Writer::getLock
//
/**
   Returns a reference to the SynchronizedObject used for controlling
   access to synchronized methods.
*/
//==============================================================================
RefPtr<SynchronizedObject> Writer::getLock() const
{
	// Note: As we are returning a RefPtr, it is imperative that
	// our reference count is not zero if the lock is pointing
	// at this object
	SynchronizedObject* pLock = m_rpLock;
	OT_DBG_ASSERT((Writer*) pLock != this || getRefCount());
	return pLock;
}

#ifdef OT_DOCUMENTATION_ONLY
//=============================================================================
//
// Documentation for pure virtual methods follows:
//
//=============================================================================

//==============================================================================
// Writer::close
//
/**
   Closes the character stream, flushing it first.
   @synchronized
*/
//==============================================================================
void Writer::close();

//==============================================================================
// Writer::write
//
/**
   Writes an array of ::CharType characters.

   @param pStr a pointer to the first ::CharType character in the array
   @param len the number of ::CharType characters to write
   @throws IOException if an I/O error occurs.
   @synchronized
*/
//==============================================================================
void Writer::write(const CharType* pStr, size_t len);

#endif //OT_DOCUMENTATION_ONLY

OT_IO_NAMESPACE_END
