//==============================================================================
// This file is part of the OpenTop C++ Library.
// (c) Copyright 2000-2002 ElCel Technology Limited
//
// This file may be distributed and/or modified under the terms of the
// GNU General Public License version 2 as published by the Free Software
// Foundation and appearing in the file LICENSE.GPL included in the
// packaging of this file.
//
// Licensees holding valid OpenTop Commercial licences may use this file
// in accordance with the OpenTop Commercial License Agreement included
// with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// See http://www.elcel.com/products/opentop/pricing.html or email
// sales@elcel.com for information about OpenTop Commercial License Agreements.
//
//==============================================================================
//
// $Revision: 1.2 $
// $Date: 2003/03/16 22:51:58 $
//
//==============================================================================
// Class BasicHttpURLConnection
/**
	@class ot::net::BasicHttpURLConnection
	
	@brief An @a specialized URLConnection class that represents a connection
	with a resource specified by a HTTP URL.

	<p>HTTP stands for Hyper Text Transfer Protocol, which is defined by
	RFCs published by the <a href="http://www.ietf.org/">Internet Engineering
	Task Force</a>.</p>

    @sa URLConnection
*/
//==============================================================================

#include "BasicHttpURLConnection.h"
#include "ProtocolException.h"

#include "ot/base/System.h"
#include "ot/base/Tracer.h"
#include "ot/base/NumUtils.h"
#include "ot/base/StringUtils.h"
#include "ot/base/debug.h"

OT_NET_NAMESPACE_BEGIN

const size_t HTTP_TIMEOUT = 30000;  // 30 second connection timeout

//==============================================================================
// BasicHttpURLConnection::BasicHttpURLConnection
//
// Construct a BasicHttpURLConnection using the passed URL and default
// (or global) settings.
//==============================================================================
BasicHttpURLConnection::BasicHttpURLConnection(const URL& url) :
	HttpURLConnection(url),
	m_rpHttp(new HttpClient)
{
	//
	// If redirects have been disabled globally, then disable them
	// for this instance
	//
	if(!GetDefaultFollowRedirects())
	{
		m_rpHttp->setFollowRedirects(false);
	}
}

//=============================================================================
// BasicHttpURLConnection::connect
// 
// Creates a socket connection to the remote host and sends the HTTP
// request headers etc.
// 
// The HttpClient will follow redirect requests if they have been enabled
// (which they are by default).
//=============================================================================
void BasicHttpURLConnection::connect()
{
	if(!isConnected())
	{
		//
		// Connect to the server (if not already connected)
		//
		m_rpHttp->connectToURL(getURL(), HTTP_TIMEOUT);

		//
		// If caching is disabled, add a Mime-type header
		//	
		if(!getUseCaches())
		{
			m_rpHttp->getRequestHeaders()->setHeaderExclusive(OT_T("Pragma"), OT_T("no-cache"));
		}
		
		//
		// Send the HTTP request, and parse the response
		// (possibly following redirects along the way)
		//
		// This will throw a variety of exceptions if it fails
		//
		m_rpHttp->sendRequest();

		//
		//  If we get this far, we have successfully connected
		//
		setConnected(true);

		//
		//  Reset the URL in case of redirects
		//
		setURL(m_rpHttp->getURL());
	}
}

//==============================================================================
// BasicHttpURLConnection::getInputStream
//
// Returns a (ref-counted) ptr to the socket's InputStream.
//==============================================================================
RefPtr<InputStream> BasicHttpURLConnection::getInputStream()
{
	connect();
	return m_rpHttp->getInputStream();
}

//==============================================================================
// BasicHttpURLConnection::getOutputStream
//
// Returns a (ref-counted) ptr to the socket's OutputStream.
//==============================================================================
RefPtr<OutputStream> BasicHttpURLConnection::getOutputStream()
{
	if(!getDoOutput())
	{
		throw ProtocolException(OT_T("URLConnection not enabled for output"));
	}

	if(isConnected())
	{
		throw IOException(OT_T("already opened for input"));
	}

	//
	// To get an early indication of a host error, 
	// connect to the server (if not already connected).  This is 
	// not strictly required (at present), but it is good practice
	// and it prepares the way for using HTTP/1.1 continue
	//
	m_rpHttp->connectToURL(getURL(), HTTP_TIMEOUT);

	//
	// Ifthe request method is still at the default (GET), then
	// it should be changed to POST.  The HttpClient doesn't do this
	// for us - it thinks we know what we're doing!
	if(m_rpHttp->getRequestMethod() == OT_T("GET"))
	{
		m_rpHttp->setRequestMethod(OT_T("POST"));
	}

	return m_rpHttp->getOutputStream();
}

//==============================================================================
// BasicHttpURLConnection::setFollowRedirects
//
//==============================================================================
void BasicHttpURLConnection::setFollowRedirects(bool bFollow)
{
	m_rpHttp->setFollowRedirects(bFollow);
}

//==============================================================================
// BasicHttpURLConnection::getFollowRedirects
//
//==============================================================================
bool BasicHttpURLConnection::getFollowRedirects() const
{
	return m_rpHttp->getFollowRedirects();
}

//==============================================================================
// BasicHttpURLConnection::setRequestProperty
//
//==============================================================================
void BasicHttpURLConnection::setRequestProperty(const String& name,
                                                const String& value)
{
	m_rpHttp->getRequestHeaders()->setHeaderExclusive(name, value);
}

//==============================================================================
// BasicHttpURLConnection::getRequestProperty
//
//==============================================================================
String BasicHttpURLConnection::getRequestProperty(const String& name) const
{
	return m_rpHttp->getRequestHeaders()->getHeader(name);
}

//==============================================================================
// BasicHttpURLConnection::getHeaderField
//
//==============================================================================
String BasicHttpURLConnection::getHeaderField(const String& name)
{
	connect();
	return m_rpHttp->getResponseHeaders()->getHeader(name);
}

//==============================================================================
// BasicHttpURLConnection::getHeaderField
//
// Note: We synthesize the first header which is the response line.
//==============================================================================
String BasicHttpURLConnection::getHeaderField(size_t index)
{
	connect();

	if(index == 0)
		return m_rpHttp->getResponseLine();
	else
		return m_rpHttp->getResponseHeaders()->getHeader(index-1);
}

//==============================================================================
// BasicHttpURLConnection::getHeaderFieldKey
//
// Note: We synthesize the first header which is the response line.
//==============================================================================
String BasicHttpURLConnection::getHeaderFieldKey(size_t index)
{
	connect();

	if(index == 0)
		return String();
	else
		return m_rpHttp->getResponseHeaders()->getHeaderKey(index-1);
}

//==============================================================================
// BasicHttpURLConnection::getHeaderFieldCount
//
// Note: We synthesize the first header which is the response line.
//==============================================================================
size_t BasicHttpURLConnection::getHeaderFieldCount()
{
	connect();
	return m_rpHttp->getResponseHeaders()->size() + 1;
}

//==============================================================================
// BasicHttpURLConnection::getErrorStream
//
//==============================================================================
RefPtr<InputStream> BasicHttpURLConnection::getErrorStream() const
{
	if(m_rpHttp->getResponseCode() >= 400)
	{
		return m_rpHttp->getInputStream();
	}
	else
	{
        return 0;
	}
}

//==============================================================================
// BasicHttpURLConnection::setRequestMethod
//
//==============================================================================
void BasicHttpURLConnection::setRequestMethod(const String& method)
{
	if(isConnected())
	{
		throw ProtocolException(OT_T("cannot set request method: already connected"));
	}

	m_rpHttp->setRequestMethod(method);
}

//==============================================================================
// BasicHttpURLConnection::getRequestMethod
//
//==============================================================================
String BasicHttpURLConnection::getRequestMethod() const
{
	return m_rpHttp->getRequestMethod();
}

//==============================================================================
// BasicHttpURLConnection::getResponseCode
//
//==============================================================================
int BasicHttpURLConnection::getResponseCode()
{
	connect();
	return m_rpHttp->getResponseCode();
}

//==============================================================================
// BasicHttpURLConnection::getResponseMessage
//
//==============================================================================
String BasicHttpURLConnection::getResponseMessage()
{
	connect();
	return m_rpHttp->getResponseMessage();
}

OT_NET_NAMESPACE_END
