//==============================================================================
// This file is part of the OpenTop C++ Library.
// (c) Copyright 2000-2002 ElCel Technology Limited
//
// This file may be distributed and/or modified under the terms of the
// GNU General Public License version 2 as published by the Free Software
// Foundation and appearing in the file LICENSE.GPL included in the
// packaging of this file.
//
// Licensees holding valid OpenTop Commercial licences may use this file
// in accordance with the OpenTop Commercial License Agreement included
// with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// See http://www.elcel.com/products/opentop/pricing.html or email
// sales@elcel.com for information about OpenTop Commercial License Agreements.
//
//==============================================================================
//
// $Revision: 1.1.1.1 $
// $Date: 2002/12/14 00:25:25 $
//
//==============================================================================

#include "FileURLConnection.h"
#include "ProtocolException.h"
#include "URL.h"
#include "URLDecoder.h"

#include "ot/base/StringUtils.h"
#include "ot/base/NumUtils.h"
#include "ot/io/IOException.h"
#include "ot/io/FileInputStream.h"
#include "ot/io/FileOutputStream.h"
#include "ot/io/File.h"

OT_NET_NAMESPACE_BEGIN

using namespace io;

//==============================================================================
// FileURLConnection::FileURLConnection
//
// Construct a FileURLConnection with a URL.  As with all URLConnections,
// construction is a two-stage process.  It is not complete until connect()
// has been called.
//==============================================================================
FileURLConnection::FileURLConnection(const URL& url) :
	BasicURLConnection(url)
{
}

//==============================================================================
// FileURLConnection::connect
//
// Perform the second-stage of construction, actually connnecting to the url
// that was provided to the constructor.
//
// For file: URLs this is a no-op.  We don't yet know whether the file
// URL is going to be used for reading or writing, so there is little
// point in checking if the file exists.  This will be done
// in getInputStream/getOutputStream.
//==============================================================================
void FileURLConnection::connect()
{
	if(!isConnected())
	{
		setConnected(true);
	}
}

//==============================================================================
// FileURLConnection::getInputStream
//
// Returns a (ref counted) ptr to the FileInputStream.
//
// If the url does not points at a directory rather than a file the JDK
// equivalent tries to be very helpful by creating an html listing of the
// directory.  This is probably a throw-back to the HotJava browser
// and has no useful place in OpenTop.
//
// Additionally, we try to be helpful by unescaping URL escape sequences.
// This may be overkill, but it is quite useful because it makes relative
// URLs behave the same locally as they would over an HTTP connection.
//==============================================================================
RefPtr<InputStream> FileURLConnection::getInputStream()
{
	connect();

	if(!m_rpInputStream)
	{
		if(m_rpOutputStream)
		{
			throw IOException(OT_T("already opened for output"));
		}

		if(!getDoInput())
		{
			throw ProtocolException(OT_T("URLConnection not enabled for input"));
		}

		File file(URLDecoder::RawDecode(getURL().getFile()));
		m_rpInputStream = new FileInputStream(file);
		String strLen = NumUtils::ToString(file.length());
		setHeaderField(OT_T("content-length"), strLen);
		DateTime modDate = file.lastModified();
		//  Format as RFC 822 eg: Thu, 25 Oct 2001 20:03:28 GMT
		setHeaderField(OT_T("last-modified"), modDate.Format(OT_T("%a, %d %b %Y %H:%M:%S GMT")));
	}

	return m_rpInputStream;
}

//==============================================================================
// FileURLConnection::getOutputStream
//
// Returns a (ref counted) ptr to the FileOutputStream.
//==============================================================================
RefPtr<OutputStream> FileURLConnection::getOutputStream()
{
	if(!getDoOutput())
	{
		throw ProtocolException(OT_T("URLConnection not enabled for output"));
	}

	if(m_rpInputStream)
	{
		throw IOException(OT_T("already opened for input"));
	}

	connect();

	if(!m_rpOutputStream)
	{
		File file(URLDecoder::RawDecode(getURL().getFile()));
		m_rpOutputStream = new FileOutputStream(file);
	}

	return m_rpOutputStream;
}

OT_NET_NAMESPACE_END
