//==============================================================================
// This file is part of the OpenTop C++ Library.
// (c) Copyright 2000-2002 ElCel Technology Limited
//
// This file may be distributed and/or modified under the terms of the
// GNU General Public License version 2 as published by the Free Software
// Foundation and appearing in the file LICENSE.GPL included in the
// packaging of this file.
//
// Licensees holding valid OpenTop Commercial licences may use this file
// in accordance with the OpenTop Commercial License Agreement included
// with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// See http://www.elcel.com/products/opentop/pricing.html or email
// sales@elcel.com for information about OpenTop Commercial License Agreements.
//
//==============================================================================
//
// $Revision: 1.1.1.1 $
// $Date: 2002/12/14 00:25:25 $
//
//==============================================================================

#include "FileURLStreamHandler.h"
#include "FileURLConnection.h"

#include "ot/base/StringUtils.h"
#include "ot/io/File.h"

OT_NET_NAMESPACE_BEGIN

using io::File;

//==============================================================================
// FileURLStreamHandler::openConnection
//
//==============================================================================
RefPtr<URLConnection> FileURLStreamHandler::openConnection(const URL& url) const
{
	return new FileURLConnection(url);
}

//==============================================================================
// FileURLStreamHandler::parsePath
//
// Specialized Helper function to determine the filename for a URL given a
// context and a (possibly) relative filename.
// This has been overloaded for the file: protocol because the URL for
// a file protocol can be a lot more relaxed and more system dependent
//==============================================================================
String FileURLStreamHandler::parsePath(const String& context, const String& path) const
{
	String ret = URLStreamHandler::parsePath(context, path);

#if defined(WIN32)

	// There is a good deal of confusion regarding what constitutes a legal
	// file: URL, especially on Windows platforms.
	// The following are all widely used: file:c:\rob    [1]
	//                                    file:///c:\rob [2]
	//                                    file://c:\rob  [3]
	//                                    file:///rob    [4]
	//                                    file:////<host>/<share>/ [5]
	// Case [3] appears invalid according to RFC1738 because the file: scheme
	// allows an optional host, which the "c:" could be interpreted as.
	// Case [1] is convenient because the resultant filename is directly
	// usable.
	// Case [4] is ambiguous because it may relate to a file in the
	// current directory of the root directory for the current drive.  In experiments
	// it seems JDK treats case [4] as refering to the root directory.  This is
	// reasonable as it mirrors the behaviour under UNIX, but means that relative
	// file URLs must be written as file:<path> without any "//".
	// 
	// 
	// To cater for case [2], we need to remove the leading "/" before
	// the local file system is asked to open it.
	if(ret.length() > 2 && ret[0] == '/' && ret[2] == ':')
	{
		ret.erase(0,1);
	}

#endif

	return ret;
}

//==============================================================================
// FileURLStreamHandler::toExternalForm
//
// Custom implementation for the @c file protocol.
//==============================================================================
String FileURLStreamHandler::toExternalForm(const URL& url) const
{
	String ret = url.getProtocol();
	ret += OT_T(":");
	
	const String& file = url.getFile();

	if((!url.getAuthority().empty()) || file.size() && file[0] == '/')
	{
		ret += OT_T("//");
		ret += url.getAuthority();
	}

	ret += url.getFile();

	if(!url.getRef().empty())
	{
		ret += OT_T("#");
		ret += url.getRef();
	}

	return ret;
}

OT_NET_NAMESPACE_END

