//==============================================================================
// This file is part of the OpenTop C++ Library.
// (c) Copyright 2000-2002 ElCel Technology Limited
//
// This file may be distributed and/or modified under the terms of the
// GNU General Public License version 2 as published by the Free Software
// Foundation and appearing in the file LICENSE.GPL included in the
// packaging of this file.
//
// Licensees holding valid OpenTop Commercial licences may use this file
// in accordance with the OpenTop Commercial License Agreement included
// with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// See http://www.elcel.com/products/opentop/pricing.html or email
// sales@elcel.com for information about OpenTop Commercial License Agreements.
//
//==============================================================================
//
// $Revision: 1.3 $
// $Date: 2003/09/10 20:17:51 $
//
//==============================================================================
//
// This class implements the URLConnection interface for FTP URLs.
//
// For convenience to client applications that wish to view HTTP/FTP
// URLs as abstract resources with similar characteristics, this class
// interprets the HTTP/1.1 Range header as an FTP restart request.
//
//==============================================================================

#include "FtpURLConnection.h"
#include "FtpClient.h"
#include "URL.h"
#include "URLDecoder.h"
#include "Socket.h"
#include "InetAddress.h"
#include "MalformedURLException.h"
#include "ProtocolException.h"

#include "ot/base/StringUtils.h"
#include "ot/base/NumUtils.h"
#include "ot/base/debug.h"
#include "ot/io/FileNotFoundException.h"
#include "ot/util/StringTokenizer.h"

OT_NET_NAMESPACE_BEGIN

using io::FileNotFoundException;
using util::StringTokenizer;

//==============================================================================
// FtpURLConnection::FtpURLConnection
//
// Construct a FtpURLConnection using the passed URL and default
// (or global) settings.
//==============================================================================
FtpURLConnection::FtpURLConnection(const URL& url) :
	BasicURLConnection(url),
	m_bDirectoryReqd(false),
	m_restartPoint(0)
{
}

//=============================================================================
// FtpURLConnection::connect
// 
// Creates a FtpClient to service the connection to the remote host
//
//=============================================================================
void FtpURLConnection::connect()
{
	if(!isConnected())
	{
		//
		// Determine if this is an anonymous or specified user.
		// If the URL contains userinfo, use that.
		//

		String user = OT_T("anonymous");
		String password = OT_T("opentop@");

		if(!getURL().getUserInfo().empty())
		{
			user = URLDecoder::RawDecode(getURL().getUserID());
			password = URLDecoder::RawDecode(getURL().getPassword());
		}

		m_rpFtpClient = new FtpClient;

		// don't wait for longer than 60 seconds - it's not really justified!
		const size_t timeout = 60000;

		m_rpFtpClient->connect(getURL().getHost(), getURL().getPort(), timeout);
		m_rpFtpClient->login(user, password);
		m_rpFtpClient->setDataConnectionTimeout(timeout);

		//
		// It's a good idea to be firewall-friendly and use passive data connections
		//
		m_rpFtpClient->setDataConnectionType(FtpClient::Passive);

		//
		// Now is a good time to massage the URL's filename
		//
		// This includes the following steps:
		// 1) unescape the entire path.  URL encoding such as "%20" is changed to " "
		// 2) check for a TYPE parameter
		// 3) isolate the filename from the path (some FTP servers won't serve up
		//    qualified paths
		// 4) update the path to remove the leading "/" (if any).  The URL syntax
		//    does not allow us to specify a path without the leading "/", but it
		//    is rare that a FTP url would be relative to the root filesystem.
		//
		// (1) Unescape the filename from the URL
		//
		String filename = URLDecoder::RawDecode(getURL().getFile());

		// (2) FTP URLS have an optional type parameter after the filename
		// file ::= file [;type=TYPE]?
		// RFC1738 states that TYPE must be "a" (ascii) "i" (image) or "d" (directory)
		// 
		// The default transmission type is binary
		//
		const size_t typePos = filename.find(OT_T(";type="));

		if(typePos != String::npos)
		{
			String strType = filename.substr(typePos+6);
			
			if(StringUtils::CompareNoCase(strType, OT_T("a")) == 0)
			{
				m_rpFtpClient->setTransferType(FtpClient::Ascii);
			}
			else if(StringUtils::CompareNoCase(strType, OT_T("i")) == 0)
			{
				m_rpFtpClient->setTransferType(FtpClient::Binary);
			}
			else
			{
				m_rpFtpClient->setTransferType(FtpClient::Ascii);
				m_bDirectoryReqd = true;
			}

			filename = filename.substr(0, typePos);
		}
		else
		{
			m_rpFtpClient->setTransferType(FtpClient::Binary);
		}

		//
		// (3) isolate the filename from the path
		//
		size_t pathPos = filename.find_last_of(OT_T("/"));
		if(pathPos != String::npos)
		{
			m_pathname = filename.substr(0, pathPos);
			if(pathPos < (filename.length() - 1))
			{
				m_filename = filename.substr(pathPos+1);
			}
		}
		else
		{
			m_filename = filename;
		}

		//
		// (4) update the path to remove the leading "/" (if any). 
		//
		if(m_pathname.length() && m_pathname.at(0) == '/')
		{
			m_pathname = m_pathname.substr(1);
		}

		setConnected(true);
	}
}

//==============================================================================
// FtpURLConnection::getInputStream
//
// Returns a (ref-counted) ptr to the FtpClient's InputStream.
//==============================================================================
RefPtr<InputStream> FtpURLConnection::getInputStream()
{
	connect();

	if(!m_rpInputStream)
	{
		try
		{
			if(m_rpOutputStream)
			{
				throw IOException(OT_T("already opened for output"));
			}

			if(!getDoInput())
			{
				throw ProtocolException(OT_T("URLConnection not enabled for input"));
			}

			//
			// change the FTP servers working directory to the appropriate place
			//
			changeDirectory();

			//
			// If the URL has specified that a directory listing is required
			// that's what we'll deliver.  
			// RFC1738 states that a NLST should be used, but this does not seem
			// appropriate when a filename has been specified as part of the URL
			// (the NLST doesn't like it), so we use LIST instead.
			//
			if(m_bDirectoryReqd)
			{
				if(m_filename.empty())
				{
					m_rpInputStream = m_rpFtpClient->listNames(m_filename);
				}
				else
				{
					m_rpInputStream = m_rpFtpClient->listDetails(m_filename);
				}
			}
			//
			// For a file transfer, we can set the content-length header, but only
			// if we are performing a binary (image) transfer, otherwise the
			// NVT encoding may change the number of bytes transmitted.
			//
			else
			{
				if(m_rpFtpClient->getTransferType() == FtpClient::Binary)
				{
					size_t fileSize = m_rpFtpClient->getFileSize(m_filename);
					setHeaderField(OT_T("content-length"), NumUtils::ToString(fileSize-m_restartPoint));
					if(m_restartPoint!=0)
					{
						// Produce a header in the form:
						// 	 Content-Range: bytes 4095-6394/6394
						String rangeHdr = OT_T("bytes ");
						rangeHdr+=NumUtils::ToString(m_restartPoint) + OT_T("-");
						rangeHdr+=NumUtils::ToString(fileSize-1) + OT_T("/");
						rangeHdr+=NumUtils::ToString(fileSize);
						setHeaderField(OT_T("Content-Range"), rangeHdr);
					}
					m_rpInputStream = m_rpFtpClient->retrieveFile(m_filename, m_restartPoint);
				}
				else
				{
					m_rpInputStream = m_rpFtpClient->retrieveFile(m_filename);
				}
			}
		}
		catch(IOException& /*e*/)
		{
		}

		if(!m_rpInputStream)
		{
			//
			// Use our representation of the filename when throwing the
			// FileNotFound exception - it gives a better indication of what has happened
			// wrt escaping, TYPE= removal etc.
			//
			String filename = m_pathname;
			if(filename.length())
				filename += OT_T("/");
			filename += m_filename;
			throw FileNotFoundException(filename);
		}
	}

	return m_rpInputStream;
}

//==============================================================================
// FtpURLConnection::getOutputStream
//
// Returns a (ref-counted) ptr to the FtpClient's OutputStream.
//==============================================================================
RefPtr<OutputStream> FtpURLConnection::getOutputStream()
{
	if(!getDoOutput())
	{
		throw ProtocolException(OT_T("URLConnection not enabled for output"));
	}

	if(m_rpInputStream)
	{
		throw IOException(OT_T("already opened for input"));
	}

	connect();

	if(!m_rpOutputStream)
	{
		if(m_filename.empty())
		{
			throw IOException(OT_T("filename missing"));
		}

		try
		{
			changeDirectory();
		}
		catch(Exception& /*e*/)
		{
			throw FileNotFoundException(m_pathname);
		}

		m_rpOutputStream = m_rpFtpClient->storeFile(m_filename);
	}

	return m_rpOutputStream;
}

//==============================================================================
// FtpURLConnection::changeDirectory
//
//==============================================================================
void FtpURLConnection::changeDirectory() const
{
	StringTokenizer tokenizer(m_pathname, OT_T("/"));
	while(tokenizer.hasMoreTokens())
	{
		if(!m_rpFtpClient->changeWorkingDirectory(tokenizer.nextToken()))
		{
			throw IOException(OT_T("unable to change directory"));
		}
	}
}

//==============================================================================
// FtpURLConnection::setRequestProperty
//
//==============================================================================
void FtpURLConnection::setRequestProperty(const String& name, const String& value)
{
	if(StringUtils::CompareNoCase(name, OT_T("Range"))==0)
	{
		//bytes=2341-
		if(value.substr(0,6) == OT_T("bytes=") && value[value.size()-1]==OT_T('-'))
		{
			m_restartPoint = NumUtils::ToLong(value.substr(6,value.size()-7));
		}
	}
}

//==============================================================================
// FtpURLConnection::getRequestProperty
//
//==============================================================================
String FtpURLConnection::getRequestProperty(const String& name) const
{
	if(m_restartPoint>0 && StringUtils::CompareNoCase(name, OT_T("Range"))==0)
	{
		String ret = OT_T("bytes=");
		ret += NumUtils::ToString(m_restartPoint) + OT_T("-");
		return ret;
	}
	else
	{
		return String();
	}
}

OT_NET_NAMESPACE_END

