//==============================================================================
// This file is part of the OpenTop C++ Library.
// (c) Copyright 2000-2002 ElCel Technology Limited
//
// This file may be distributed and/or modified under the terms of the
// GNU General Public License version 2 as published by the Free Software
// Foundation and appearing in the file LICENSE.GPL included in the
// packaging of this file.
//
// Licensees holding valid OpenTop Commercial licences may use this file
// in accordance with the OpenTop Commercial License Agreement included
// with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// See http://www.elcel.com/products/opentop/pricing.html or email
// sales@elcel.com for information about OpenTop Commercial License Agreements.
//
//==============================================================================
//
// $Revision: 1.2 $
// $Date: 2003/03/16 22:51:58 $
//
//==============================================================================
// http://www.ietf.org/rfc/rfc959.txt
//==============================================================================

#include "HttpClient.h"
#include "Authenticator.h"
#include "HttpChunkedInputStream.h"
#include "HttpURLConnection.h" // for response codes
#include "InetAddress.h"
#include "MimeHeaderParser.h"
#include "ProtocolException.h"
#include "Socket.h"
#include "ServerSocket.h"
#include "URLEncoder.h"

#include "ot/base/AutoBuffer.h"
#include "ot/base/NumUtils.h"
#include "ot/base/System.h"
#include "ot/base/StringUtils.h"
#include "ot/base/stringfunc.h"
#include "ot/base/Tracer.h"
#include "ot/base/debug.h"
#include "ot/io/ByteArrayOutputStream.h"
#include "ot/io/FileNotFoundException.h"
#include "ot/io/IOException.h"
#include "ot/io/InputStreamReader.h"
#include "ot/io/OutputStreamWriter.h"
#include "ot/util/AttributeListParser.h"
#include "ot/util/Base64.h"
#include "ot/util/MessageFormatter.h"
#include "ot/util/StringTokenizer.h"

#include <map>

OT_NET_NAMESPACE_BEGIN

using namespace util;
using namespace io;

const int HTTP_PORT = 80;
const int PROXY_PORT = 8080;
const int MAX_REDIRECTS = 10; // HTTP 1.0 used to suggest 5

//==============================================================================
// HttpClient::HttpClient
//
/**
   Constructs a new HttpClient with default values.
*/
//==============================================================================
HttpClient::HttpClient() :
	m_rpRequestHeaders(new MimeHeaderSequence),
	m_rpResponseHeaders(new MimeHeaderSequence),
	m_requestMethod(OT_T("GET")),
	m_nResponseCode(-1),
	m_bFollowRedirects(true),
	m_nProxyPort(0),
	m_timeoutMS(0)
{
	if(System::GetPropertyBool(OT_T("http.proxySet"), false))
	{
		m_proxyHost = System::GetProperty(OT_T("http.proxyHost"));
		m_nProxyPort = (int)System::GetPropertyLong(OT_T("http.proxyPort"), PROXY_PORT);
	}
}

//==============================================================================
// HttpClient::getDefaultPort
//
/**
   Returns the default port for HTTP servers which is 80.
*/
//==============================================================================
int HttpClient::getDefaultPort() const
{
	return HTTP_PORT;
}

//==============================================================================
// HttpClient::postConnect
//
/**
   Protected function called when a socket connection has been established with 
   the TCP/IP network server.
*/
//==============================================================================
void HttpClient::postConnect(const String& /*server*/,
                             int /*port*/,
                             size_t /*timeoutMS*/)
{
	const String encoding = OT_T("ISO-8859-1");

	m_rpWriter = new OutputStreamWriter(
		TcpNetworkClient::getOutputStream().get(), encoding);
}

//==============================================================================
// HttpClient::connectToURL
//
/**
   Establishes a TCP/IP socket connection with the host that will
   satisfy requests for the provided URL.

   This may connect to the host name contained within the URL, or to 
   a proxy server if one has been set.

   This function does not send any information to the remote server
   after the connection is established.

   @param url a fully-qualified http URL for the required resource
   @param timeoutMS the time-out value in milliseconds.  A value of zero indicates
          that the client will wait forever, or until the underlying
		  operating system decides that the connection cannot be established.
   @throws IOException if an error occurs trying to connect to the server
   @throws ProtocolException if an invalid response is received from the server
   @throws UnknownHostException if an error occurs trying to resolve the host name
   @throws SocketTimeoutException if the specified timeout value expires while
           attempting to connect to the server
   @sa sendRequest()
*/
//==============================================================================
void HttpClient::connectToURL(const URL& url, size_t timeoutMS)
{
	//
	// Guard against multiple connections
	//
	if(isConnected())
		return;

	//
	// Check we've been passed a valid URL
	//
	if(StringUtils::CompareNoCase(url.getProtocol().substr(0,4), OT_T("http")) != 0)
	{
		throw IllegalArgumentException(OT_T("not a http URL"));
	}

	//
	// Save the URL and timeout for the future
	// (in case we need them again e.g. when redirected to another host)
	//
	m_url = url;
	m_timeoutMS = timeoutMS;

	String sHost = url.getHost();
	int port = url.getPort();

	//
	// If we need to use a proxy server, then connect to that,
	// otherwise create a connection to the specified host.
	//
	if(m_proxyHost.empty())
	{
		TcpNetworkClient::connect(sHost, port, timeoutMS);
	}
	else
	{
		TcpNetworkClient::connect(m_proxyHost, m_nProxyPort, timeoutMS);
	}
}

//=============================================================================
// HttpClient::sendRequest
// 
// Send the http request to the remote host.
// 
// The request is made up of a request line followed by request
// header fields.  Eg:-
// GET filename HTTP/1.1
// Host: hostname[:port]
// Connection: close
// 
// The response from the server will contain a number of header
// fields followed by the requested data.  The headers are parsed
// and removed leaving the InputStream to point at the requested
// data.
//
// Note: HTTP Redirection
// ----------------------
// By default we will follow HTTP redirects.  These are communicated
// to us by a 3xx HTTP response code and the presence of a "Location" header
// field.  A 3xx response code without a Location header is an error.
// Redirection may be an iterative process, so it continues until
// we receive a 200 OK response orthe maximum number of redirects is exceeded.
//
// We do not process Location headers when accompanying a 200 OK response.
//=============================================================================
int HttpClient::sendRequest()
{
	if(!isConnected())
	{
		throw ProtocolException(OT_T("Http client not connected"));
	}

	int redirectCount=0;
	const String CRLF = OT_T("\r\n");

	//
	// Add a User-Agent header identifying OpenTop, unless the application
	// has inserted its own header
	//
	String sUserAgent = OT_T("opentop ");
	sUserAgent += System::GetVersionAsString();
	m_rpRequestHeaders->setHeaderIfAbsent(OT_T("User-Agent"), sUserAgent);

	while(redirectCount == 0 || (m_bFollowRedirects && redirectCount < MAX_REDIRECTS))
	{
		if(!isConnected())
			connectToURL(m_url, m_timeoutMS);

		String request = m_requestMethod + OT_T(" ");
		
		if(m_proxyHost.empty())
		{
			if(m_url.getFile().empty())
			{
				request += OT_T("/"); 
			}
			else
			{
				request += URLEncoder::RawEncode(m_url.getFile());
			}
			request += OT_T(" HTTP/1.1");

			// As we don't yet support persistent connections
			// we must include a Connection: close header
			m_rpRequestHeaders->setHeaderExclusive(OT_T("Connection"), OT_T("close"));
		}
		else
		{
			// We'll drop down to HTTP 1.0 for the proxy server
			// I don't know why, but this is what JDK does from examining
			// the httpd log
			//
			// The URL should probably not contain the user information
			// part or the ref part when it is transmitted to the server,
			// so we'll assemble a synthesized URL missing those components
			//
			request += m_url.getProtocol();
			request += OT_T("://");
			request += m_url.getHost();
			request += URLEncoder::RawEncode(m_url.getPath());
			if(!m_url.getQuery().empty())
			{
				request += OT_T("?");
				request += URLEncoder::RawEncode(m_url.getQuery());
			}

			request += OT_T(" HTTP/1.0");

			// As we don't yet support persistent connections
			// we must include a Connection: close header
			m_rpRequestHeaders->setHeaderExclusive(OT_T("Proxy-Connection"), OT_T("close"));
		}

		//
		// If the URL contains a port, then add that to the Host header
		//
		String sHost = m_url.getHost();
		if(m_url.getPort() != -1)
		{
			sHost += OT_T(":");
			sHost += NumUtils::ToString(m_url.getPort());
		}
		m_rpRequestHeaders->setHeaderExclusive(OT_T("Host"), sHost);

		Tracer::Trace(Tracer::Net, Tracer::Low, request);

		//
		// If we have created an output stream (on demand)
		// then close it and use its length to generate a 
		// "content-length" header
		//
		ByteArrayOutputStream* pOS = reinterpret_cast<ByteArrayOutputStream*>(m_rpOutputStream.get());
		if(pOS)
		{
			m_rpOutputStream->close();
			m_rpRequestHeaders->setHeaderExclusive(
				OT_T("content-length"),
				NumUtils::ToString(pOS->size()));
		}

		//
		// Write the request to the HTTP server
		//
		m_rpWriter->write(request + CRLF);

		//
		// Write the MIME-type headers to the HTTP server
		//
		m_rpRequestHeaders->writeHeaders(m_rpWriter.get());

		//
		// Indicate the end of the headers with an empty line
		//
		m_rpWriter->write(CRLF);
		m_rpWriter->flush();

		//
		// Write the contents of the OutputStream (if any).
		// Note: this of course by-passes the writer and writes 
		// directly to the byte stream
		//
		if(pOS)
		{
			pOS->writeTo(TcpNetworkClient::getOutputStream().get());
			TcpNetworkClient::getOutputStream()->flush();
		}

		//
		// Read the result line and attached headers...
		//
		parseReturnedHeaders();

		if(m_bFollowRedirects && m_nResponseCode >= 300 && m_nResponseCode < 400)
		{
			disconnect();

			++redirectCount;

			if(redirectCount > MAX_REDIRECTS)
			{
				throw IOException(OT_T("Redirection limit reached"));
			}

			String newLocation = m_rpResponseHeaders->getHeader(OT_T("Location"));

			if(newLocation.empty())
			{
				const String& errMsg = MessageFormatter::Format(
					OT_T("Location header missing from HTTP redirect response: {0} ({1}) for URL: {2}"),
					NumUtils::ToString(m_nResponseCode),
					m_responseMessage,
					m_url.toExternalForm());
				throw ProtocolException(errMsg);
			}
			else
			{
				//
				// Combine the new location with our URL
				//
				m_url = URL(m_url, newLocation);

				//
				// New locations will possibly need different authentication,
				// so we should reset our origin authentication header (if any)
				//
				m_rpRequestHeaders->removeAllHeaders(OT_T("Authorization"));	
			}
		}
		else if(m_nResponseCode == HttpURLConnection::HTTP_UNAUTHORIZED)
		{
			//
			// If we haven't already done so, attempt to create an 
			// Authentication header.  If this fails
			// (due to application not passing the credentials), then
			// we treat it as an error.
			//
			// If we already have one then the server is rejecting it so
			// we have an error anyway.
			//
			if(m_rpRequestHeaders->containsHeader(OT_T("Authorization"))
			   || !createAuthorizationHeader())
			{
				break;
			}
			disconnect();
		}
		else if(m_nResponseCode == HttpURLConnection::HTTP_PROXY_AUTH)
		{
			//
			// If we haven't already done so, attempt to create an 
			// Authentication header.  If this fails
			// (due to application not passing the credentials), then
			// we treat it as an error.
			//
			// If we already have one then the server is rejecting it so
			// we have an error anyway.
			//
			if(m_rpRequestHeaders->containsHeader(OT_T("Proxy-Authorization"))
			   || !createAuthorizationHeader())
			{
				break;
			}
			disconnect();
		}
		else
		{
			break;
		}
	}

	const String& tfrEncoding = 
		m_rpResponseHeaders->getHeader(OT_T("Transfer-Encoding"));

	if(StringUtils::CompareNoCase(tfrEncoding, OT_T("chunked")) ==0)
	{
		m_rpInputStream = new HttpChunkedInputStream(
			TcpNetworkClient::getInputStream().get());
	}
	else
	{
		m_rpInputStream = TcpNetworkClient::getInputStream();
	}

	//
	// Any error code 300 or above becomes an eexception at this point.
	//
	if(m_nResponseCode >= 300) 
	{
		//
		// We don't disconnect from the server because the application
		// may wish to read from the error stream after catching
		// the exception we are about to throw!
		//disconnect();

		if(m_nResponseCode == HttpURLConnection::HTTP_NOT_FOUND)
		{
			throw FileNotFoundException(m_url.toExternalForm());
		}
		else
		{
			const String& errMsg = MessageFormatter::Format(
				OT_T("server returned HTTP response: {0} ({1}) for URL: {2}"),
				NumUtils::ToString(m_nResponseCode),
				getResponseMessage(),
				m_url.toExternalForm());
			throw IOException(errMsg);
		}
	}

	return m_nResponseCode;
}

//==============================================================================
// HttpClient::createAuthenticationHeader
//
// In response to a HTTP_UNAUTHORISED response from the HTTP server,
// this function will attempt to generate an Authentication header
// to satisfy the server.
//==============================================================================
bool HttpClient::createAuthorizationHeader()
{
	const bool bProxy = (m_nResponseCode == HttpURLConnection::HTTP_PROXY_AUTH);

	const String sHeader = bProxy
	                     ? OT_T("Proxy-Authenticate")
	                     : OT_T("WWW-Authenticate");
	
	const String& authResponse = m_rpResponseHeaders->getHeader(sHeader);

	if(authResponse.empty())
	{
		// assume there's no need for an error stream
		// in this case
		disconnect();

		const String& errMsg = MessageFormatter::Format(
			OT_T("{0} header missing from HTTP response: {1} ({2}) for URL: {3}"),
			sHeader,
			NumUtils::ToString(m_nResponseCode),
			getResponseMessage(),
			getURL().toExternalForm());
		throw ProtocolException(errMsg);
	}

	//
	// The authentication header is constructed like a tagged attribute
	// list.  The AttributeListParser can deal with this.
	//
	AttributeListParser parser;
	if(!parser.parseString(authResponse))
	{
		const String& errMsg = MessageFormatter::Format(
			OT_T("{0} header value: {1} is invalid"),
			sHeader, authResponse);
		throw ProtocolException(errMsg);
	}

	const String& scheme = parser.getAttributeValue(0);
	const String& realm = parser.getAttributeValueICase(OT_T("realm"));

	//
	// Obtain a userid and password for the scheme/realm returned
	// from the HTTP server
	//
	String user, password;

	if(!Authenticator::RequestPasswordAuthentication(getServerAddress().get(),
		getServerPort(), OT_T("HTTP"),
		realm, scheme, bProxy,
		user, password))
	{
		//
		// If the registered Authenticator cannot supply a user/password
		// then we cannot continue.  This is signalled by returning
		// false to the sendRequest() function.
		//
		return false;
	}

	//
	// According to RFC2617 HTTP Authentication: Basic and Digest Access Authentication
	// For "Basic" authentication, the user and password are concatentated with a
	// colon separator before being encoded in base64
	//
	// According to RFC 2068 (HTTP/1.1) the Username and Password are defined as
	// TEXT productions and are therefore supposed to be encoded in ISO-8859-1
	// before being Base64-encoded.
	//
	// TODO: schemes may be listed, we need to parse the value
	//
	if(StringUtils::CompareNoCase(scheme, OT_T("Basic")) == 0)
	{
		const String sColon = OT_T(":");
		String headerValue = OT_T("Basic ");

		String combo = user + sColon + password;
		ByteString bytes = StringUtils::ToLatin1(combo);

		const size_t base64Size = 
			Base64::GetEncodedLength((const Byte*)bytes.data(),
			                         (const Byte*)bytes.data()+bytes.size());

		ArrayAutoPtr<Byte> apBuffer(new Byte[base64Size]);
		Byte* pToNext;
		if(Base64::Encode((const Byte*)bytes.data(),
		                  (const Byte*)bytes.data()+bytes.size(),
		                  apBuffer.get(),
		                  apBuffer.get()+base64Size,
		                  pToNext) != Base64::ok)
		{
			throw IOException(OT_T("unable to base64 encode password"));
		}

		OT_DBG_ASSERT(base64Size ==  (size_t)(pToNext-apBuffer.get()));
		
		headerValue += StringUtils::FromLatin1((const char*)apBuffer.get(), pToNext-apBuffer.get());

		const String sReqHeader = bProxy
		                        ? OT_T("Proxy-Authorization")
			                    : OT_T("Authorization");

		m_rpRequestHeaders->setHeaderExclusive(sReqHeader, headerValue);

		return true;
	}
	else
	{
		//
		// Only basic authentication is supported at present.  By failing
		// to create an authentication header, and returning false, we
		// signal to the caller that the authenticate response should be
		// treated as an error.
	}

	return false;
}

//=============================================================================
// HttpClient::parseReturnedHeaders
// 
// A typical http response will take the form:-
// HTTP/1.1 200 OK
// Date: Wed, 25 Oct 2000 16:54:02 GMT
// Age: 57718
// Server: Apache/1.3b5
// Last-Modified: Sat, 23 Sep 2000 15:00:57 GMT
// Accept-Ranges: bytes
// Content-Type: text/html
// Content-Length: 11660
// Etag: "122b12-2d8c-39ccc5a9"
// Via: 1.1 www-cache.pol.co.uk (NetCache 4.1R6), 1.1 www-cache.pol.co.uk (NetCache
// 
// <the data>
//
// We make use of the MimeHeaderParser utility module to parse the headers
// for us.
//=============================================================================
void HttpClient::parseReturnedHeaders()
{
	OT_DBG_ASSERT(isConnected());

	const RefPtr<InputStream>& rpInputStream = TcpNetworkClient::getInputStream();

	//
	// reset our state information
	//
	m_nResponseCode = 0;
	m_responseMessage.erase();

	MimeHeaderParser::ReadLineLatin1(rpInputStream.get(), m_responseLine);

	//
	// Parse the response line:
	// HTTP/n.n nnn <ssss>?
	//
	size_t codePos = m_responseLine.find(OT_T(' '));
	size_t msgPos = String::npos;
	if(codePos != String::npos)
	{
		msgPos = m_responseLine.find(OT_T(' '), codePos+1);
	}
	if(msgPos != String::npos)
	{
		String strCode = m_responseLine.substr(codePos+1, msgPos-(codePos+1));
		m_nResponseCode = NumUtils::ToLong(strCode);
		m_responseMessage = m_responseLine.substr(msgPos+1);
	}
	else
	{
		m_responseMessage.erase();
	}

	m_rpResponseHeaders = MimeHeaderParser::ParseHeaders(rpInputStream.get());
}

//==============================================================================
// HttpClient::getInputStream
//
// Returns a (ref-counted) ptr to the socket's InputStream.
//==============================================================================
RefPtr<InputStream> HttpClient::getInputStream() const
{
	return m_rpInputStream;
}

//==============================================================================
// HttpClient::getOutputStream
//
// Returns an OutputStream that can be used by the application for
// POST or PUT operations.
//==============================================================================
RefPtr<OutputStream> HttpClient::getOutputStream() const
{
	if(!m_rpOutputStream)
	{
		const_cast<HttpClient*>(this)->m_rpOutputStream = 
			new ByteArrayOutputStream;
	}
	return m_rpOutputStream;
}

//==============================================================================
// HttpClient::setRequestMethod
//
//==============================================================================
void HttpClient::setRequestMethod(const String& method)
{
	m_requestMethod = method;
}

//==============================================================================
// HttpClient::getRequestMethod
//
//==============================================================================
String HttpClient::getRequestMethod() const
{
	return m_requestMethod;
}

//==============================================================================
// HttpClient::getResponseCode
//
//==============================================================================
int HttpClient::getResponseCode() const
{
	return m_nResponseCode;
}

//==============================================================================
// HttpClient::getResponseMessage
//
//==============================================================================
const String& HttpClient::getResponseMessage() const
{
	return m_responseMessage;
}

//==============================================================================
// HttpClient::setProxyHost
//
/**
   Sets a proxy host to use for connections from this HttpClient instance.
*/
//==============================================================================
void HttpClient::setProxyHost(const String& host)
{
	m_proxyHost = host;
}

//==============================================================================
// HttpClient::setProxyPort
//
/**
   Sets the port number on the proxy host to use for connections from this
   HttpClient instance.
*/
//==============================================================================
void HttpClient::setProxyPort(int port)
{
	m_nProxyPort = port;
}

//==============================================================================
// HttpClient::getProxyHost
//
/**
   Returns the name of the proxy host that will be used for this
   HttpClient instance.
*/
//==============================================================================
String HttpClient::getProxyHost() const
{
	return m_proxyHost;
}

//==============================================================================
// HttpClient::getProxyPort
//
/**
   Returns the proxy host port number that will be used for this
   HttpClient instance.
*/
//==============================================================================
int HttpClient::getProxyPort() const
{
	return m_nProxyPort;
}

//==============================================================================
// HttpClient::getResponseHeaders
//
/**
   Returns a sequence of Mime-type headers that were received in
   response from the HTTP server.
*/
//==============================================================================
RefPtr<MimeHeaderSequence> HttpClient::getResponseHeaders() const
{
	return m_rpResponseHeaders;
}

//==============================================================================
// HttpClient::getRequestHeaders
//
/**
   Returns a modifiable sequence of Mime-type headers that will be used to form
   a request to the HTTP server.
*/
//==============================================================================
RefPtr<MimeHeaderSequence> HttpClient::getRequestHeaders() const
{
	return m_rpRequestHeaders;
}

//==============================================================================
// HttpClient::getFollowRedirects
//
/**
   Returns a flag indicating if HTTP redirects will be followed.

   @returns @c true if redirect requests will be followed (the default);
            @c false otherwise
   @sa setFollowRedirects()
*/
//==============================================================================
bool HttpClient::getFollowRedirects() const
{
	return m_bFollowRedirects;
}

//==============================================================================
// HttpClient::setFollowRedirects
//
/**
   Sets a flag indicating if HTTP redirects will be followed.

   @param bFollow @c true for redirect requests to be followed (the default);
                  @c false prevents this behaviour.
   @sa getFollowRedirects()
*/
//==============================================================================
void HttpClient::setFollowRedirects(bool bFollow)
{
	m_bFollowRedirects = bFollow;
}

//==============================================================================
// HttpClient::getURL
//
/**
   Returns the URL that was last connect.  This will reflect the latest
   URL, updated as a result of any HTTP redirect processing.
*/
//==============================================================================
URL HttpClient::getURL() const
{
	return m_url;
}

//==============================================================================
// HttpClient::getResponseLine
//
/**
   Returns the response line from the HTTP server.
*/
//==============================================================================
const String& HttpClient::getResponseLine() const
{
	return m_responseLine;
}

OT_NET_NAMESPACE_END
