//==============================================================================
// This file is part of the OpenTop C++ Library.
// (c) Copyright 2000-2002 ElCel Technology Limited
//
// This file may be distributed and/or modified under the terms of the
// GNU General Public License version 2 as published by the Free Software
// Foundation and appearing in the file LICENSE.GPL included in the
// packaging of this file.
//
// Licensees holding valid OpenTop Commercial licences may use this file
// in accordance with the OpenTop Commercial License Agreement included
// with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// See http://www.elcel.com/products/opentop/pricing.html or email
// sales@elcel.com for information about OpenTop Commercial License Agreements.
//
//==============================================================================
//
// $Revision: 1.1.1.1 $
// $Date: 2002/12/14 00:25:30 $
//
//==============================================================================

#include "MimeHeaderParser.h"

#include "ot/base/StringUtils.h"
#include "ot/base/Tracer.h"

OT_NET_NAMESPACE_BEGIN

//==============================================================================
// MimeHeaderSequence::parseInputStream
//
// Clear the headers before parsing an InputStream for MIME headers.
// 
// Mime-type headers comprise a keyword followed by ":" followed
// by whitespace and the value.  If a line begins with witespace it is
// a continuation of the preceeding line.
//
// MIME headers are delimited by an empty line.
//=============================================================================
RefPtr<MimeHeaderSequence> MimeHeaderParser::ParseHeaders(InputStream* pInputStream)
{
	if(!pInputStream) throw NullPointerException();
	
	RefPtr<MimeHeaderSequence> rpRet(new MimeHeaderSequence);

	Tracer::Trace(Tracer::Net, Tracer::Low, OT_T("reading headers:"));

	//
	// Until we reach the data part of the response we know we are
	// dealing with iso-8859-1 characters
	//
	String line;

	while (ReadLineLatin1(pInputStream, line) != InputStream::EndOfFile)
	{
		if(line.empty())
		{
			// we have reached the MIME headers delimiter
			break;
		}
		else
		{
			Tracer::Trace(Tracer::Net, Tracer::Low, line);

			//
			// Check for continuation of preceding header
			//
			if(UnicodeCharacterType::IsSpace(line[0]))
			{
				if(!rpRet->m_headerFields.empty())
				{
					//
					// The following strange syntax came about trying to
					// get ospace stl to work with workshop 4.2
					// - it didn't like m_headerFields[m_headerFields->size()-1]
					//
					MimeHeaderSequence::HeaderFieldEntry& entry = 
						*(rpRet->m_headerFields.end()-1);

					entry.second += StringUtils::StripWhiteSpace(line, StringUtils::both);
				}
			}
			else
			{
				size_t delimPos = line.find(OT_T(':'));
				String key;
				String value;
				if(delimPos != String::npos)
				{
					key = line.substr(0, delimPos);
					// strip any leading and trailing white-space
					value = StringUtils::StripWhiteSpace(line.substr(delimPos+1), StringUtils::both);
				}
				else
				{
					value = StringUtils::StripWhiteSpace(line, StringUtils::trailing);
				}
				rpRet->insertHeader(key, value);
			}
		}
	}
	return rpRet;
}

//==============================================================================
// MimeHeaderParser::ReadLineLatin1
//
//==============================================================================
long MimeHeaderParser::ReadLineLatin1(InputStream* pInputStream, String& retLine)
{
	retLine.erase();

	long charCount=0;
	int x;

	while( (x=pInputStream->read()) != InputStream::EndOfFile)
	{
		if(x == '\n')
		{
			return charCount;
		}
		else if(x != '\r')
		{
			Character ch((UCS4Char)x);
			ch.appendToString(retLine);
			++charCount;
		}
	}
	return InputStream::EndOfFile;
}

OT_NET_NAMESPACE_END
