//==============================================================================
// This file is part of the OpenTop C++ Library.
// (c) Copyright 2000-2002 ElCel Technology Limited
//
// This file may be distributed and/or modified under the terms of the
// GNU General Public License version 2 as published by the Free Software
// Foundation and appearing in the file LICENSE.GPL included in the
// packaging of this file.
//
// Licensees holding valid OpenTop Commercial licences may use this file
// in accordance with the OpenTop Commercial License Agreement included
// with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// See http://www.elcel.com/products/opentop/pricing.html or email
// sales@elcel.com for information about OpenTop Commercial License Agreements.
//
//==============================================================================
//
// $Revision: 1.1.1.1 $
// $Date: 2002/12/14 00:25:30 $
//
//==============================================================================

#include "MimeHeaderSequence.h"

#include "ot/base/Tracer.h"
#include "ot/base/StringUtils.h"

#include <functional>
#include <algorithm>

OT_NET_NAMESPACE_BEGIN

struct KeyEquals : std::binary_function<std::pair<String, String>, String, bool>
{
	bool operator()(const std::pair<String, String>& lhs, const String& rhs) const
	{
		return (StringUtils::CompareNoCase(lhs.first, rhs) == 0);
	}
};

//==============================================================================
// MimeHeaderSequence::getHeader
//
//==============================================================================
String MimeHeaderSequence::getHeader(const String& key) const
{
	int i = findHeader(key);
	if(i == -1)
	{
		return String();
	}
	else
	{
		return m_headerFields[i].second;
	}
}

//==============================================================================
// MimeHeaderSequence::getHeader
//
//==============================================================================
String MimeHeaderSequence::getHeader(size_t index) const
{
	if(index<m_headerFields.size())
	{
		return m_headerFields[index].second;
	}
	else
	{
		return String();
	}
}

//==============================================================================
// MimeHeaderSequence::getHeaderKey
//
//==============================================================================
String MimeHeaderSequence::getHeaderKey(size_t index) const
{
	if(index<m_headerFields.size())
	{
		return m_headerFields[index].first;
	}
	else
	{
		return String();
	}
}

//==============================================================================
// MimeHeaderSequence::size
//
//==============================================================================
size_t MimeHeaderSequence::size() const
{
	return m_headerFields.size();
}

//==============================================================================
// MimeHeaderSequence::setHeader
//
// Sets a header field, overwriting any existing value
//==============================================================================
void MimeHeaderSequence::setHeaderExclusive(const String& key, const String& value)
{
	removeAllHeaders(key);
	insertHeader(key, value);
}

//==============================================================================
// MimeHeaderSequence::setHeaderIfAbsent
//
//==============================================================================
bool MimeHeaderSequence::setHeaderIfAbsent(const String& key, const String& value)
{
	if(findHeader(key) == -1)
	{
		insertHeader(key, value);
		return true;
	}
	else
	{
		return false;
	}
}

//==============================================================================
// MimeHeaderSequence::findHeader
//
//==============================================================================
int MimeHeaderSequence::findHeader(const String& key) const
{
	for(size_t i=0; i<m_headerFields.size(); i++)
	{
		if(StringUtils::CompareNoCase(m_headerFields[i].first, key)==0)
		{
			return i;
		}
	}
	return -1;
}

//==============================================================================
// MimeHeaderSequence::containsHeader
//
//==============================================================================
bool MimeHeaderSequence::containsHeader(const String& key) const
{
	return (findHeader(key) != -1);
}

//==============================================================================
// MimeHeaderSequence::insertHeader
//
//==============================================================================
void MimeHeaderSequence::insertHeader(const String& key, const String& value)
{
	m_headerFields.push_back(std::make_pair(key, value));
}

//==============================================================================
// MimeHeaderSequence::removeAllHeaders
//
//==============================================================================
void MimeHeaderSequence::removeAllHeaders(const String& key)
{
	HeaderFieldVector::iterator last = 
		std::remove_if(m_headerFields.begin(),
		               m_headerFields.end(), 
		               std::bind2nd(KeyEquals(), key));

	m_headerFields.erase(last, m_headerFields.end());
}

//==============================================================================
// MimeHeaderSequence::clear
//
//==============================================================================
void MimeHeaderSequence::clear()
{
	m_headerFields.clear();
}

//==============================================================================
// MimeHeaderSequence::writeHeaders
//
// Writes all the headers to the supplied Writer.
//==============================================================================
void MimeHeaderSequence::writeHeaders(Writer* pWriter)
{
	if(!pWriter) throw NullPointerException();
	const String sep = OT_T(": ");
	const String CRLF = OT_T("\r\n");

	Tracer::Trace(Tracer::Net, Tracer::Low, OT_T("writing headers:"));

	for(HeaderFieldVector::iterator i=m_headerFields.begin();
		i != m_headerFields.end(); ++i)
	{
		String header = (*i).first + sep + (*i).second; 

		Tracer::Trace(Tracer::Net, Tracer::Low, header);

		pWriter->write(header + CRLF);
	}
}

OT_NET_NAMESPACE_END
