//==============================================================================
// This file is part of the OpenTop C++ Library.
// (c) Copyright 2000-2002 ElCel Technology Limited
//
// This file may be distributed and/or modified under the terms of the
// GNU General Public License version 2 as published by the Free Software
// Foundation and appearing in the file LICENSE.GPL included in the
// packaging of this file.
//
// Licensees holding valid OpenTop Commercial licences may use this file
// in accordance with the OpenTop Commercial License Agreement included
// with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// See http://www.elcel.com/products/opentop/pricing.html or email
// sales@elcel.com for information about OpenTop Commercial License Agreements.
//
//==============================================================================
//
// $Revision: 1.4 $
// $Date: 2003/09/04 21:10:53 $
//
//==============================================================================
/**
	@class ot::net::SocketDescriptor
	
	@brief An abstract base class representing socket handles.

	By using reference-counted SocketDescriptors, @OpenTop is able to share
	socket handles between multiple objects, preventing the socket from being
	closed until all references to the SocketDescriptor have been removed.

	This is a SynchronizedObject to allow OpenTop to synchronize some socket
	operations so that they may be safely called by multiple threads.
*/
//==============================================================================

#include "SocketDescriptor.h"
#include "NetUtils.h"

#include "ot/base/NumUtils.h"
#include "ot/base/Tracer.h"
#include "ot/io/IOException.h"

OT_NET_NAMESPACE_BEGIN

using io::IOException;

//==============================================================================
// SocketDescriptor::SocketDescriptor
//
/**
   Constructs a SocketDescriptor from an operating system socket handle.
*/
//==============================================================================
SocketDescriptor::SocketDescriptor(OSSocketDescriptorType sockfd) :
	ResourceDescriptor(true /*auto close*/),
	m_fd(sockfd),
	m_flags(0)
{
}

//==============================================================================
// SocketDescriptor::~SocketDescriptor
//
/**
   Destructor.

   Calls getAutoClose() and closes the related resource if it returns @c true.
*/
//==============================================================================
SocketDescriptor::~SocketDescriptor()
{
	if(getAutoClose())
	{
		try
		{
			close();
		}
		catch(Exception& /*e*/)
		{
		}
	}
}

//==============================================================================
// SocketDescriptor::close
//
/**
   Closes the socket.
*/
//==============================================================================
void SocketDescriptor::close()
{
	Tracer::Trace(Tracer::Net, Tracer::Medium, String(OT_T("Closing socket: ")) + toString());

	int rc=
#if defined(WIN32)
	::closesocket(m_fd);
#else
	::close(m_fd);
#endif

	if(rc != 0)
	{
		static const String err(OT_T("error closing socket: "));
		String errMsg = err + NetUtils::GetSocketErrorString();
		throw IOException(errMsg);
	}

	OT_SYNCHRONIZED
	m_flags |= DescriptorClosed;
	setAutoClose(false);
}

//==============================================================================
// SocketDescriptor::getFD
//
/**
	Returns an operating system handle for the socket.
*/
//==============================================================================
SocketDescriptor::OSSocketDescriptorType SocketDescriptor::getFD() const
{
	return m_fd;
}

//==============================================================================
// SocketDescriptor::getSocketFlags
//
/**
   Returns the current socket flags.

   @sa setSocketFlags
   @mtsafe
*/
//==============================================================================
int SocketDescriptor::getSocketFlags() const
{
	OT_SYNCHRONIZED

	return m_flags;
}

//==============================================================================
// SocketDescriptor::setSocketFlags
//
/**
   Sets the socket flags.

   Each open socket has a set of flags associated with it for use by internal
   @OpenTop routines.

   @param flags the new flag value
   
   @mtsafe
*/
//==============================================================================
void SocketDescriptor::setSocketFlags(int flags)
{
	OT_SYNCHRONIZED
	m_flags = flags;
}

//==============================================================================
// SocketDescriptor::toString
//
//==============================================================================
String SocketDescriptor::toString() const
{
	return NumUtils::ToString(m_fd);
}

//==============================================================================
// SocketDescriptor::setSocketFlags
//
/**
   Sets the socket flags.

   Each open socket has a set of flags associated with it for use by internal
   @OpenTop routines.

   @param flagsToSet a value whose bits are ORed with the existing flag bits to
          arrive at a new value.
   @param flagsToUnset a value whose bits are NANDedd with the existing flag bits to
          arrive at a new value.
   @returns the new flag value
   @mtsafe
*/
//==============================================================================
int SocketDescriptor::modifySocketFlags(int flagsToSet, int flagsToUnset)
{
	OT_SYNCHRONIZED
	m_flags = ((m_flags | flagsToSet) & ~flagsToUnset);

	//
	// If the AutoCloseEnabled flag is set and the socket has not yet been closed
	// and input is at eof (or shutdown which is the same thing)
	// then the socket is automatically closed.
	//
	if((m_flags & (ShutdownInput | AutoCloseEnabled)) == (ShutdownInput | AutoCloseEnabled) &&
	   (m_flags & (HasOutputStream|DescriptorClosed)) == 0)
	{
		try
		{
			close();
		}
		catch(Exception& /*e*/)
		{
		}
	}

	return m_flags;
}

OT_NET_NAMESPACE_END
