//==============================================================================
// This file is part of the OpenTop C++ Library.
// (c) Copyright 2000-2002 ElCel Technology Limited
//
// This file may be distributed and/or modified under the terms of the
// GNU General Public License version 2 as published by the Free Software
// Foundation and appearing in the file LICENSE.GPL included in the
// packaging of this file.
//
// Licensees holding valid OpenTop Commercial licences may use this file
// in accordance with the OpenTop Commercial License Agreement included
// with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// See http://www.elcel.com/products/opentop/pricing.html or email
// sales@elcel.com for information about OpenTop Commercial License Agreements.
//
//==============================================================================
//
// $Revision: 1.1.1.1 $
// $Date: 2002/12/14 00:25:39 $
//
//==============================================================================
/**
	@class ot::net::URLEncoder
	
	@brief A class module containing static methods for encoding strings
	into a form that can be transmitted as a URL.
*/
//==============================================================================

#include "URLEncoder.h"

#include "ot/base/ArrayAutoPtr.h"
#include "ot/base/StringUtils.h"
#include "ot/base/StringIterator.h"

#include <stdio.h>  
#include <ctype.h>  
#include <algorithm>  

OT_NET_NAMESPACE_BEGIN

//==============================================================================
// URLEncoder::Encode
//
/**
   Converts a Unicode string into the MIME @c x-www-form-urlencoded format.

   To convert a String, each Unicode character is examined in turn: 

   - The ASCII characters 'a' through 'z', 'A' through 'Z', '0' through '9', 
      and ".", "-", "*", "_" remain the same. 
   - The space character ' '(U+20) is converted into a plus sign '+'. 
   - All other characters are converted into their UTF-8 equivalent and the
     subsequent bytes are encoded as the 3-byte string "%xy", 
     where xy is the two-digit hexadecimal representation of the byte. 
*/
//==============================================================================
String URLEncoder::Encode(const String& uri)
{
	const char included[] = {'.', '-', '*', '_'};
	const char* pInclEnd = included + sizeof(included);

	String sRet;

	//
	// 1. Convert the URL to UTF-8.  Of course, it may already be encoded
	//    as UTF-8 if that is our internal string encoding.
	//
	ByteString utf8 = StringUtils::ToUTF8(uri);

	sRet.reserve(utf8.length());
	const Byte* pEnd =  (const Byte*)utf8.data() + utf8.length();

	for(const Byte* pByte=(const Byte*)utf8.data(); pByte<pEnd; ++pByte)
	{
		if((isalnum(*pByte)
			|| std::find(included, pInclEnd, *pByte) != pInclEnd))
		{
			sRet += CharType(*pByte);
		}
		else
		{
			//
			// 2. Escape any disallowed characters
			//
			char buffer[10];
			sprintf(buffer, "%%%02X", (unsigned)*pByte);
			sRet += StringUtils::FromLatin1(buffer);
		}
	}
	
	return sRet;
}

//==============================================================================
// URLEncoder::IdempotentEncode
//
/**
   Converts a Unicode string into an escaped form.

   This method performs a similar function to Encode(), but the algorithm
   has been modified to ignore certain escape characters thereby making
   it idempotent.

   The algorithm employed is described in the W3C
   <a href="http://www.w3.org/TR/REC-xml#sec-entity-decl">
   XML recommendation section 4.2.2, External Entities</a>.
*/
//
// URI references require encoding and escaping of certain characters.
// The disallowed characters include all non-ASCII characters, plus the
// excluded characters listed in Section 2.4 of [RFC 2396], except for
// the number sign (#) and percent sign (%) characters and the square bracket
// characters re-allowed in [RFC 2732].
//
// The excluded characters listed in Section 2.4 of IETF RFC 2396:
//
//  control: 0x00-0x1F and 0x7F
//  space: 0x20
//  delims: "<" | ">" | "#" | "%" | <">
//  unwise      = "{" | "}" | "|" | "\" | "^" | "[" | "]" | "`"
//
// 1) Each disallowed character is converted to UTF-8 as 1 or more bytes
// 2) Any bytes corresponding to a disallowed character are escaped
//    with the URI escaping mechanism (that is, converted to %HH, where
//    HH is the hexadecimal notation of the byte value)
// 3) The original character is replaced by the resulting character sequence
//
// Note that this normalization process is idempotent: repeated normalization
// does not change a normalized URI reference.
//==============================================================================
String URLEncoder::RawEncode(const String& uri)
{
	const char excluded[] = {'<', '>', '"', '{', '}', '|', '\\', '^', '\''};
	const char* pExclEnd = excluded + sizeof(excluded);

	String sRet;

	//
	// 1. Convert the URL to UTF-8.  Of course, it may already be encoded
	//    as UTF-8 if that is our internal string encoding.
	//
	ByteString utf8 = StringUtils::ToUTF8(uri);

	sRet.reserve(utf8.length());
	const Byte* pEnd =  (const Byte*)utf8.data() + utf8.length();

	for(const Byte* pByte=(const Byte*)utf8.data(); pByte<pEnd; ++pByte)
	{
		if((*pByte > 0x20 && *pByte < 0x7F)
			&& std::find(excluded, pExclEnd, *pByte) == pExclEnd)
		{
			sRet += CharType(*pByte);
		}
		else
		{
			//
			// 2. Escape any disallowed characters
			//
			char buffer[10];
			sprintf(buffer, "%%%02X", (unsigned)*pByte);
			sRet += StringUtils::FromLatin1(buffer);
		}
	}
	
	return sRet;
}

OT_NET_NAMESPACE_END
