//==============================================================================
// This file is part of the OpenTop C++ Library.
// (c) Copyright 2000-2002 ElCel Technology Limited
//
// This file may be distributed and/or modified under the terms of the
// GNU General Public License version 2 as published by the Free Software
// Foundation and appearing in the file LICENSE.GPL included in the
// packaging of this file.
//
// Licensees holding valid OpenTop Commercial licences may use this file
// in accordance with the OpenTop Commercial License Agreement included
// with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// See http://www.elcel.com/products/opentop/pricing.html or email
// sales@elcel.com for information about OpenTop Commercial License Agreements.
//
//==============================================================================
//
// $Revision: 1.3 $
// $Date: 2003/09/05 22:21:33 $
//
//==============================================================================
// Class XMLReaderFactory
/**
	@class ot::sax::XMLReaderFactory
	
	@brief Factory class for creating an XMLReader.
	
	<p>This class contains a static function for creating an XMLReader: CreateXMLReader().
	This function delegates the create request to the registered XMLReaderFactory instance
	thereby enabling the application to register a factory object and, in doing so,
	control the kind of XMLReader produced.</p>
	<p>If an instance of XMLReaderFactory is not registered by the application, an
	instance is created automatically the first time CreateXMLReader() or GetInstance()
	is called.</p>
	
    @code
    RefPtr<XMLReader> rpReader = XMLReaderFactory::CreateXMLReader();
    @endcode
*/
//==============================================================================

#include "XMLReaderFactory.h"
#include "SAXParser.h"

#include "ot/base/System.h"
#include "ot/base/ObjectManager.h"
#include "ot/base/FastMutex.h"

OT_SAX_NAMESPACE_BEGIN

//==================================================================
// Multi-threaded locking strategy
//
// XMLReaderFactory is a global object, created on demand.
// To ensure that resources are not created by multiple
// concurrent threads, the pointers are protected using a
// single static mutex.
//
// Update access to static variables is mutex protected, but to 
// minimise the runtime cost, read access is not protected.  This
// gives an exposure to the so-called "relaxed" memory model that
// exists on some multi-processor machines.  We minimise this
// exposure by declaring the static variables as 'volatile'.
//==================================================================
#ifdef OT_MT
	FastMutex XMLReaderFactoryMutex;
#endif //OT_MT


XMLReaderFactory* OT_MT_VOLATILE XMLReaderFactory::s_pInstance = NULL; // mutex protected

//==============================================================================
// XMLReaderFactory::GetInstance
//
/**
   Returns the currently registered XMLReaderFactory.  If a factory has not
   yet been registered, this method creates a default XMLReaderFactory and
   registers that.
   @mtsafe
*/
//==============================================================================
RefPtr<XMLReaderFactory> XMLReaderFactory::GetInstance()
{
	if(s_pInstance == NULL)
	{
		SetInstance(new XMLReaderFactory);
	}
	OT_DBG_ASSERT(s_pInstance!=0);
	return s_pInstance;
}

//==============================================================================
// XMLReaderFactory::SetInstance
//
/**
   Allows an application to register a custom XMLReaderFactory object
   which will be used to create instances of XMLReader.

   Like other @OpenTop global objects, XMLReaderFactory is a ManagedObject.
   This enables the passed object to be registered with the system's
   ObjectManager, thereby freeing the application from having to manage the lifetime
   of the factory object.  In other words, the application does not need to maintain
   the object's reference-count once it has been registered.

   A typical application may do the following:-

   @code
   XMLReaderFactory::SetInstance(new MyXMLReaderFactory);
   @endcode

   @param pFactory a pointer to an instance of XMLReaderFactory or null
          to unregister the currently registered factory.
   @mtsafe
*/
//==============================================================================
void XMLReaderFactory::SetInstance(XMLReaderFactory* pFactory)
{
	//==================================================================
	// Multi-threaded locking strategy
	//
	// See top of file.
	//==================================================================

	if(pFactory)
	{
		System::GetObjectManager().registerObject(pFactory);
	}

	XMLReaderFactory* pExisting;
	// create a scope for the mutex lock
	{
		OT_AUTO_LOCK(FastMutex, XMLReaderFactoryMutex);
		pExisting = s_pInstance;
		s_pInstance = pFactory;
	}

	if(pExisting) 
	{
		System::GetObjectManager().unregisterObject(pExisting);
	}
}

//==============================================================================
// XMLReaderFactory::CreateXMLReader
//
/**
   Static function to create an instance of an XMLReader.

   This static function simply delegates the request to the currently registered
   XMLReaderFactory object's doCreate() method.  If the application has not
   registered an instance of XMLReaderFactory, a default factory is registered
   automatically.

   @mtsafe
*/
//==============================================================================
RefPtr<XMLReader> XMLReaderFactory::CreateXMLReader()
{
	return GetInstance()->doCreate();
}

//==============================================================================
// XMLReaderFactory::doCreate
//
/**
   Creates an instance of XMLReader.

   An application may register a class @a derived from XMLReaderFactory
   to create whatever derivative of XMLReader it requires.
*/
//==============================================================================
RefPtr<XMLReader> XMLReaderFactory::doCreate() const
{
	return new SAXParser;
}

OT_SAX_NAMESPACE_END
