//==============================================================================
// This file is part of the OpenTop C++ Library.
// (c) Copyright 2000-2002 ElCel Technology Limited
//
// This file may be distributed and/or modified under the terms of the
// GNU General Public License version 2 as published by the Free Software
// Foundation and appearing in the file LICENSE.GPL included in the
// packaging of this file.
//
// Licensees holding valid OpenTop Commercial licences may use this file
// in accordance with the OpenTop Commercial License Agreement included
// with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// See http://www.elcel.com/products/opentop/pricing.html or email
// sales@elcel.com for information about OpenTop Commercial License Agreements.
//
//==============================================================================
//
// $Revision: 1.1.1.1 $
// $Date: 2002/12/14 00:25:51 $
//
//==============================================================================

#ifdef WIN32

#include "Win32Utils.h"

#include "ot/base/Win32Exception.h"
#include "ot/base/NumUtils.h"
#include "ot/base/NullPointerException.h"
#include "ot/base/StringUtils.h"
#include "ot/base/SystemUtils.h"
#include "ot/base/Tracer.h"
#include "ot/base/debug.h"

#include <stdlib.h>

OT_UTIL_NAMESPACE_BEGIN

//==============================================================================
// Win32Utils::ExpandFiles
//
// Given a file specification (eg *.xml), expand it to a list of filenanes
// using the Win32 functions: FindFirstFile/FindNextFile.
//==============================================================================
std::list<String> Win32Utils::ExpandFiles(const String& filespec)
{
	StringUtils::TCharAutoPtr pFilespec = StringUtils::ToWin32String(filespec);

	std::list<String> fileList;

	WIN32_FIND_DATA nextInfo;

	//
	// Find the first matching file.  An INVALID_HANDLE_VALUE is returned if no files match
	//
	HANDLE hContext = ::FindFirstFile(pFilespec.get(), &nextInfo);

	if(hContext == INVALID_HANDLE_VALUE)
	{
		throw Win32Exception(::GetLastError());
	}
	else
	{
		String dirPart;
		size_t dirPos = filespec.find_last_of(OT_T("/\\"));
		if(dirPos != String::npos)
		{
			dirPart = filespec.substr(0, dirPos);
		}

		String sSep = OT_T("\\");

		do
		{
			String file = StringUtils::FromWin32String(nextInfo.cFileName);
			static String sDot(OT_T("."));
			static String sDotDot(OT_T(".."));
			if(file != sDot && file != sDotDot)
			{
				if(!dirPart.empty())
				{
					file = dirPart + sSep + file;
				}
				fileList.push_back(file);
			}
		}
		while(::FindNextFile(hContext, &nextInfo));
		
		::FindClose(hContext);
	}
	fileList.sort();
	return fileList;
}

//==============================================================================
// Win32Utils::SystemTimeToDateTime
//
//==============================================================================
DateTime Win32Utils::SystemTimeToDateTime(SYSTEMTIME* const pSysTime)
{
	OT_DBG_ASSERT(pSysTime);
	if(!pSysTime) throw NullPointerException();

	return DateTime(pSysTime->wYear, pSysTime->wMonth, pSysTime->wDay,
	                pSysTime->wHour, pSysTime->wMinute, pSysTime->wSecond,
	                pSysTime->wMilliseconds);
}

//==============================================================================
// Win32Utils::DateTimeToSystemTime
//
//==============================================================================
void Win32Utils::DateTimeToSystemTime(const DateTime& time, SYSTEMTIME* pSysTime)
{
	OT_DBG_ASSERT(pSysTime);
	if(!pSysTime) throw NullPointerException();
	UShort year, month, day, dayOfWeek, hour, minute, second, millisecond;
	time.getDate(year, month, day, dayOfWeek);
	time.getTime(hour, minute, second, millisecond);
	pSysTime->wYear = year;
	pSysTime->wMonth = month;
	pSysTime->wDay = day;
	pSysTime->wDayOfWeek = dayOfWeek;
	pSysTime->wHour = hour;
	pSysTime->wMinute = minute;
	pSysTime->wSecond = second;
	pSysTime->wMilliseconds = millisecond;
}

//==============================================================================
// Win32Utils::TraceSystemCall
//
//==============================================================================
void Win32Utils::TraceSystemCall(short nSection, short nLevel, const String& message, DWORD errNo)
{
	static String sCode = OT_T(" rc=");
	static String sComma = OT_T(", ");
	if(errNo==0) errNo = ::GetLastError();

	String traceMsg = message + sCode 
	                  + NumUtils::ToString(errNo)
					  + sComma + SystemUtils::GetWin32ErrorString(errNo);

	Tracer::Trace(nSection, nLevel, traceMsg);
}

OT_UTIL_NAMESPACE_END
#endif //WIN32

