//==============================================================================
// This file is part of the OpenTop C++ Library.
// (c) Copyright 2000-2002 ElCel Technology Limited
//
// This file may be distributed and/or modified under the terms of the
// GNU General Public License version 2 as published by the Free Software
// Foundation and appearing in the file LICENSE.GPL included in the
// packaging of this file.
//
// Licensees holding valid OpenTop Commercial licences may use this file
// in accordance with the OpenTop Commercial License Agreement included
// with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// See http://www.elcel.com/products/opentop/pricing.html or email
// sales@elcel.com for information about OpenTop Commercial License Agreements.
//
//==============================================================================
//
// $Revision: 1.3 $
// $Date: 2003/09/05 22:21:33 $
//
//==============================================================================

#include "ParserFactory.h"
#include "XMLInputSource.h"
#include "ParserImpl.h"
#include "ot/base/System.h"
#include "ot/base/ObjectManager.h"
#include "ot/base/FastMutex.h"

OT_XML_NAMESPACE_BEGIN

//==================================================================
// Multi-threaded locking strategy
//
// ParserFactory is a singleton object, created on demand.
// To ensure that singleton resources are not created by multiple
// concurrent threads, the pointers are protexted using a
// single static mutex.
//
// To minimize the time that the mutex is held, we do
// not lock the mutex for the creation of a singleton object,
// only when a complete object has been created and we need to swap
// the address into our static pointer variable.
//==================================================================
#ifdef OT_MT
	FastMutex ParserFactoryMutex;
#endif //OT_MT

ParserFactory* OT_MT_VOLATILE ParserFactory::s_pInstance = 0; // mutex protected

ParserFactory::ParserFactory()
{
}

ParserFactory::~ParserFactory()
{
}

//==============================================================================
// ParserFactory::GetInstance
//
//==============================================================================
RefPtr<ParserFactory> ParserFactory::GetInstance()
{
	if(s_pInstance == NULL)
	{
		SetInstance(new ParserFactory);
	}
	OT_DBG_ASSERT(s_pInstance!=0);
	return s_pInstance;
}

//==============================================================================
// ParserFactory::SetInstance
//
//==============================================================================
void ParserFactory::SetInstance(ParserFactory* pFactory)
{
	//==================================================================
	// Multi-threaded locking strategy
	//
	// See top of file.
	//==================================================================

	if(pFactory)
	{
		System::GetObjectManager().registerObject(pFactory);
	}

	ParserFactory* pExisting;

	// create a scope for the mutex lock
	{
		OT_AUTO_LOCK(FastMutex, ParserFactoryMutex);
		pExisting = s_pInstance;
		s_pInstance = pFactory;
	}

	if(pExisting) 
	{
		System::GetObjectManager().unregisterObject(pExisting);
	}
}

//==============================================================================
// ParserFactory::CreateXMLParser
//
/**
   Creates a Parser by delegating the request to the currently registered
   ParserFactory object.
*/
//==============================================================================
RefPtr<Parser> ParserFactory::CreateXMLParser()
{
	return GetInstance()->doCreate();
}

//==============================================================================
// ParserFactory::doCreate
//
// Return an instance of an XML Parser
//==============================================================================
RefPtr<Parser> ParserFactory::doCreate() const
{
	return new ParserImpl();
}

OT_XML_NAMESPACE_END

