//==============================================================================
// This file is part of the OpenTop C++ Library.
// (c) Copyright 2000-2002 ElCel Technology Limited
//
// This file may be distributed and/or modified under the terms of the
// GNU General Public License version 2 as published by the Free Software
// Foundation and appearing in the file LICENSE.GPL included in the
// packaging of this file.
//
// Licensees holding valid OpenTop Commercial licences may use this file
// in accordance with the OpenTop Commercial License Agreement included
// with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// See http://www.elcel.com/products/opentop/pricing.html or email
// sales@elcel.com for information about OpenTop Commercial License Agreements.
//
//==============================================================================
//
// $Revision: 1.1.1.1 $
// $Date: 2002/12/14 00:27:03 $
//
//==============================================================================

#include "XMLFilterReader.h"

OT_XML_NAMESPACE_BEGIN

XMLFilterReader::XMLFilterReader(Reader* pReader) :
	FilterReader(pReader),
	m_bStandaloneCR(false)
{
}

XMLFilterReader::~XMLFilterReader()
{
}

//==============================================================================
// XMLFilterReader::read
//
// Call the underlying Reader's read method, and then normalize newlines.
//==============================================================================
IntType XMLFilterReader::read()
{
	IntType ret = FilterReader::read();

	if(ret == 0x0D)
	{
		ret = 0x0A;
		m_bStandaloneCR = true;
	}
	else if(ret == 0x0A && m_bStandaloneCR)
	{
		m_bStandaloneCR = false;
		ret = read();
	}

	return ret;
}

//==============================================================================
// XMLFilterReader::read
//
// Call the underlying Reader's read method, and then normalize newlines.
//==============================================================================
long XMLFilterReader::read(CharType* pBuffer, size_t bufLen)
{
	long charsRead=0;

	//
	// loop to guard against the situation where just one
	// character is returned and normalized away - resulting in
	// a potential invalid zero return.
	//
	while(charsRead==0)
	{
		charsRead = FilterReader::read(pBuffer, bufLen);
		
		if(charsRead > 0)
		{
			charsRead = normalizeNewLines(pBuffer, charsRead);
		}
		else
		{
			break;
		}
	}
	return charsRead;
}

//==============================================================================
// XMLFilterReader::readAtomic
//
//==============================================================================
long XMLFilterReader::readAtomic(CharType* pBuffer, size_t bufLen)
{
	long charsRead=0;

	//
	// loop to guard against the situation where just one
	// character is returned and normalized away - resulting in
	// a potential invalid zero return.
	//
	while(charsRead==0)
	{
		charsRead = FilterReader::readAtomic(pBuffer, bufLen);
		if(charsRead > 0)
		{
			charsRead = normalizeNewLines(pBuffer, charsRead);
		}
		else
		{
			break;
		}
	}
	return charsRead;
}

Character XMLFilterReader::readAtomic()
{
	static Character newLine('\n');
	Character ret = FilterReader::readAtomic();
	UCS4Char ch = ret.toUnicode();

	if(ch == 0x0D)
	{
		ret = newLine;
		m_bStandaloneCR = true;
	}
	else if(ch == 0x0A && m_bStandaloneCR)
	{
		m_bStandaloneCR = false;
		ret = readAtomic();
	}

	return ret;
}

//==============================================================================
// XMLFilterReader::normalizeNewLines
//
// Normalize end-of-line characters as per the W3 XML 1.0 recommendation
// section 2.11
//
// Replace the 2-character sequence "0x0D" "0x0A" with "0x0A" and
// standalone "0x0D" with "0x0A"
//
//==============================================================================
long XMLFilterReader::normalizeNewLines(CharType* pBuffer, size_t bufLen)
{
	CharType* pEnd = pBuffer + bufLen;
	CharType* pNextOut = pBuffer;

	for(CharType* pNextIn = pBuffer; pNextIn < pEnd; ++pNextIn)
	{
		const CharType nextChar = *pNextIn;

		if(nextChar == 0x0D)
		{
			*pNextOut++ = 0x0A;
			//
			// We have found a 0x0D (carriage return to you and me)
			// which we have silently converted to a newline
			// If a newline follows immediately afterwards then we want to
			// delete it.  If we are before the last character then we can
			// simply check the buffer, otherwise we need to get another.
			//
			if(pNextIn+1 < pEnd)
			{
				if(*(pNextIn+1) == 0x0A)
				{
					// next character is effectively deleted simply by incrementing ptr
					// so that it is not copied like all the rest
					pNextIn++;
				}
			}
			else
			{
				//
				// We are in the situation where the last character in the 
				// buffer was a 0x0D (cr).  We have changed it into a
				// newline (0x0A) but we need to know what comes next to see
				// if the next character is a newline which should be silently
				// dropped
				//
				// To do this we just set a marker which is tested next time
				// we do a normalization
				
				m_bStandaloneCR = true;
				break;
			}
		}
		else if(!(nextChar == 0x0A && m_bStandaloneCR))
		{
			*pNextOut++ = nextChar;
		}

		m_bStandaloneCR = false;
	}
	
	return (pNextOut - pBuffer);
}
	

OT_XML_NAMESPACE_END
