//==============================================================================
// This file is part of the OpenTop C++ Library.
// (c) Copyright 2000-2002 ElCel Technology Limited
//
// This file may be distributed and/or modified under the terms of the
// GNU General Public License version 2 as published by the Free Software
// Foundation and appearing in the file LICENSE.GPL included in the
// packaging of this file.
//
// Licensees holding valid OpenTop Commercial licences may use this file
// in accordance with the OpenTop Commercial License Agreement included
// with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// See http://www.elcel.com/products/opentop/pricing.html or email
// sales@elcel.com for information about OpenTop Commercial License Agreements.
//
//==============================================================================
//
// $Revision: 1.1.1.1 $
// $Date: 2002/12/14 00:27:07 $
//
//==============================================================================

#include "CatalogSet.h"
#include "CatalogFile.h"
#include "CatalogResolver.h"
#include "CatalogEventHandler.h"

OT_XMLCAT_NAMESPACE_BEGIN

CatalogSet::CatalogSet(const CatalogResolver& resolver) :
	m_resolver(resolver)
{
}

//==============================================================================
// CatalogSet::resolveEntity
//
// Iterate thru all the CatalogFiles attempting to find a match.  As soon as a 
// positive match is found, or the request has been delegated to another
// CatalogSet, we quit.
//==============================================================================
bool CatalogSet::resolveEntity(const String& publicId, const String& systemId,
                               const CatalogFile::List& catList,
                               String& returnUri,
							   bool& bReturnDelegated) const
{
	bReturnDelegated = false;
	CatalogFiles::const_iterator i=m_catalogFiles.begin();
	for(; i!=m_catalogFiles.end() && !bReturnDelegated; ++i)
	{
		const RefPtr<CatalogFile>& rpCatalogFile = *i;

		if(rpCatalogFile->resolveEntity(publicId, systemId, catList,
		                                returnUri, bReturnDelegated))
		{
			return true;
		}
	}
	return false;
}

//==============================================================================
// CatalogSet::resolveURI
//
// Iterate thru all the CatalogSet parts (the CatalogSet list in the parlance
// of the Oasis specification) attempting to find a match.  As soon as a 
// positive match is found, or the request has been delegated to another
// CatalogSet, we quit.
//==============================================================================
bool CatalogSet::resolveURI(const String& uri, const CatalogFile::List& catList,
                            String& returnUri,
							bool& bReturnDelegated) const
{
	bReturnDelegated = false;
	CatalogFiles::const_iterator i=m_catalogFiles.begin();
	for(; i!=m_catalogFiles.end() && !bReturnDelegated; ++i)
	{
		const RefPtr<CatalogFile>& rpCatalogFile = *i;
		if(rpCatalogFile->resolveURI(uri, catList, returnUri, bReturnDelegated))
		{
			return true;
		}
	}

	return false;
}
	
//==============================================================================
// CatalogSet::addFile
//
//==============================================================================
void CatalogSet::addFile(const URL& catalogFile, const CatalogFile* pFrom)
{
	//
	// Test if the file already exists in our set.  If so, it 
	// cannot be allowed in because it would imply one of
	// a) A duplicate nextCatalog element
	// b) A circular nextCatalog reference
	//

	CatalogFiles::const_iterator i=m_catalogFiles.begin();
	for(; i!=m_catalogFiles.end(); ++i)
	{
		const RefPtr<CatalogFile>& rpCatalogFile = *i;
		if(rpCatalogFile->getURL().sameFile(catalogFile))
		{
			// Error! duplicate
			CatalogEventHandler* pHandler = getResolver().getEventHandler();
			if(pHandler)
			{
				String errMsg = OT_T("the catalog entry file: ");
				errMsg += catalogFile.toExternalForm();
				errMsg += OT_T(" is already contained in the set of catalog files");
				String fileName;
				if(pFrom)
				{
					fileName = pFrom->getURL().getFile();
				}
				pHandler->onCatalogError(CatalogEventHandler::Error, 0, errMsg, fileName, 0, 0);
			}
			return;
		}
	}

	RefPtr<CatalogFile> rpNewFile(new CatalogFile(catalogFile, *this));
	m_catalogFiles.push_back(rpNewFile);
}

//==============================================================================
// CatalogSet::getResolver
//
//==============================================================================
const CatalogResolver& CatalogSet::getResolver() const
{
	return m_resolver;
}

OT_XMLCAT_NAMESPACE_END
