//==============================================================================
// This file is part of the OpenTop C++ Library.
// (c) Copyright 2000-2003 ElCel Technology Limited
//
// This file may be distributed and/or modified under the terms of the
// GNU General Public License version 2 as published by the Free Software
// Foundation and appearing in the file LICENSE.GPL included in the
// packaging of this file.
//
// Licensees holding valid OpenTop Commercial licences may use this file
// in accordance with the OpenTop Commercial License Agreement included
// with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// See http://www.elcel.com/products/opentop/pricing.html or email
// sales@elcel.com for information about OpenTop Commercial License Agreements.
//
//==============================================================================
//
// $Revision: 1.2 $
// $Date: 2003/09/08 21:00:14 $
//
//==============================================================================
//
// SAX test driver program
//
//==============================================================================

#include "SAXCountHandler.h"

#include "ot/auxil/BasicOption.h"
#include "ot/auxil/CommandLineParser.h"
#include "ot/auxil/MemCheckSystemMonitor.h"
#include "ot/base/Exception.h"
#include "ot/io/Console.h"
#include "ot/io/FileInputStream.h"
#include "ot/io/FileDescriptor.h"
#include "ot/net/URL.h"
#include "ot/sax/XMLReader.h"
#include "ot/sax/XMLReaderFactory.h"
#include "ot/sax/SAXFeatures.h"

using namespace ot;
using namespace auxil;
using namespace sax;
using namespace io;

void showUsage(const String& programName)
{
	
	Console::Out()->print(OT_T("Usage: "));
	Console::Out()->println(programName + OT_T(" [option]... [file]..."));
	Console::Out()->println();

	Console::Out()->println(OT_T("  -e, --noextentities disables the reading of external entities"));
	Console::Out()->println(OT_T("  -h, --help          display this message and terminate"));
	Console::Out()->println(OT_T("  -v, --validate      perform validation tests on the input file"));
	Console::Out()->println(OT_T("  -w, --warnings      perform additional warning tests"));
	Console::Out()->println();
	Console::Out()->println(OT_T("  If file is '-' then read from standard input"));
}

int main(int argc, char* argv[]) 
{
	MemCheckSystemMonitor monitor;

	//
	// Parse the command line for options and filenames
	//
	CommandLineParser cmdlineParser;

	BasicOption optNoExt(OT_T("noextentities"), 'e', BasicOption::none);
	BasicOption optHelp(OT_T("help"), 'h', BasicOption::none);
	BasicOption optValidate(OT_T("validate"), 'v', BasicOption::none);
	BasicOption optWarnings(OT_T("warnings"), 'w', BasicOption::none);

	cmdlineParser.addOption(&optNoExt);
	cmdlineParser.addOption(&optHelp);
	cmdlineParser.addOption(&optValidate);
	cmdlineParser.addOption(&optWarnings);

	try
	{
		cmdlineParser.parse(argc, argv);
	}
	catch (CommandLineException& e)
	{
		Console::Err()->println(cmdlineParser.getProgramName() + String(OT_T(": ")) + e.getMessage());
		Console::Err()->print(OT_T("Try "));
		Console::Err()->println(cmdlineParser.getProgramName() + OT_T(" --help"));
		return (1);
	}

	if(optHelp.isPresent())
	{
		showUsage(cmdlineParser.getProgramName());
		return (0);
	}

	RefPtr<XMLReader> rpXMLReader = XMLReaderFactory::CreateXMLReader();
	RefPtr<DefaultHandler> rpHandler = new SAXCountHandler();

	rpXMLReader->setErrorHandler(rpHandler.get());
	rpXMLReader->setContentHandler(rpHandler.get());

	bool bSuccess = true;

	//==========================================================================
	// Set the required features of the SAX Parser
	//==========================================================================
	bool bValidate = optValidate.isPresent();
	bool bEntities = (!optNoExt.isPresent()) || bValidate;
	rpXMLReader->setFeature(SAXFeatures::namespaces, true);
	rpXMLReader->setFeature(SAXFeatures::validation, optValidate.isPresent());
	rpXMLReader->setFeature(SAXFeatures::external_parameter_entities, bEntities);
	rpXMLReader->setFeature(SAXFeatures::external_general_entities, bEntities);
	rpXMLReader->setFeature(SAXFeatures::warnings, optWarnings.isPresent());

	std::list<String> files = cmdlineParser.getFilenames(argc, argv);

	if(!files.empty())
	{
		for(std::list<String>::const_iterator i=files.begin(); i!=files.end(); ++i)
		{
			const String& uri = *i;

			try
			{
				if(uri == OT_T("-"))
				{
					RefPtr<InputStream> rpInputStream = new FileInputStream(FileDescriptor::In().get());
					RefPtr<InputSource> rpInputSource = new InputSource(rpInputStream.get());
					rpInputSource->setSystemId(OT_T("stdin"));
					rpXMLReader->parse(rpInputSource.get());
				}
				else
				{
					rpXMLReader->parse(uri);
				}
			}
			catch(Exception& e)
			{
				Console::Err()->println(uri + String(OT_T(" : Fatal error: ")) + e.getMessage());
				bSuccess = false;
			}
		}
	}
	else 
	{
		showUsage(cmdlineParser.getProgramName());
	}
	
	return (bSuccess ? 0 : 1);
}

