//==============================================================================
// This file is part of the OpenTop C++ Library.
// (c) Copyright 2000-2002 ElCel Technology Limited
//
// This file may be distributed and/or modified under the terms of the
// GNU General Public License version 2 as published by the Free Software
// Foundation and appearing in the file LICENSE.GPL included in the
// packaging of this file.
//
// Licensees holding valid OpenTop Commercial licences may use this file
// in accordance with the OpenTop Commercial License Agreement included
// with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// See http://www.elcel.com/products/opentop/pricing.html or email
// sales@elcel.com for information about OpenTop Commercial License Agreements.
//
//==============================================================================
//
// $Revision: 1.3 $
// $Date: 2003/09/09 21:14:56 $
//
//==============================================================================
//
// inetcopy Command-line Driver
// 
//==============================================================================

#include "ot/base/NumUtils.h"
#include "ot/base/StringUtils.h"
#include "ot/base/Exception.h"
#include "ot/base/System.h"
#include "ot/io/Console.h"
#include "ot/io/FileNotFoundException.h"
#include "ot/io/FileOutputStream.h"
#include "ot/io/FileDescriptor.h"
#include "ot/io/File.h"
#include "ot/net/URL.h"
#include "ot/net/URLConnection.h"
#include "ot/net/MalformedURLException.h"
#include "ot/auxil/BooleanOption.h"
#include "ot/auxil/CommandLineParser.h"
#include "ot/auxil/FileMessageFactory.h"
#include "ot/auxil/MemCheckSystemMonitor.h"
#include "ot/auxil/NetAccessHelper.h"
#include "ot/auxil/TraceHelper.h"

using namespace ot;
using namespace ot::net;
using namespace ot::io;
using namespace ot::auxil;

const CharType* sProgramTitle = OT_T("Internet Copy (inetcopy)");
const CharType* sCopyright = OT_T("Copyright (c) 2002 - ElCel Technology Limited (http://www.elcel.com)");
const String sHTTP = OT_T("http://");
const String sInetCopy = OT_T("inetcopy");

void showVersion(const String& /*programName*/)
{
	String sVersion = System::GetVersionAsString();
	Console::Out()->println(sProgramTitle + String(OT_T(" using OpenTop version ")) + sVersion);
	Console::Out()->println();
	Console::Out()->println(sCopyright);
}

void showUsage(const String& programName)
{
	Console::Out()->println(String(OT_T("Usage: ")) + programName + OT_T(" [option]... URL file"));
	Console::Out()->println(String(OT_T("       ")) + programName + OT_T(" [option]... URL ... directory"));
	Console::Out()->println();
	Console::Out()->println(OT_T("Copies URL resource to local file(s) or standard output."));
	Console::Out()->println();
	Console::Out()->println(OT_T("  -h, --help                display this help"));
	Console::Out()->println(OT_T("  -n, --nocopy              do not copy the URL, useful for testing a URL"));
	Console::Out()->println(OT_T("                            without downloading the data"));
	Console::Out()->println(OT_T("  -H, --headers             display protocol header information"));
	Console::Out()->println(OT_T("  -p, --password PASS       HTTP password"));
	Console::Out()->println(OT_T("  -P, --proxypassword PASS  password for proxy server authentication"));
	Console::Out()->println(OT_T("  -q, --quiet               do not display progress information messages"));
	Console::Out()->println(OT_T("  --httpproxy SERVER[:PORT] HTTP proxy server.  If the PORT number is not "));
	Console::Out()->println(OT_T("                            specified then port 8080 is used by default"));
	Console::Out()->println(OT_T("  -r, --restart             attempt to restart a previous transfer"));
	Console::Out()->println(OT_T("  -u, --user USER           HTTP userid"));
	Console::Out()->println(OT_T("  -U, --proxyuser USER      userid for proxy server authentication"));
	Console::Out()->println(OT_T("  -v, --verbose             display additional information"));
	Console::Out()->println(OT_T("  -V, --version             display version number and exit"));
	Console::Out()->println();
	Console::Out()->println(OT_T("  If file is '-' then echo the file to standard output"));
	Console::Out()->println(OT_T("Return codes: 0=success, 1=error"));
	Console::Out()->println(sCopyright);
}

int main(int argc, char* argv[])
{
	//
	// Facilitate clean termination and memory leak detection by creating
	// a MemCheckSystemMonitor on the stack.  The destructor will ensure
	// clean termination even in the event of an uncaught exception.
	//
	MemCheckSystemMonitor monitor;

	//
	// Install I18N messages if the appropriate environment is present
	//
	FileMessageFactory::Install();

	// 
	// Use the OpenTop CommandLineParser to provide POSIX-compliant command-line
	// handling.
	//
	CommandLineParser cmdlineParser;

	//
	// Create command option objects for all the recognised command line flags
	//
	BasicOption optNoCopy(OT_T("nocopy"), 'n', BasicOption::none);
	BasicOption optHeaders(OT_T("headers"), 'H', BasicOption::none);
	BasicOption optHttpProxy(OT_T("httpproxy"), 0, BasicOption::mandatory);
	BasicOption optHelp(OT_T("help"), 'h', BasicOption::none);
	BasicOption optProxyPassword(OT_T("proxypassword"), 'P', BasicOption::mandatory);
	BasicOption optPassword(OT_T("password"), 'p', BasicOption::mandatory);
	BasicOption optRestart(OT_T("restart"), 'r', BasicOption::none);
	BasicOption optTrace(OT_T("trace"), 0, BasicOption::mandatory);
	BasicOption optUserid(OT_T("userid"), 'u', BasicOption::mandatory);
	BasicOption optProxyUserid(OT_T("proxyuserid"), 'U', BasicOption::mandatory);
	BasicOption optQuiet(OT_T("quiet"), 'q', BasicOption::none);
	BasicOption optVerbose(OT_T("verbose"), 'v', BasicOption::none);
	BasicOption optVersion(OT_T("version"), 'V', BasicOption::none);

	//
	// Add the option objects to the CommandLineParser...
	//
	cmdlineParser.addOption(&optNoCopy);
	cmdlineParser.addOption(&optHeaders);
	cmdlineParser.addOption(&optHttpProxy);
	cmdlineParser.addOption(&optHelp);
	cmdlineParser.addOption(&optPassword);
	cmdlineParser.addOption(&optProxyPassword);
	cmdlineParser.addOption(&optProxyUserid);
	cmdlineParser.addOption(&optQuiet);
	cmdlineParser.addOption(&optRestart);
	cmdlineParser.addOption(&optTrace);
	cmdlineParser.addOption(&optUserid);
	cmdlineParser.addOption(&optVerbose);
	cmdlineParser.addOption(&optVersion);

	//
	// ...and then parse the command line, setting the option values into
	// their respecting command option objects.
	//
	try
	{
		cmdlineParser.parse(argc, argv);
	}
	catch (CommandLineException& e)
	{
		Console::Err()->println(cmdlineParser.getProgramName() + String(OT_T(": ")) + e.getMessage());
		Console::Err()->println();
		Console::Err()->println(String(OT_T("Try ")) + cmdlineParser.getProgramName() + OT_T(" --help"));
		return (1);
	}

	const bool bQuiet = optQuiet.isPresent();
	const bool bVerbose = optVerbose.isPresent();
	const bool bNoCopy = optNoCopy.isPresent();
	const bool bRestart = optRestart.isPresent();

	int firstPositional = cmdlineParser.getFirstPositionalArg();

	if(optHelp.isPresent() || firstPositional == argc)
	{
		showUsage(cmdlineParser.getProgramName());
		return (0);
	}

	if(optVersion.isPresent())
	{
		showVersion(cmdlineParser.getProgramName());
		return (0);
	}

	//
	// Use a nice and helpful NetAccessHelper to set up the boiler-plate
	// stuff for proxies, userids, passwords etc.  This is necessary
	// for HTTP access to resources that require authentication or must be accessed
	// via a proxy.
	//
	RefPtr<NetAccessHelper> rpNetHelper = new NetAccessHelper(
		optHttpProxy.getArgument(),
		optProxyUserid.getArgument(),
		optProxyPassword.getArgument(),
		optUserid.getArgument(),
		optPassword.getArgument());
	Authenticator::SetDefault(rpNetHelper.get());
	
	const int lastPositional = (bNoCopy ? argc-1 : argc-2);

	String dest;
	bool bStandardOut = false;
	bool bDestDirectory = false;

	if(!bNoCopy)
	{
		dest = StringUtils::FromNativeMBCS(argv[lastPositional+1]);
		bStandardOut = (dest == OT_T("-"));
		if(!bStandardOut)
		{
			File destDirectory(dest);
			bDestDirectory = destDirectory.isDirectory();
		}
	}

	if(lastPositional<firstPositional)
	{
		showUsage(cmdlineParser.getProgramName());
		return (0);
	}
	else if(lastPositional>firstPositional && !bDestDirectory)
	{
		Console::Err()->println(dest + OT_T(" is not a directory"));
		return (0);
	}

	URL contextUrl;

	try
	{
		//
		// Use a TraceHelper from the auxil namespace to set up a standard tracing output
		//
		if(optTrace.isPresent())
		{
			Tracer::SetTracer(new TraceHelper(sInetCopy, optTrace.getArgument()));
		}

		//
		// The CommandLineParser lets us know the first positional argument
		// (ie the first non-flag argument).  We process each one of these
		// in turn, treating all but the last as a URL that locates a resource
		// to be copied.  The last paremeter designates the destination.  This
		// may be a local file or directory.  If multiple source URLs are 
		// specified, the destination must refer to a directory.
		//
		// Note: It would be reasonably straight forward t modify this program
		// to allow the destination to be a URL also.  This is left as an
		// exercise for the reader!
		//
		for(int i=firstPositional; i<=lastPositional; i++)
		{
			//
			// Command line parameters are translated into Unicode strings
			// before being interpreted as URLs
			//
			String spec = StringUtils::FromNativeMBCS(argv[i]);
			if(bVerbose)
			{
				Console::Out()->println(String(OT_T("processing URL: ")) + spec);
			}

			URL url;

			try
			{
				url = URL(contextUrl, spec);
			}
			catch(MalformedURLException& /*e*/)
			{
				url = URL(sHTTP + spec);
			}

			contextUrl = url;

			RefPtr<URLConnection> rpUrlConn = url.openConnection();

			//
			// If this is a restart request, find out the binary size of
			// the local file
			//
			size_t fileSize=0;
			if(!bStandardOut && bRestart)
			{
				if(bDestDirectory)
				{
					String fileName = url.getFile();
					if(fileName.empty())
						fileName = url.getHost();
					File local(fileName);
					fileName = local.getName();
					File newFile(dest, fileName);
					try
					{
						fileSize = newFile.length();
					}
					catch(FileNotFoundException& /*e*/) {}
				}
				else
				{
					File newFile(dest);
					try
					{
						fileSize = newFile.length();
					}
					catch(FileNotFoundException& /*e*/) {}

				}
				if(fileSize)
				{
					rpUrlConn->setRequestProperty(OT_T("Range"), String(OT_T("bytes=")) + NumUtils::ToString(fileSize) + OT_T("-"));
				}
			}



			rpUrlConn->connect();

			if(bVerbose)
			{
				Console::Out()->println(OT_T("connected to URL: "));
				Console::Out()->println(rpUrlConn->getURL().toExternalForm());
				DateTime fileDate = rpUrlConn->getLastModified();
				if(fileDate.isValid())
				{
					Console::Out()->println(String(OT_T("File last modified: ")) + fileDate.toString());
				}
			}

			if(optHeaders.isPresent())
			{
				for(size_t i=0; i<rpUrlConn->getHeaderFieldCount(); i++)
				{
					String hdr = rpUrlConn->getHeaderField(i);
					if(rpUrlConn->getHeaderFieldKey(i).empty())
						Console::Out()->print(OT_T("null"));
					else
						Console::Out()->print(rpUrlConn->getHeaderFieldKey(i));

					Console::Out()->print(OT_T(": "));
					Console::Out()->println(hdr);
				}
			}

			if(!optNoCopy.isPresent())
			{
				RefPtr<OutputStream> rpOStream;
				String fileName;

				bool bAppend = (bRestart && fileSize > 0);

				if(bStandardOut)
				{
					rpOStream = new FileOutputStream(FileDescriptor::Out().get());
				}
				else if(bDestDirectory)
				{
					fileName = url.getFile();
					if(fileName.empty())
					{
						fileName = url.getHost();
					}
					File local(fileName);
					fileName = local.getName();
					File newFile(dest, fileName);
					rpOStream = new FileOutputStream(newFile.getPath(), bAppend);
				}
				else
				{
					rpOStream = new FileOutputStream(dest, bAppend);
				}


				RefPtr<InputStream> rpIStream = rpUrlConn->getInputStream();
				Byte buffer[2048];
				size_t byteCount = 0;
				while(true)
				{
					long bytesRead = rpIStream->read(buffer, sizeof(buffer));
					if(bytesRead > 0)
					{
						byteCount += bytesRead;
						rpOStream->write(buffer, bytesRead);
						
						//
						// print hatches '#' if we are not redirecting
						// the output to std out
						//
						if(bVerbose && !bStandardOut)
						{
							Console::Out()->print(OT_T("#"));
							Console::Out()->flush();
						}
					}
					else if(bytesRead == InputStream::EndOfFile)
					{
						break;
					}
					else
					{
						OT_DBG_ASSERT(false);
					}
				}
				// clean up after printing all those hatches
				if(bVerbose && !bStandardOut)
				{
					Console::Out()->println();
				}

				// print byte count
				if(!bQuiet && !bStandardOut)
				{
					Console::Out()->print(byteCount);
					Console::Out()->print(OT_T(" bytes copied from "));
					Console::Out()->print(url.toExternalForm());
					Console::Out()->println(String(OT_T(" to ")) + fileName);
				}
			} //if(!optNoCopy.isPresent())
		} //while
	}
	catch(Exception& e)
	{
		Console::Err()->println(e.toString());
	}

	return 0;
}

