//==============================================================================
// This file is part of the OpenTop C++ Library.
// (c) Copyright 2000-2003 ElCel Technology Limited
//
// This file may be distributed and/or modified under the terms of the
// GNU General Public License version 2 as published by the Free Software
// Foundation and appearing in the file LICENSE.GPL included in the
// packaging of this file.
//
// Licensees holding valid OpenTop Commercial licences may use this file
// in accordance with the OpenTop Commercial License Agreement included
// with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// See http://www.elcel.com/products/opentop/pricing.html or email
// sales@elcel.com for information about OpenTop Commercial License Agreements.
//
//==============================================================================
//
// $Revision: 1.3 $
// $Date: 2003/09/08 21:00:14 $
//
//==============================================================================
//
// OpenTop Sample Application: tcpclient
//
// This simple console application will initiate a TCP connection to a specified
// host/port and send it a "Hello World!" message.  It will then (optionally)
// wait for a reply and echo this out to the console.
//
//==============================================================================

#include "ot/net/InetAddress.h"
#include "ot/net/ServerSocket.h"
#include "ot/net/Socket.h"
#include "ot/base/NumUtils.h"
#include "ot/base/Exception.h"
#include "ot/base/StringUtils.h"
#include "ot/base/System.h"
#include "ot/base/Tracer.h"
#include "ot/base/Thread.h"
#include "ot/io/Console.h"
#include "ot/auxil/MemCheckSystemMonitor.h"
#include "ot/auxil/CommandLineParser.h"
#include "ot/auxil/BooleanOption.h"
#include "ot/auxil/TraceHelper.h"

#ifndef _MSC_VER
#include <signal.h>
#endif

using namespace ot;
using namespace ot::net;
using namespace ot::io;
using namespace ot::auxil;

bool bContinue = true;
const size_t ClientTimeout = 20000; // 20 secs

#define COUT Console::cout()
#define CERR Console::cerr()

void showUsage(const String& programName)
{
	COUT << OT_T("Usage: ") << programName << OT_T(" [option]... ") << endl << endl;
	COUT << OT_T("Test server.") << endl << endl;

	COUT << OT_T("  -h, --help           display this help") << endl;
	COUT << OT_T("  -l, --linger <secs>  enable the SO_LINGER option") << endl;
	COUT << OT_T("  -p, --port           port to connect to") << endl;
	COUT << OT_T("  -p, --reply          wait for a reply (and echo it to standard out)") << endl;
	COUT << OT_T("  -s, --server         server to connect to") << endl;
	COUT << OT_T("  -t, --timeout        milliseconds to wait for connection") << endl;
	COUT << OT_T("Copyright (c) 2003 - ElCel Technology Limited (http://www.elcel.com)") << endl;
}

int main(int argc, char* argv[])
{
	MemCheckSystemMonitor monitor;

	//
	// Set up a command line parser to recognise our options...
	// 
	BasicOption optHelp(OT_T("help"), 'h', BasicOption::none);
	BasicOption optLinger(OT_T("linger"), 'l', BasicOption::mandatory);
	BasicOption optPort(OT_T("port"), 'p', BasicOption::mandatory);
	BooleanOption optReply(OT_T("reply"), 'r', false);
	BasicOption optServer(OT_T("server"), 's', BasicOption::mandatory);
	BasicOption optTimeout(OT_T("timeout"), 't', BasicOption::mandatory);
	BasicOption optTrace(OT_T("trace"), 0, BasicOption::mandatory);

	CommandLineParser cmdlineParser;
	cmdlineParser.addOption(&optHelp);
	cmdlineParser.addOption(&optLinger);
	cmdlineParser.addOption(&optPort);
	cmdlineParser.addOption(&optReply);
	cmdlineParser.addOption(&optServer);
	cmdlineParser.addOption(&optTimeout);
	cmdlineParser.addOption(&optTrace);

	//
	// ...parse the cpmmand line, placing the command options into their 
	// respective objects.
	//
	try
	{
		cmdlineParser.parse(argc, argv);
	}
	catch (CommandLineException& e)
	{
		CERR << cmdlineParser.getProgramName() << OT_T(": ") << e.getMessage() << endl << endl;
		CERR << OT_T("Try ") << cmdlineParser.getProgramName() << OT_T(" --help") << endl;
		return (1);
	}

	//
	// Determine local host info
	//
	RefPtr<InetAddress> rpLocal = InetAddress::GetLocalHost();
	COUT << OT_T("Client running on host ") << rpLocal->toString() << endl;

	if(optHelp.isPresent())
	{
		showUsage(cmdlineParser.getProgramName());
		return (0);
	}

	//
	// Determine the host and port to which we shall try and connect
	//
	int port = 0;
	if(optPort.isPresent())
	{
		port = NumUtils::ToInt(optPort.getArgument());
	}

	String host = OT_T("localhost");
	if(optServer.isPresent())
	{
		host = optServer.getArgument();
	}

	try
	{
		//
		// If we have been asked to enable tracing, create a TraceHelper (from the auxil namespace)
		// This little helper is initialized with a string from which it extracts what tracing
		// should be performed, and to where the output should be written.
		//
		if(optTrace.isPresent())
		{
			Tracer::SetTracer(new TraceHelper(cmdlineParser.getProgramName(), optTrace.getArgument()));
		}

		COUT << OT_T("Connecting to server: ") << host << OT_T(" on port: ") << port << endl;

		//
		// Here we demonstate two methods of connecting to a server.  
		//
		// The first method uses an overloaded socket constructor to specify the host and port
		// and attempts to connect immediately.
		//
		// The second method uses the default constructor, enabling us to set a timeout for the 
		// connect method.
		//
		RefPtr<Socket> rpSocket;
		
		size_t timeoutMS = 0;
		if(optTimeout.isPresent())
			timeoutMS = NumUtils::ToInt(optTimeout.getArgument());

		if(timeoutMS)
		{
			// as a timeout has been specified, we must use the "construct then connect" model
			rpSocket = new Socket;

			// Note that connect() is not overloaded to take a host name and a timeout value,
			// so we must obtain an InetAddress for the host.

			RefPtr<InetAddress> rpAddr = InetAddress::GetByName(host);
			rpSocket->connect(rpAddr.get(), port, timeoutMS);
		}
		else
		{
			rpSocket = new Socket(host, port);
		}

		//
		// If the --linger option has been specified, use its (mandatory) value
		// to set the SO_LINGER option
		//
		if(optLinger.isPresent())
		{
			int linger = NumUtils::ToInt(optLinger.getArgument());
			rpSocket->setSoLinger(true, linger);
		}

		//
		// Send a simple message to the server.
		// Note we are sending raw bytes in this case, not Unicode characters.
		//
		const char * hello = "Hello World!";
		rpSocket->getOutputStream()->write((const Byte*)hello, strlen(hello));

		//
		// If a reply is expected, then wait for it - but use the timeout value
		// to limit the length of time we'll wait.
		//
		if(optReply.getValue())
		{
			if(timeoutMS)
				rpSocket->setSoTimeout(timeoutMS);

			Byte buffer[80];
			long bytesRead;
			
			COUT << OT_T("Reply from server: ") ;

			while( (bytesRead = rpSocket->getInputStream()->read(buffer, sizeof(buffer))) != InputStream::EndOfFile)
			{
				COUT << StringUtils::FromLatin1((const char*)buffer, bytesRead);
			}

			COUT << endl;
		}
	}
	catch(Exception& e)
	{
		CERR << e.toString() << endl;
	}

	COUT << OT_T("Terminating client") << endl;

	return 0;
}
