//==============================================================================
// This file is part of the OpenTop C++ Library.
// (c) Copyright 2000-2003 ElCel Technology Limited
//
// This file may be distributed and/or modified under the terms of the
// GNU General Public License version 2 as published by the Free Software
// Foundation and appearing in the file LICENSE.GPL included in the
// packaging of this file.
//
// Licensees holding valid OpenTop Commercial licences may use this file
// in accordance with the OpenTop Commercial License Agreement included
// with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// See http://www.elcel.com/products/opentop/pricing.html or email
// sales@elcel.com for information about OpenTop Commercial License Agreements.
//
//==============================================================================
//
// $Revision: 1.2 $
// $Date: 2003/09/08 21:00:15 $
//
//==============================================================================
//
// tcpserver sample application
//
// Class ClientHandler
// -------------------
//
// An instance of this class is created whenever a new TCP socket connection
// is established with a client.  
// 
// This class handles all communication with the client over the socket.  It
// will run either synchronously or concurrently depending on the threading
// model.  In single-threaded mode the run() method is executed within the
// context of the main thread, so control is not returned to the main program
// until the client connection terminates.
//
//==============================================================================

#include "ClientHandler.h"
#include "Listener.h"

#include "ot/base/String.h"
#include "ot/base/StringUtils.h"
#include "ot/base/System.h"
#include "ot/base/Thread.h"
#include "ot/io/BufferedReader.h"
#include "ot/io/BufferedInputStream.h"
#include "ot/io/Console.h"
#include "ot/io/FileOutputStream.h"
#include "ot/io/InterruptedIOException.h"
#include "ot/io/InputStreamReader.h"
#include "ot/io/OutputStreamWriter.h"
#include "ot/net/InetAddress.h"

using namespace ot;
using namespace ot::net;
using namespace ot::io;

const size_t ClientTimeout = 0; // no timeout
const bool bEcho = false;
const bool bLog = true;
const bool bWelcomeMessage = true;

#define COUT Console::cout()
#define CERR Console::cerr()

ClientHandler::ClientHandler(Socket* pSocket, Listener* pListener) :
	m_rpListener(pListener),
	m_rpSocket(pSocket)
{
}

void ClientHandler::run()
{
	COUT << OT_T("Connected to client: ") << m_rpSocket->toString() << endl;

#ifdef OT_MT
	COUT << OT_T("The client thread is: ") << Thread::CurrentThread()->getName() << endl;
#endif // OT_MT

	RefPtr<OutputStream> rpSockOut = m_rpSocket->getOutputStream();
	RefPtr<OutputStream> rpLog = new FileOutputStream(OT_T("server.log"));
	RefPtr<Writer> rpSockWriter = new OutputStreamWriter(m_rpSocket->getOutputStream().get());
	RefPtr<InputStream> rpInput = new BufferedInputStream(m_rpSocket->getInputStream().get());
	
	try
	{
		if(ClientTimeout != 0)
			m_rpSocket->setSoTimeout(ClientTimeout);

		//
		// If requested, send out a welcome message to the client informing
		// them that they have connected to the OpenTop test server
		//
		if(bWelcomeMessage)
		{
			String welcome = OT_T("Welcome to the test server using OpenTop version ");
			welcome += System::GetVersionAsString() + OT_T(" running on ");
			welcome += InetAddress::GetLocalHost()->getHostName() + OT_T("\r\n");
			rpSockWriter->write(welcome);
			rpSockWriter->flush();
		}

		Byte buffer[1024];

		while(true)
		{
			int bytesRead = rpInput->read(buffer, sizeof(buffer));
			if(bytesRead == InputStream::EndOfFile)
			{
				break;
			}
			else
			{	
				COUT << StringUtils::FromLatin1((char*)buffer, bytesRead) << endl;

				if(bLog)
				{
					rpLog->write(buffer, bytesRead);
					rpLog->flush();
				}

				if(bEcho)
				{
					rpSockOut->write(buffer, bytesRead);
					rpSockOut->flush();
				}
			}
		}
	}
	catch(InterruptedIOException& /*e*/)
	{
		CERR << OT_T("Client idle for too long") << endl;
		if(bEcho)
		{
			try {
				static const String idle = OT_T("Too slow, sorry!\r\n");
				rpSockWriter->write(idle);
				rpSockWriter->flush();
			} catch(Exception&) {}
		}
	}
	catch(Exception& e)
	{
		CERR << OT_T("Client connection failed: ") << e.toString() << endl;
	}
	COUT << OT_T("Client connection terminated") << endl;
}
