//==============================================================================
// This file is part of the OpenTop C++ Library.
// (c) Copyright 2000-2003 ElCel Technology Limited
//
// This file may be distributed and/or modified under the terms of the
// GNU General Public License version 2 as published by the Free Software
// Foundation and appearing in the file LICENSE.GPL included in the
// packaging of this file.
//
// Licensees holding valid OpenTop Commercial licences may use this file
// in accordance with the OpenTop Commercial License Agreement included
// with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// See http://www.elcel.com/products/opentop/pricing.html or email
// sales@elcel.com for information about OpenTop Commercial License Agreements.
//
//==============================================================================
//
// $Revision: 1.2 $
// $Date: 2003/09/08 21:00:15 $
//
//==============================================================================
//
// tcpserver sample application
//
//==============================================================================

#include "Listener.h"

#include "ot/net/InetAddress.h"
#include "ot/net/ServerSocket.h"
#include "ot/net/Socket.h"
#include "ot/base/NumUtils.h"
#include "ot/base/Exception.h"
#include "ot/base/System.h"
#include "ot/base/Thread.h"
#include "ot/base/Tracer.h"
#include "ot/io/Console.h"
#include "ot/auxil/MemCheckSystemMonitor.h"
#include "ot/auxil/CommandLineParser.h"
#include "ot/auxil/BooleanOption.h"
#include "ot/auxil/TraceHelper.h"

#ifndef _MSC_VER
#include <signal.h>
#endif

using namespace ot;
using namespace ot::net;
using namespace ot::io;
using namespace ot::auxil;

bool bContinue = true;
const size_t ClientTimeout = 20000; // 20 secs

#define COUT Console::cout()
#define CERR Console::cerr()

void showUsage(const String& programName)
{
	COUT << OT_T("Usage: ") << programName << OT_T(" [option]... ") << endl << endl;
	COUT << OT_T("Test server.") << endl << endl;

	COUT << OT_T("  -h, --help                display this help") << endl;
	COUT << OT_T("  -p, --port                port to listen on") << endl;
	COUT << OT_T("  -r, --reusaddr            set the SO_REUSADDR option") << endl;
	COUT << OT_T("  -t, --timeout             milliseconds to wait for connection") << endl;
	COUT << OT_T("Copyright (c) 2002 - ElCel Technology Limited (http://www.elcel.com)") << endl;
}

int main(int argc, char* argv[])
{
	MemCheckSystemMonitor monitor;

	CommandLineParser cmdlineParser;

	BasicOption optHelp(OT_T("help"), 'h', BasicOption::none);
	BasicOption optReusaddr(OT_T("reusaddr"), 'r', BasicOption::none);
	BasicOption optPort(OT_T("port"), 'p', BasicOption::mandatory);
	BasicOption optTimeout(OT_T("timeout"), 't', BasicOption::mandatory);
	BasicOption optTrace(OT_T("trace"), 0, BasicOption::mandatory);

	cmdlineParser.addOption(&optHelp);
	cmdlineParser.addOption(&optPort);
	cmdlineParser.addOption(&optReusaddr);
	cmdlineParser.addOption(&optTimeout);
	cmdlineParser.addOption(&optTrace);

	try
	{
		cmdlineParser.parse(argc, argv);
	}
	catch (CommandLineException& e)
	{
		CERR << cmdlineParser.getProgramName() << OT_T(": ") << e.getMessage() << endl << endl;
		CERR << OT_T("Try ") << cmdlineParser.getProgramName() << OT_T(" --help") << endl;
		return (1);
	}

	//
	// If we have been asked to enable tracing, so it
	//
	if(optTrace.isPresent())
	{
		Tracer::SetTracer(new TraceHelper(cmdlineParser.getProgramName(), optTrace.getArgument()));
	}

	//
	// Determine local host info
	//
	RefPtr<InetAddress> rpLocal = InetAddress::GetLocalHost();
	COUT << OT_T("Server running on host ") << rpLocal->toString() << endl;

	if(optHelp.isPresent())
	{
		showUsage(cmdlineParser.getProgramName());
		return (0);
	}

	int port = 0;
	if(optPort.isPresent())
	{
		port = NumUtils::ToInt(optPort.getArgument());
	}

#if defined(OT_MT) && !defined(WIN32)
	Thread::SetInterruptSignal(SIGALRM);
#endif

	try
	{
		//
		// W. Richard Stevens' Unix Network Programming
		// suggests always setting SO_REUSADDR for TCP Servers.
		// This must be set before bind() so we must create an unbound socket
		// and then call bind() after setting the option
		//
		RefPtr<ServerSocket> rpServerSocket = new ServerSocket();
		rpServerSocket->setReuseAddress(optReusaddr.isPresent());
		rpServerSocket->bind(port);

		COUT << OT_T("Listening on socket: ") << rpServerSocket->toString() << endl;

		if(optTimeout.isPresent())
		{
			size_t timeout = NumUtils::ToInt(optTimeout.getArgument());
			rpServerSocket->setSoTimeout(timeout);
		}

		RefPtr<Listener> rpListener = new Listener(rpServerSocket.get());

		//
		// In multi-threaded builds the listener object is a Thread, but in single-threaded
		// builds it is a simple implemententation of Runnable.
		//
#ifdef OT_MT
		rpListener->start();
		rpListener->join();
#else
		rpListener->run();
#endif
	}
	catch(Exception& e)
	{
		CERR << e.toString() << endl;
	}

	COUT << OT_T("Terminating server") << endl;

	return 0;
}
