//==============================================================================
// This file is part of the OpenTop C++ Library.
// (c) Copyright 2000-2003 ElCel Technology Limited
//
// This file may be distributed and/or modified under the terms of the
// GNU General Public License version 2 as published by the Free Software
// Foundation and appearing in the file LICENSE.GPL included in the
// packaging of this file.
//
// Licensees holding valid OpenTop Commercial licences may use this file
// in accordance with the OpenTop Commercial License Agreement included
// with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// See http://www.elcel.com/products/opentop/pricing.html or email
// sales@elcel.com for information about OpenTop Commercial License Agreements.
//
//==============================================================================
//
// $Revision: 1.4 $
// $Date: 2003/09/08 21:00:15 $
//
//==============================================================================
//
// OpenTop Sample Application: tcpclient
//
// This simple console application will initiate a TCP connection to a specified
// host/port and send it a "Hello World!" message.
//
//==============================================================================

#include "ot/net/InetAddress.h"
#include "ot/net/DatagramPacket.h"
#include "ot/net/DatagramSocket.h"
#include "ot/base/NumUtils.h"
#include "ot/base/Exception.h"
#include "ot/base/StringUtils.h"
#include "ot/base/System.h"
#include "ot/base/Tracer.h"
#include "ot/io/Console.h"
#include "ot/auxil/MemCheckSystemMonitor.h"
#include "ot/auxil/CommandLineParser.h"
#include "ot/auxil/BooleanOption.h"
#include "ot/auxil/TraceHelper.h"

#ifndef _MSC_VER
#include <signal.h>
#endif

using namespace ot;
using namespace ot::net;
using namespace ot::io;
using namespace ot::auxil;

bool bContinue = true;
const size_t ClientTimeout = 20000; // 20 secs

#define COUT Console::cout()
#define CERR Console::cerr()

void showUsage(const String& programName)
{
	COUT << OT_T("Usage: ") << programName << OT_T(" [option]... ") << endl << endl;
	COUT << OT_T("Test server.") << endl << endl;

	COUT << OT_T("  -h, --help           display this help") << endl;
	COUT << OT_T("  -l, --localport      local port to bind to") << endl;
	COUT << OT_T("  -p, --port           port to send to") << endl;
	COUT << OT_T("  -r, --reply          wait for a reply (and echo it to standard out)") << endl;
	COUT << OT_T("  -s, --server         server to send to, if not specified broadcasts") << endl;
	COUT << OT_T("                       to local subnet") << endl;
	COUT << OT_T("  -t, --timeout        milliseconds to wait for reply") << endl;
	COUT << OT_T("Copyright (c) 2003 - ElCel Technology Limited (http://www.elcel.com)") << endl;
}

int main(int argc, char* argv[])
{
	MemCheckSystemMonitor monitor;

	//
	// Set up a command line parser to recognise our options...
	// 
	BasicOption optHelp(OT_T("help"), 'h', BasicOption::none);
	BasicOption optLocalPort(OT_T("localport"), 'l', BasicOption::mandatory);
	BasicOption optPort(OT_T("port"), 'p', BasicOption::mandatory);
	BooleanOption optReply(OT_T("reply"), 'r', false);
	BasicOption optServer(OT_T("server"), 's', BasicOption::mandatory);
	BasicOption optTimeout(OT_T("timeout"), 't', BasicOption::mandatory);
	BasicOption optTrace(OT_T("trace"), 0, BasicOption::mandatory);

	CommandLineParser cmdlineParser;
	cmdlineParser.addOption(&optHelp);
	cmdlineParser.addOption(&optLocalPort);
	cmdlineParser.addOption(&optPort);
	cmdlineParser.addOption(&optReply);
	cmdlineParser.addOption(&optServer);
	cmdlineParser.addOption(&optTimeout);
	cmdlineParser.addOption(&optTrace);

	//
	// ...parse the cpmmand line, placing the command options into their 
	// respective objects.
	//
	try
	{
		cmdlineParser.parse(argc, argv);
	}
	catch (CommandLineException& e)
	{
		CERR << cmdlineParser.getProgramName() << OT_T(": ") << e.getMessage() << endl << endl;
		CERR << OT_T("Try ") << cmdlineParser.getProgramName() << OT_T(" --help") << endl;
		return (1);
	}

	//
	// Determine local host info
	//
	RefPtr<InetAddress> rpLocal = InetAddress::GetLocalHost();
	COUT << OT_T("Client running on host ") << rpLocal->toString() << endl;

	if(optHelp.isPresent())
	{
		showUsage(cmdlineParser.getProgramName());
		return (0);
	}

	try
	{
		//
		// If we have been asked to enable tracing, create a TraceHelper (from the auxil namespace)
		// This little helper is initialized with a string from which it extracts what tracing
		// should be performed, and to where the output should be written.
		//
		if(optTrace.isPresent())
		{
			Tracer::SetTracer(new TraceHelper(cmdlineParser.getProgramName(), optTrace.getArgument()));
		}

		//
		// Determine the address and port to which we shall try and send
		//
		int port = 0;
		if(optPort.isPresent())
		{
			port = NumUtils::ToInt(optPort.getArgument());
		}

		int localPort = 0;
		if(optLocalPort.isPresent())
		{
			localPort = NumUtils::ToInt(optLocalPort.getArgument());
		}

		RefPtr<InetAddress> rpSendAddr;
		if(optServer.isPresent())
			rpSendAddr = InetAddress::GetByName(optServer.getArgument());
		else
			rpSendAddr = InetAddress::GetLocalBroadcast();

		COUT << OT_T("Sending UDP packet to server: ") 
		     << rpSendAddr->getHostName() 
		     << OT_T(" on port: ") << port << endl;

		RefPtr<DatagramSocket> rpDatagramSocket = new DatagramSocket;
		
		size_t timeoutMS = 0;
		if(optTimeout.isPresent())
			timeoutMS = NumUtils::ToInt(optTimeout.getArgument());

		if(timeoutMS)
			rpDatagramSocket->setSoTimeout(timeoutMS);

		//
		// Allocate a datagram packet with a 1024 byte internal buffer
		//
		DatagramPacket packet(1024);
		packet.setAddress(rpSendAddr.get());
		packet.setPort(port);

		//
		// Send a simple message to the server.
		// Note we are sending raw bytes in this case, not Unicode characters.
		//
		const char* hello = "Hello World!";
		strcpy((char*)packet.getData(), hello);
		packet.setLength(strlen(hello));

		rpDatagramSocket->send(packet);

		//
		// If a reply is expected, then wait for it (up to the timeout value is specified)
		//
		if(optReply.getValue())
		{
			COUT << OT_T("Waiting for reply on interface: ") 
			     << rpDatagramSocket->getLocalAddress()->toString() 
			     << OT_T(" port: ") 
			     << NumUtils::ToString(rpDatagramSocket->getLocalPort()) << endl;

			//
			// Set the length of the receive packet large enough for any expected reply
			// (otherwise the received msg will be truncated)
			//
			packet.setLength(packet.getBufferSize());
			rpDatagramSocket->receive(packet);

			COUT << OT_T("Reply from server: ") << StringUtils::FromLatin1((const char*)packet.getData(), packet.getLength()) << endl;
		}
	}
	catch(Exception& e)
	{
		CERR << e.toString() << endl;
	}

	COUT << OT_T("Terminating client") << endl;

	return 0;
}
