//==============================================================================
// This file is part of the OpenTop C++ Library.
// (c) Copyright 2000-2003 ElCel Technology Limited
//
// This file may be distributed and/or modified under the terms of the
// GNU General Public License version 2 as published by the Free Software
// Foundation and appearing in the file LICENSE.GPL included in the
// packaging of this file.
//
// Licensees holding valid OpenTop Commercial licences may use this file
// in accordance with the OpenTop Commercial License Agreement included
// with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// See http://www.elcel.com/products/opentop/pricing.html or email
// sales@elcel.com for information about OpenTop Commercial License Agreements.
//
//==============================================================================
//
// $Revision: 1.2 $
// $Date: 2003/09/08 21:00:15 $
//
//==============================================================================
//
// OpenTop Sample Application: udpserver
//
// This simple console application will wait for UDP datagram packets
// and (optionally) echo the message contents back to the sender.
//
// This application can also be used to test IP multicasting by joining to a
// specified multicast group address.
//
//==============================================================================

#include "ot/net/InetAddress.h"
#include "ot/net/DatagramPacket.h"
#include "ot/net/DatagramSocket.h"
#include "ot/net/MulticastSocket.h"
#include "ot/base/NumUtils.h"
#include "ot/base/Exception.h"
#include "ot/base/StringUtils.h"
#include "ot/base/System.h"
#include "ot/base/Tracer.h"
#include "ot/io/Console.h"
#include "ot/auxil/MemCheckSystemMonitor.h"
#include "ot/auxil/CommandLineParser.h"
#include "ot/auxil/BooleanOption.h"
#include "ot/auxil/TraceHelper.h"

#ifndef _MSC_VER
#include <signal.h>
#endif

using namespace ot;
using namespace ot::net;
using namespace ot::io;
using namespace ot::auxil;

bool bContinue = true;
const size_t ClientTimeout = 20000; // 20 secs

#define COUT Console::cout()
#define CERR Console::cerr()

void showUsage(const String& programName)
{
	COUT << OT_T("Usage: ") << programName << OT_T(" [option]... ") << endl << endl;
	COUT << OT_T("  -e, --[no]echo       echo received messages back to sender") << endl;
	COUT << OT_T("                       (toggle switch enabled by default)") << endl;
	COUT << OT_T("  -h, --help           display this help") << endl;
	COUT << OT_T("  -i, --interface      local IP address to bind to") << endl;
	COUT << OT_T("  -m, --multicast      multicast group to join, specified as a class D") << endl;
	COUT << OT_T("                       address (239.x.x.x).  See RFC2365 for details.") << endl;
	COUT << OT_T("  -p, --port           local port to bind to") << endl;
	COUT << OT_T("  -t, --timeout        milliseconds to wait for message") << endl;
	COUT << OT_T("Copyright (c) 2003 - ElCel Technology Limited (http://www.elcel.com)") << endl;
}

int main(int argc, char* argv[])
{
	MemCheckSystemMonitor monitor;

	//
	// Set up a command line parser to recognise our options...
	// 
	BasicOption optHelp(OT_T("help"), 'h', BasicOption::none);
	BasicOption optPort(OT_T("port"), 'p', BasicOption::mandatory);
	BooleanOption optEcho(OT_T("echo"), 'e', true);
	BasicOption optInterface(OT_T("interface"), 'i', BasicOption::mandatory);
	BasicOption optMulticast(OT_T("multicast"), 'm', BasicOption::mandatory);
	BasicOption optTimeout(OT_T("timeout"), 't', BasicOption::mandatory);
	BasicOption optTrace(OT_T("trace"), 0, BasicOption::mandatory);

	CommandLineParser cmdlineParser;
	cmdlineParser.addOption(&optHelp);
	cmdlineParser.addOption(&optPort);
	cmdlineParser.addOption(&optEcho);
	cmdlineParser.addOption(&optInterface);
	cmdlineParser.addOption(&optMulticast);
	cmdlineParser.addOption(&optTimeout);
	cmdlineParser.addOption(&optTrace);

	//
	// ...parse the cpmmand line, placing the command options into their 
	// respective objects.
	//
	try
	{
		cmdlineParser.parse(argc, argv);
	}
	catch (CommandLineException& e)
	{
		CERR << cmdlineParser.getProgramName() << OT_T(": ") << e.getMessage() << endl << endl;
		CERR << OT_T("Try ") << cmdlineParser.getProgramName() << OT_T(" --help") << endl;
		return (1);
	}

	//
	// Determine local host info
	//
	RefPtr<InetAddress> rpLocal = InetAddress::GetLocalHost();
	COUT << OT_T("UDP server running on host ") << rpLocal->toString() << endl;

	if(optHelp.isPresent())
	{
		showUsage(cmdlineParser.getProgramName());
		return (0);
	}

	try
	{
		//
		// If we have been asked to enable tracing, create a TraceHelper (from the auxil namespace)
		// This little helper is initialized with a string from which it extracts what tracing
		// should be performed, and to where the output should be written.
		//
		if(optTrace.isPresent())
		{
			Tracer::SetTracer(new TraceHelper(cmdlineParser.getProgramName(), optTrace.getArgument()));
		}

		//
		// Determine the i/face and port to which we shall bind
		//
		int port = 0;
		if(optPort.isPresent())
		{
			port = NumUtils::ToInt(optPort.getArgument());
		}

		RefPtr<InetAddress> rpInterface;
		if(optInterface.isPresent())
			rpInterface = InetAddress::GetByName(optInterface.getArgument());
		else
			rpInterface = InetAddress::GetAnyHost();

		RefPtr<DatagramSocket> rpDatagramSocket;
	
		//
		// If a multicast group has been specified, then attempt to join the group
		// so that we receive messages destined for the group.
		//
		// Note: For testing it is advisable to use a local address in the range
		// 239.255.0.0/16 (as described in RFC 2365).
		//
		if(optMulticast.isPresent())
		{
			RefPtr<InetAddress> rpMulticastAddr = InetAddress::GetByName(optMulticast.getArgument());
			RefPtr<MulticastSocket> rpMultiSocket = new MulticastSocket(port);
			if(rpInterface)
				rpMultiSocket->joinGroup(rpMulticastAddr.get(), rpInterface.get());
			else
				rpMultiSocket->joinGroup(rpMulticastAddr.get());
			rpDatagramSocket = rpMultiSocket.get();
		}
		else
		{
			rpDatagramSocket = new DatagramSocket(port, rpInterface.get());
		}
					
		size_t timeoutMS = 0;
		if(optTimeout.isPresent())
			timeoutMS = NumUtils::ToInt(optTimeout.getArgument());

		if(timeoutMS)
			rpDatagramSocket->setSoTimeout(timeoutMS);

		//
		// Allocate a datagram packet with a 1024 byte internal buffer
		//
		DatagramPacket packet(1024);

		while(true)
		{
			COUT << OT_T("Waiting for message on interface: ") 
			     << rpDatagramSocket->getLocalAddress()->toString() 
			     << OT_T(" port: ") 
			     << NumUtils::ToString(rpDatagramSocket->getLocalPort()) << endl;

			//
			// Set the length of the receive packet large enough for any expected reply
			// (otherwise the received msg will be truncated)
			//
			packet.setLength(packet.getBufferSize());
			rpDatagramSocket->receive(packet);

			COUT << OT_T("Received message from host: ") 
			     << packet.getAddress()->getHostName()
				 << OT_T(" Data: ")
			     << StringUtils::FromLatin1((const char*)packet.getData(), packet.getLength()) << endl;

			if(optEcho.getValue())
			{
				//
				// It's a piece of cake to send the packet back to the originating
				// client, the DatagramPacket already contains all the required 
				// information.
				// 
				rpDatagramSocket->send(packet);
			}
		}
	}
	catch(Exception& e)
	{
		CERR << e.toString() << endl;
	}

	COUT << OT_T("Terminating UDP server") << endl;

	return 0;
}
